/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	19-Oct-2011
 */

#include "userdump.h"

// ================================= UserDump ==================================
UserDump::UserDump(const char* filename, const char* mode, Format fmt)
{
	// default don't read/write anything
	flagSource = flagEnergy = flagTrack = false;
	if (filename != nullptr)
		open(filename, mode, fmt);
} // UserDump

/** UserDump */
UserDump::~UserDump()
{
	if (file) close();
} // UserDump

/** open */
bool UserDump::open(const char* filename, const char* amode, Format fmt)
{
	char mode[5];
	int ptr = 0;
	_format = fmt;
	mode[ptr++] = amode[0];
	mode[ptr] = 0;
	switch (fmt) {
		case Format::Fluka:
		case Format::Moira:
			mode[ptr++] = 'b';
			mode[ptr] = 0;
			break;
		default:
			/* do nothing */;
	}
	if (file.open(filename,mode)) {
		// on success read/write all unless changed by the user
		flagSource = flagEnergy = flagTrack = true;
		return true;
	}
	return false;
} // open

/** Read or skip next event from mgread structure */
bool UserDump::readEvent(int filter)
{
	if (_format==Format::Fluka) {
		char buffer[20];

		FortranParser parser(buffer, sizeof(buffer));

		// Read header
		if (file.read(parser) != 20) return false;
		event.generic.ndum = parser.readInt();
		event.generic.mdum = parser.readInt();
		event.generic.jdum = parser.readInt();
		event.generic.edum = parser.readFloat();
		event.generic.wdum = parser.readFloat();

		/* check event */
		if (event.generic.ndum > 0) {			// Tracking event
			type = MGDRAW_TRACK;
			if (filter & MGDRAW_TRACK)
				return readTracking();
			else
				file.skip();
		} else
		if (event.generic.ndum == 0) {			// Energy deposition event
			type = MGDRAW_ENERGY;
			if (filter & MGDRAW_TRACK)
				return readEnergy();
			else
				file.skip();
		} else {
			type = MGDRAW_SOURCE;
			event.source.ncase = -event.generic.ndum;	// Make positive
			if (filter & MGDRAW_SOURCE)
				return readSource();
			else
				file.skip();
		}
	} else
		assert(false);
	return true;
} // readEvent

/** readTracking */
bool UserDump::readTracking()
{
	assert(_format == Format::Fluka);
	/* read ntrack points */
	int size = sizeof(TrackPos) * (event.track.ntrack+1) +
		   sizeof(float) * event.track.mtrack +
		   sizeof(float);

	if (!file.mustBe(size)) return false;
	track.clear();
	for (int i=0; i<=event.track.ntrack; i++) {
		TrackPos pos;
		file.readraw(&pos, sizeof(pos), 1);
		track.push_back(pos);
	}
	for (int i=0; i<event.track.mtrack; i++) {
		float dh;
		file.readraw(&dh, sizeof(dh), 1);
	}
	file.readraw(&event.track.ctrack, sizeof(float), 1);
	if (!file.mustBe(size)) return false;

	return true;
} // readTracking

/** readEnergy */
bool UserDump::readEnergy()
{
	assert(_format == Format::Fluka);
	char buffer[16];
	FortranParser parser(buffer, sizeof(buffer));

	if (file.read(parser) != 16) return false;

	event.energy.x    = parser.readFloat();
	event.energy.y    = parser.readFloat();
	event.energy.z    = parser.readFloat();
	event.energy.edep = parser.readFloat();
	return true;
} // readEnergy

/** readSource */
bool UserDump::readSource()
{
	assert(_format == Format::Fluka);
	/* read npfluka particles */
	int size = sizeof(SourceParticle) * event.source.npfluka;
	if (!file.mustBe(size)) return false;

	stack.clear();

	for (int i=0; i<event.source.npfluka; i++) {
		SourceParticle part;
		size_t sz = fread(&part, sizeof(part), 1, file.handle);
		if (sz != 1) return false;
		stack.push_back(part);
	}
	return file.mustBe(size);
} // readSource

/** Write events to mgread structure */
void UserDump::writeHeader(int ndum, int mdum, int particle, float energy, float weight)
{
	assert(_format == Format::Fluka);

	char buffer[32];
	FortranParser parser(buffer, sizeof(buffer));
	parser.write(ndum);
	parser.write(mdum);
	parser.write(particle);
	parser.write(energy);
	parser.write(weight);
	file.write(parser);
} // writeHeader

/** Write events to mgread structure */
//void UserDump::writeEvent(int ndum, int mdum, int particle, float weight,
//			float prex,  float prey,  float prez,
//			float postx, float posty, float postz,
//			float energy)
//{
//	char	buffer[32];
//
//	writeHeader(ndum, mdum, particle, weight, energy);
//
//	if (ndum>0)
//		writeTracking(prex, prey, prez, postx, posty, postz, energy);
//	//else
//	//if (!ndum)
//		////std::cout << "Energy Dep" << std::endl;
//	else
//		writeSource(particle, energy, weight, prex, prey, prez);
//} // writeEvent

/** writeTracking */
void UserDump::writeTracking(float prex,  float prey,  float prez,
			     float postx, float posty, float postz,
			     float edep, float length)
{
	if (_format==Format::Fluka) {
		char buffer[128];
		FortranParser parser(buffer, sizeof(buffer) );
		parser.write(prex);
		parser.write(prey);
		parser.write(prez);
		parser.write(postx);
		parser.write(posty);
		parser.write(postz);
		parser.write(edep);
		parser.write(length);
		file.write(parser);
	} else
		assert(false);
} // writeTracking

/** write on source particle */
void UserDump::writeSource(	int particle, float energy, float weight,
				float x,  float y,  float z,
				float tx, float ty, float tz)
{
	if (_format==Format::Fluka) {
		char buffer[128];
		FortranParser parser(buffer, sizeof(buffer));
		parser.write(particle);
		parser.write(energy);
		parser.write(weight);
		parser.write(x);
		parser.write(y);
		parser.write(z);
		parser.write(tx);
		parser.write(ty);
		parser.write(tz);
		file.write(parser);
	} else
		assert(false);
} // writeSource
