////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#pragma once
#ifndef BXDF_BASE_H
#define BXDF_BASE_H

// FIXME: this should go into a general tools header
#define UNUSED(expr) (void)(expr)

#include <cmath>

#include "graph/color.h"
#include "math/randsampler.h"
#include "math/vector.h"
#include "math/random.h"
#include "geometry/ray.h"
#include "geometry/material.h"

inline double CosTheta( const Vector& Vec ) { return Vec.z; }
inline double SinTheta( const Vector& Vec ) { return sqrt( std::max( 0.0, 1.0 - Vec.z * Vec.z ) ); }
inline double SinTheta2( const Vector& Vec ) { return 1.0 - CosTheta( Vec ) * CosTheta( Vec ); }
inline double CosPhi( const Vector& Vec ) { return ( Vec.x / SinTheta( Vec ) ); }
inline double SinPhi( const Vector& Vec ) { return ( Vec.y / SinTheta( Vec ) ); }

class CBxDFBase
{
public:
	enum BxDFType {
		BSDF_REFLECTION = 1 << 0,
		BSDF_TRANSMISSION = 1 << 1,
		BSDF_DIFFUSE = 1 << 2,
		BSDF_GLOSSY = 1 << 3,
		BSDF_SPECULAR = 1 << 4,
		BSDF_ALL_TYPES = BSDF_DIFFUSE | BSDF_GLOSSY | BSDF_SPECULAR,
		BSDF_ALL_REFLECTION = BSDF_REFLECTION | BSDF_ALL_TYPES,
		BSDF_ALL_TRANSMISSION = BSDF_TRANSMISSION | BSDF_ALL_TYPES,
		BSDF_ALL = BSDF_ALL_REFLECTION | BSDF_ALL_TRANSMISSION
	};


	CBxDFBase( BxDFType Type ) : m_Type( Type ) {};
	virtual ~CBxDFBase( void ) {};

	// check if the current BxDF matches a certain type
	bool MatchesType( BxDFType Flags ) const { return (m_Type & Flags); };

	// f( Out, In) evaluates the BxDF for the in & outgoing pair of rays
	// Out = ray coming FROM the camera, In = ray going towards the light
	virtual double f( const Ray& Out, const Ray& In, const Vector& Normal, const Material& RenderMat, const Color& LightColor, double Power, float Shade = 1.0 ) = 0;

	// return corresponding PDF
	virtual double GetPDF( const Vector& Out, const Vector& In, const Vector& Normal ) const = 0;

	// Sample_f samples an incoming ray matching to the outgoing ray + its Pdf. Furthermore it evaluates the
	// BxDF according to In & Out and returns the color. This is used for reflection, refraction etc. which will
	// continue recursively.
	virtual double Sample_f( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, Random& Sampler, BxDFType Type, const Color& LightColor = Color( 1.0, 1.0, 1.0 ), double Power = 1.0, float Shade = 1.0 )
	{
		UNUSED( Type );

		Vector Intersection = Out.hit( -SMALL3D3 );
		Vector Dir, TmpVec;
		double pRand[2];

		Sampler.realqmcND( 2, pRand );

		// sample incoming ray
		Dir = CBaseSampler::cosineSampleHemisphere( Normal, pRand[0], pRand[1] );

		In.use_clip = Out.use_clip;
		RaySegment new_segment( Intersection, Dir, Out.prevZone() );
		new_segment.tmin = SMALL3D3;
		new_segment.tmax = INFINITE;
		In.push( new_segment );

		In.incDepth();

		// determine pdf which needs to be normalized to the hemi-sphere
		Pdf = 1.0 / PI;

		// default light color was set to white
		return f( Out, In, Normal, RenderMat, LightColor, Power, Shade );
	}


protected:
	const BxDFType m_Type;
};


inline Vector SphericalDirection( double SinTheta, double CosTheta, double Phi )
{
	return Vector( SinTheta * cos( Phi ), SinTheta * sin( Phi ), CosTheta );
}

inline Vector SphericalDirection( double SinTheta, double CosTheta, double Phi, const Vector& X, const Vector& Y, const Vector& Z )
{
	return X * SinTheta * cos( Phi ) + Y * SinTheta * sin( Phi ) + Z * CosTheta;
}

inline double SphericalTheta( const Vector& Vec ) {
	return acos( std::min<double>( std::max<double>( Vec.z, -1.0 ), 1.0 ) );
}

inline double SphericalPhi( const Vector& Vec ) {
	double p = atan2( Vec.y, Vec.x );
	return (p < 0.0) ? p + 2.0 * PI : p;
}


#endif
// vim:ts=2
