////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "bxdfdispersivetransmission.h"
#include "bxdffresnel.h"
#include "spectrahelpers.h"

#include <algorithm>

CBxDFDispersiveSpecTransmission::CBxDFDispersiveSpecTransmission( double DispersionPower, unsigned int TotalNrOfSamples, const Color& TransmissionColor /* = Color( 1.0, 1.0, 1.0 ) ) */ ) : CBxDFSpecTransmission( TransmissionColor ), m_DispersionPower( DispersionPower ),
m_CauchyA(0), m_CauchyB(0), m_CurSampleIndex( 0 ), m_TotalNrOfSamples( TotalNrOfSamples)
{
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CBxDFDispersiveSpecTransmission::Sample_f( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, Random& Sampler, BxDFType Type, const Color& LightColor /*= Color( 1.0, 1.0, 1.0 )*/, double Power /*= 1.0*/, float Shade /*= 1.0 */ )

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  UNUSED( Shade ); UNUSED( Power ); UNUSED( LightColor );

  if( Type == CBxDFBase::BSDF_TRANSMISSION )
    return Sample_f_Transmission( Out, In, Pdf, Normal, RenderMat, Sampler);

  return 0.0;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CBxDFDispersiveSpecTransmission::Sample_f_Transmission( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, Random& Sampler )

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
  Vector OutDir = Out.dir();
  Vector RefrNormal( Normal );

  // Figure out which $\eta$ is incident and which is transmitted
  double cosI = -RefrNormal.dot( OutDir );
  bool   bEntering = cosI > 0.;
  double EtaI = 1.0;            // could get this from ray
  double EtaT = RenderMat.ior();

  //////////////////////
  // handle dispersion
  // TODO: implement dispersion jittering by using a random offset
  double JitterOffset = Sampler.real(); // 0.5;
  Color  DispCol = m_TransmissionColor;
  m_CurSampleIndex++;

  /* A table of Cauchy coefficients for common optical materials is shown below: (http://en.wikipedia.org/wiki/Cauchy's_equation)
    Material A B (micro-m2)
    Fused silica 1.4580 0.00354
    Borosilicate glass BK7 1.5046 0.00420  --> B = 4200
    Hard crown glass K5 1.5220 0.00459     --> B = 4590
    Barium crown glass BaK4 1.5690 --> B = 5310
    Barium flint glass BaF10 1.6700 --> B = 7430
    Dense flint glass SF10 1.7280 --> B = 1342
  */

  //m_CauchyA = 1.5220;
  //m_CauchyB = 4590;
  //EtaT = CSpectraHelpers::GetIORColor( (m_CurSampleIndex + JitterOffset) / m_TotalNrOfSamples, m_CauchyA, m_CauchyB, DispCol);

  // Or we can calculate it like yafray. even though this is computationally expensive and not necessarily worth it
  //
  // ATTENTION: Triggering of the calculation of the cauchy coefficients in the PhysMat container class!
  //
  CSpectraHelpers::CauchyCoefficients( RenderMat.ior(), m_DispersionPower, m_CauchyA, m_CauchyB);
  EtaT = CSpectraHelpers::GetIORColor( (m_CurSampleIndex + JitterOffset) / m_TotalNrOfSamples, m_CauchyA, m_CauchyB, DispCol);
  EtaT = 1.0 + (EtaT / 100.0);

  //
  ///////////////////////

  if( !bEntering) {
    std::swap( EtaI, EtaT);
    EtaT = 1.0;                     // as a first assumption we go from the dense medium into air/vacuum
    RefrNormal = RefrNormal * -1.0;
  }

  double n = EtaI / EtaT;
  Pdf = 1.0;

  // Compute transmitted ray direction
  double cosT2 = 1.0 - n * n * (1.0 - cosI * cosI);
  if( cosT2 > 0.0) {
    // CVector T = (OutDir * n) + RefrNormal * (n * cosI - sqrtf( cosT2 )); // create refraction ray
    // removed slow the calculation of the square root
    Vector T = ( OutDir + RefrNormal * cosI) * n - RefrNormal * cosI;

	  // sample refraction ray
		In = Out;
		In.init();
		In.incDepth();	// to avoid infinite loops

		RaySegment NewSegment( Out.hit( SMALL3D3 ), T, Out.prevZone() );
		NewSegment.tmin = SMALL3D3;
		NewSegment.tmax = INFINITE;

		In.push( NewSegment );

  }
  else {
		In.depth = 32000;                                  // flag ray to be terminated
		return 0.0;
  }

  double F = m_pFresnel->Evaluate( cosI, EtaI, EtaT );

  // For the moment we fix the transmission color at white
  double FresnelRef = (EtaI * EtaI) / (EtaT * EtaT) * (1.0 - F );

  return FresnelRef / (-RefrNormal).dot( In.dir() );
}
// vim:ts=2
