////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// F A R M     -  Flair advanced rendering module
//
// Copyright European Organization for Nuclear Research( CERN )
// All rights reserved
//
// Authors: Christian.Theis@cern.ch
//			    Vasilis.Vlachoudis@cern.ch
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "bxdfphong.h"
#include "bxdffresnel.h"

CBxDFPhong::CBxDFPhong( void ) : CBxDFBase( CBxDFBase::BSDF_ALL ), m_pFresnelObj( new CFresnelDiel )
{
}

CBxDFPhong::~CBxDFPhong( void )
{
	delete m_pFresnelObj;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CBxDFPhong::f( const Ray& Out, const Ray& In, const Vector& Normal, const Material& RenderMat, const Color& LightColor, double Power, float Shade /*= 1.0*/ )

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( Out );
	UNUSED( LightColor );

	double ResColor = 1.0 * Shade;

	// add specular component
	if( RenderMat.shine() > 0.0 ) {
		double Dot = Out.dir() * Normal;
		Vector Reflect = (-2.0 * Dot) * Normal + Out.dir();
		double Spec = In.dir() * Reflect;

		if( Spec > 0.0 )
			ResColor += Power * pow( Spec, RenderMat.shine() );
	}

	return ResColor;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CBxDFPhong::Sample_f( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, Random& Sampler, BxDFType Type, const Color& LightColor /*= Color( 1.0, 1.0, 1.0 )*/, double Power /*= 1.0*/, float Shade /*= 1.0 */)

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( Power );

	if( Type == CBxDFBase::BSDF_REFLECTION )
		return Sample_f_Reflection( Out, In, Pdf, Normal, RenderMat, Sampler );
	else if( Type == CBxDFBase::BSDF_TRANSMISSION )
		return Sample_f_Transmission( Out, In, Pdf, Normal, RenderMat, Sampler );
	else
		return CBxDFBase::Sample_f( Out, In, Pdf, Normal, RenderMat, Sampler, Type, LightColor, Shade );

	return 0.0;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CBxDFPhong::Sample_f_Reflection( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, Random& Sampler )
////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( Sampler );

	Pdf = 1.0f;

	// CT: we only render reflections if we are at the beauty pass (step = 1)
	double dot = Normal * Out.dir();
	if( dot < 0.0 ) {
		Vector reflect = (-2.0 * dot) * Normal + Out.dir();

		// Find intersection with light
		In = Out;
		In.init();
		In.incDepth();	// to avoid infinite loops
		In.push( RaySegment( Out.hit( -SMALL3D4 * 10.0 ), reflect, Out.prevZone() ) );

		return f( Out, In, Normal, RenderMat );
	}

	return 0.0;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

double CBxDFPhong::Sample_f_Transmission( const Ray& Out, Ray& In, float& Pdf, const Vector& Normal, const Material& RenderMat, Random& Sampler )

////////////////////////////////////////////////////////////////////////////////////////////////////////////
{
	UNUSED( Sampler );

	Vector RefrNormal( Normal );
	Vector OutDir = Out.dir();

	// Figure out which $\eta$ is incident and which is transmitted
	double cosI = -RefrNormal.dot( OutDir );
	bool   bEntering = cosI > 0.;
	double EtaI = 1.0;			// to be physically precise this IOR should be carried by the outgoing ray --> TODO:
	double EtaT = RenderMat.ior();
	double n = 0.0;

	// check if the ray is incident, otherwise reverse the object normal
	if( !bEntering ) {
		std::swap( EtaI, EtaT );
		EtaT = 1.0;                     // as a first assumption we go from the dense medium into air/vacuum
		RefrNormal = RefrNormal * -1.0;
	}

	n = EtaI / EtaT;
	Pdf = 1.0;

	double cosT2 = 1.0 - n * n * (1.0 - cosI * cosI);
	if( cosT2 > 0.0f ) {
		Vector T = (OutDir * n) + RefrNormal * (n * cosI - sqrt( cosT2 )); // create refraction ray
		// TODO: removed the slow the calculation of the square root
		// Vector T = ( OutDir + RefrNormal * cosI) * n - RefrNormal * cosI;

		In = Out;
		In.init();
		In.incDepth();	// to avoid infinite loops

		RaySegment NewSegment( Out.hit( SMALL3D3) , T, Out.prevZone() );
		NewSegment.tmin = SMALL3D3;
		NewSegment.tmax = INFINITE;

		In.push( NewSegment );

	}
	else {  // total reflection
		In.depth = 32000;                                  // flag ray to be terminated
		return 0.0;
	}

	// Fresnel transmission,
	double Rho = 1.0 / fabs( cosI ) * (EtaI * EtaI) / (EtaT * EtaT) * (1.0 - m_pFresnelObj->Evaluate( cosI, EtaI, EtaT ));
	return Rho;

}
// vim:ts=2
