/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#ifndef __GEOMETRY_CBODY_H
#define __GEOMETRY_CBODY_H

#include "gbody.h"
#include "vbody.h"

class VZone;
#ifdef EXPERIMENTAL
class GeometryEngine;
#endif

/* ============================== CBody =============================== */
/** CBody cached viewer body class
 * WARNING: Do not mix 3D and 2D operations at the same time
 */
class CBody {
private:
const	GBody*	gbody;		// reference to Gbody
const	VBody*	vbody;		// reference to Vbody

	bool	checkInside;	// Result of the last inside check
	int	_checkId;	// instance checked Id
	int*	gCheckId;	// global current check Id

public:
	bool	tinverse;	// if [tmin,tmax] solution should be
				// used in inverse e.g. for concave objects
				// or cones
	double	tmin, tmax;	// ray intersection min, max distances
	Array<VZone*>	zones;  // zones that refer to this body

#if _DEBUG>2
private:
	double	xx,yy,zz;	// STRONG debugging remember position
	double	ddx,ddy,ddz;	// ... and direction
#endif

public:
	CBody(GBody* gb=NULL, VBody* vb=NULL) : gCheckId(NULL) { init(gb, vb); }

	void	init(GBody *gb=NULL, VBody *vb=NULL) {
				gbody = gb;
				vbody = vb;
				_checkId = -1;
				zones.clear();
			}

const	std::string&	name()	const;
const	char*	cname()		const;
const	GBody*	body()		const	{ return gbody; }
	int	id()		const	{ return gbody->id(); }

	// Inside2D version before use please take care of incBodyCheckId2D()
	bool	inside2D(
#ifdef EXPERIMENTAL
			 GeometryEngine* engine,
#endif
			 const double  x, const double  y, const double  z,
			 const double dx, const double dy, const double dz);

	// FIXME duplication should be merged with the inside(x,y,z, dx,dy,dz, t=0)
	bool	inside( const double  x, const double  y, const double  z,
			const double dx, const double dy, const double dz) {
				if (*gCheckId != _checkId) {
					checkInside = gbody->inside(x,y,z, dx,dy,dz);
					checked();
				}
				return checkInside;
			}

	/**
	 * Find if a point (x,y,z)+t*(dx,dy,dz) is inside the body or not, using the
	 * information from the distance
	 * Returns: true  inside
	 *          false outside
	 */
	bool	inside( const double  x, const double  y, const double  z,
			const double dx, const double dy, const double dz,
			const double t) {
				if (*gCheckId != _checkId) {
					tinverse = gbody->distance(x,y,z, dx,dy,dz, &tmin,&tmax);
					_checkId = *gCheckId;
#if _DEBUG>2
					xx = x;
					yy = y;
					zz = z;
					ddx = dx;
					ddy = dy;
					ddz = dz;
				} else {
					assert(x==xx);
					assert(y==yy);
					assert(z==zz);
					assert(dx==ddx);
					assert(dy==ddy);
					assert(dz==ddz);
#endif
				}

				if (tinverse)
					return (bool)(t<tmin || tmax<t);
				else
					return (bool)(tmin<=t && t<=tmax);
			}

	bool	inside(const Vector& p, const Vector &d, const double t)
			{ return inside(p.x,p.y,p.z, d.x,d.y,d.z, t); }

	/** distance cached
	 * calculates the tmin, tmax and tinverse variables from the intersection
	 * of the ray with the body
	 * @param x,y,z		position vector
	 * @param dx,dy,dz	direction vector (has to be normalized)
	 */
	bool	distance(const double  x, const double  y, const double  z,
			 const double dx, const double dy, const double dz) {
				if (*gCheckId != _checkId) {
					tinverse  = gbody->distance(x,y,z, dx,dy,dz, &tmin,&tmax);
					_checkId  = *gCheckId;
#if _DEBUG>2
					xx = x;
					yy = y;
					zz = z;
					ddx = dx;
					ddy = dy;
					ddz = dz;
				} else {
					assert(x==xx);
					assert(y==yy);
					assert(z==zz);
					assert(dx==ddx);
					assert(dy==ddy);
					assert(dz==ddz);
#endif
				}
				return (bool)(tmin<tmax);
			}
	bool	distance(const Vector& p, const Vector &d)
			{ return distance(p.x,p.y,p.z, d.x,d.y,d.z); }

	// Checkid
	void	checkId(int* gid)	{ gCheckId = gid; _checkId = -1; }
	void	resetCheckId()		{ _checkId  = -1; }
	void	checked()		{ _checkId  = *gCheckId; }
	bool	isChecked()		{ return *gCheckId == _checkId; }

	/** add zones that refers to this body
	 * @param zone	to be added to the list
	 */
	void	addZone(VZone* zone)	{ zones.add(zone); }
	size_t	memory()	const	{ return sizeof(CBody) + zones.memory(); }
}; // CBody
#endif
