/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#include "engine.h"

#include <iomanip>

#include "ray.h"
#include "kernel.h"
#include "math/point.h"
#include "graph/viewport.h"

using std::cout;
using std::endl;

/* ========================== GeometryEngine ========================== */
/** GeometryEngine */
GeometryEngine::GeometryEngine(const Geometry* geo, GeometryKernel* k) :
	_geometry(geo),
	_kernel(k),
	zonesSorted(32),
	boundingVolTree( nullptr ),
	irurAccel(32)
{
	gBodyCheckId    = 0;
	gBodyMaxCheckId = 0;
	nClipBodies     = 0;
	// Accelerators
	irurAccel.compare(CBodyOrderAccel::compare);
	lastIRURCheckID = -1;
} // GeometryEngine

/** ~GeometryEngine */
GeometryEngine::~GeometryEngine()
{
	bodies.clear();
	zonesSorted.clear();
} /* ~GeometryEngine */

/** view */
ViewPort& GeometryEngine::view() const
{
	return _kernel->view;
} // view

/** getRegion */
VRegion* GeometryEngine::getRegion(const int id)
{
	return _kernel->getRegion(id);
} // getRegion

/** cleanup
 * Do not delete the data, simply destroy the links V[Body/Region] -> G[Body/Region]
 */
void GeometryEngine::cleanup()
{
	for (auto body : bodies)
		body.init(NULL);
} // cleanup

#ifdef EXPERIMENTAL
/** clearProximity
 * increase checkid of proximity bodies and clear the proximity list
 */
void GeometryEngine::clearProximity()
{
	for (auto prox : proximity) {
		DUMP(cout << "clearProximity(" << prox->name() << ")" << endl);
		prox->resetCheckId();
	}
	proximity.clear();
} // clearProximity
#endif

// Convert the Bodies,Zones,Regions to Array(VBody,VZone,VRegion)
// do not delete anything but reassign
//
/** derive - derive the VBodies, VZones and VRegions from Geometry
 * It initializes the list of bodies, zones and regions in the Viewer,
 * note that it deletes all previous objects and creates new ones
 */
void GeometryEngine::derive()
{
	// WARNING: bodies and regions vectors can contain more or equal
	// number of VBodies vs _geometry->bodies/regions

	// Synchronize the bodies
	bodies.clear();
	bodies.resize(_geometry->nBodies());
	for (size_t i=0; i<_geometry->nBodies(); i++) {
		GBody* gbody = _geometry->bodies[i];
		VBody* vbody = _kernel->bodies[i];
		assert(gbody == vbody->body());
		bodies[i].init(gbody, vbody);
		bodies[i].checkId(&gBodyCheckId);
	}

	// Rebuild zones caching array
	zonesSorted.clear();
	for (size_t i=0; i<_geometry->nRegions(); i++) {
		VRegion* region = _kernel->getRegion(i);
		for (int j=0; j<region->zones().size(); j++) {
			VZone* zone = region->zones()[j];
			zonesSorted.add(zone);
			fillCBodyZoneReferences(zone);
		}
	}

	buildBVH();
} // derive

/** fillCBodyZoneReferences */
void GeometryEngine::fillCBodyZoneReferences(VZone* zone)
{
	// Fill Body/Zone references
	for (int i=0; i<zone->size(); i++) {
		const GBody* gbody = zone->gexpr(i);
		if (gbody->isBody())
			getBody(gbody)->addZone(zone);
	}
} // fillCBodyZoneReferences

/** find zone where point is
 * WARNING this routine is optimized for points close to the 2D viewport ONLY!
 * @param x,y,z		location to search
 * @param dx,dy,dz	direction in case we are on the boundary
 * @param zone          candidate zone to be checked first
 * @return first zone that belongs to x,y,z with direction (dx,dy,dz)
 */
VZone* GeometryEngine::where2D(const double  x, const double  y, const double  z,
				const double dx, const double dy, const double dz,
				VZone* zone)
{
	if (zone)
		if (zone->inside2D(this, x,y,z, dx,dy,dz)) return zone;

	for (int i=_geometry->regions.size(); i--; ) {
		VZone* vzone = getRegion(i)->inside2D(this, x,y,z, dx,dy,dz);
		if (vzone) return vzone;
	}
	return NULL;
} // where2D

/** find zone where point is
 * WARNING this routine is optimized for points close to the 2D viewport ONLY!
 * @param p		location structure to search
 * @return first zone that belongs to x,y,z with direction (dx,dy,dz)
 */
int GeometryEngine::where2D(ZoneOfPoint* p)
{
	// FIXME use the "valid/invalid" information
	// if body is not modified but from the previous check
	// then do not check it's zones...
	// WARNING: e.g the second body can be modified, by the check below
	//              then is important to keep it!!!
	p->init();
#ifdef EXPERIMENTAL
	clearProximity();
#endif
	if (p->nOld != 0) {
		// First check all zones referring to that body
		for (int i=p->from; i<=p->to; i++) {
			VBody* body = (*p->V)[i].body;
			if (body==NULL) continue;
			CBody* cbody = getBody(body);
			for (int j=0; j<cbody->zones.size() && p->more(); j++) {
				IFSTAT(stat("where_body"));
				VZone* zone = cbody->zones[j];
				if (zone->inside2D(this, p->x,p->y,p->z, p->dx,p->dy,p->dz)) {
					IFSTAT(stat("where_body_found"));
					p->add(zone);
				}
			}
		}

		// loop over body zones
		for (int i=0; i<getBody(p->body)->zones.size() && p->more(); i++) {
			IFSTAT(stat("where_body"));
			VZone* zone = getBody(p->body)->zones[i];
			if (zone->inside2D(this, p->x,p->y,p->z, p->dx,p->dy,p->dz)) {
				IFSTAT(stat("where_body_found"));
				p->add(zone);
			}
		}

		// Then all zones in zoneOld
		for (int i=0; i<p->nOld && p->more(); i++) {
			IFSTAT(stat("where_zoneold"));
			if (p->zoneOld[i]->inside2D(this, p->x,p->y,p->z, p->dx,p->dy,p->dz)) {
				IFSTAT(stat("where_zoneold_found"));
				p->add(p->zoneOld[i]);
			}
		}
	}

	// Nothing found check all the remaining zones
	if (p->n==0) {
		IFSTAT(stat("where_zone_none"));
		// We should check for touching Conics and
		// scan all regions sharing a conic...
		// otherwise it wont find possible errors
		// with touching surfaces
		// Check everything...
		//
		// Check all zones sorted
		for (int i=zonesSorted.size(); i-- && p->more(); ) {
			VZone* zone = zonesSorted[i];
			if (zone->inside2D(this, p->x,p->y,p->z, p->dx,p->dy,p->dz)) {
				// promote zone
				if (i<zonesSorted.top()) {
					zonesSorted.erase(i);
					zonesSorted.add(zone);
				}
				p->add(zone);
			}
		}
	}

	return p->n;
} // where2D

/** find zone where a 3D point is
 * @param x,y,z		location to search
 * @param dx,dy,dz	direction in case we are on the boundary
 * @param zone          candidate zone to be checked first
 * @return first zone that belongs to x,y,z with direction (dx,dy,dz)
 */
// FIXME duplication should be merged with the whereRay(x,y,z, dx,dy,dz, t=0)
VZone* GeometryEngine::where(const double  x, const double  y, const double  z,
			     const double dx, const double dy, const double dz,
			     VZone* zone)
{
	if (zone)
		if (zone->inside(this, x,y,z, dx,dy,dz)) return zone;

#if defined( USE_BBOX) && defined(USE_BBOX_HIERARCHIES)
	VZone* pZone = boundingVolTree->inside( Vector( x, y, z ) );
	if (pZone) return pZone;
#else
	for (int i=_geometry->nRegions(); i--; ) {
		VZone* vzone = getRegion(i)->inside(this, x,y,z, dx,dy,dz);
		if (vzone) return vzone;
	}
	return NULL;
#endif
} // where

/** find zone where point is in the transformed system
 * @param x,y,z		location to search
 * @param dx,dy,dz	direction in case we are on the boundary
 * @param t		distance to check
 * @param zone		current zone to check
 * @param body		check zones linked to body
 * @return first region that belongs to (x+dx*t,y+dy*t,z+dz*t)
 */
VZone* GeometryEngine::whereRay(const double  x, const double  y, const double  z,
				const double dx, const double dy, const double dz,
				const double t, VZone* zone, CBody* body)
{
	IFSTAT(stat("whereRay.01 call"));

#ifdef USE_BBOX
	auto px = x + dx*t;
	auto py = y + dy*t;
	auto pz = z + dz*t;
	const double tol=SMALL4*(Abs(px)+Abs(py)+Abs(pz));
#endif
	// Check current zone if specified
	if (zone) {
		IFSTAT(stat("whereRay.02 zoneCheck"));
		if (zone->bbox().isValid() and
		    zone->bbox().inside(px,py,pz,tol) and
		    zone->inside(this, x,y,z, dx,dy,dz, t)) {
			IFSTAT(stat("whereRay.03_zoneHit"));
			return zone;
		}
	}

	// Check zones linked to body if any
	if (body) {
		IFSTAT(stat("whereRay.04_bodyCheck"));
		for (int i=body->zones.size(); i--; ) {
			VZone* zzz = body->zones[i];
			IFSTAT(stat("whereRay.05_body-zoneCheck"));
#ifdef USE_BBOX
			if (zzz->bbox().isValid() and
			    not zzz->bbox().inside(px,py,pz,tol)) {
				assert(not zzz->inside(this, x,y,z, dx,dy,dz, t));
				IFSTAT(stat("whereRay.06_body-zone-bbox"));
				continue;
			}
#endif
			if (zzz->inside(this, x,y,z, dx,dy,dz, t)) {
				IFSTAT(stat("whereRay.07_body-Hit"));
				IFSTAT(stat.inc("whereRay.08_body-zoneIndex_sum",
						(body->zones.size() - i)));
				if (i<body->zones.top()) {
					IFSTAT(stat("whereRay.09_body-zone-promote"));
					// promote zone
					body->zones.erase(i);
					body->zones.add(zzz);
				}
				return zzz;
			}
		}
	}

#if defined( USE_BBOX) && defined(USE_BBOX_HIERARCHIES)
	VZone* pZone = boundingVolTree->inside(Vector(x, y, z));
	if (pZone) return pZone;
#else
	// Check everything...
	IFSTAT(stat("whereRay.10_check-all"));
	for (int i=zonesSorted.size(); i--;) {
		VZone* zzz = zonesSorted[i];
		IFSTAT(stat("whereRay.11_check-all_zone"));
#ifdef USE_BBOX
		if (zzz->bbox().isValid() and
		    not zzz->bbox().inside(px,py,pz,tol)) {
			//if (zzz->inside(this, x,y,z, dx,dy,dz, t))
			//	zzz->bbox().inside(px,py,pz,tol);
			assert(not zzz->inside(this, x,y,z, dx,dy,dz, t));
			IFSTAT(stat("whereRay.12_check-all_bbox"));
			continue;
		}
#endif
		if (zzz->inside(this, x,y,z, dx,dy,dz, t)) {
			IFSTAT(stat("whereRay.13_check-all_hit"));
			//int idx = iter.index()+1;
			//IFSTAT(stat.inc("whereRay_zoneIndex_sum",
			//		(zones.size() - idx)));
			if (i<zonesSorted.top()) {
				// promote zone
				zonesSorted.erase(i);
				zonesSorted.add(zzz);
				if (body) body->zones.add(zzz);
			}
			return zzz;
		}
		// No need to check the zones
	}
	IFSTAT(stat("whereRay.14_error"));
#endif
	return NULL;
} // whereRay

/** fillIRURegionAccel
 * Fill Accelerator structure for intersectRayUndefinedRegion
 * Initializes a list of bodies ordered by the distance of their intersections
 * with the current ray.
 * Checks if last checkID is still valid and if so skips all computations
 */
void GeometryEngine::fillIRURegionAccel(Ray* ray)
{
	if (gBodyCheckId == lastIRURCheckID) return; // Previous list still valid
	lastIRURCheckID = gBodyCheckId;

	irurAccel.clear();
	RaySegment& segment = ray->segment();
	double tmin = segment.tmin;
	double tmax = segment.tmax;

	for (size_t ib=0; ib<_geometry->nBodies(); ib++) {
		CBody* body = getBody(ib);
		if (body->zones.empty()) continue;
		if (body->distance(segment.pos, segment.dir)) {
			CBodyOrderAccel accel;
			if (InRangeOpen(tmin, body->tmin, tmax)) { // Add entry point
				accel.t = body->tmin;
				accel.body = body;
				irurAccel.add(accel);
			}
			if (InRangeOpen(tmin, body->tmax, tmax)) { // Add exit point
				accel.t = body->tmax;
				accel.body = body;
				irurAccel.add(accel);
			}
		}
	}
} // fillIRURegionAccel

/** intersectRayUndefinedRegion
 * assume an empty region as transparent and check for next intersection
 * @param ray	ray structure filled
 * @return false if ray arrives to end without hitting anything
 *	   true  if ray hits any defined region
 *
 * NOTE: Similar logic has the method Zone::distance
 */
bool GeometryEngine::intersectRayUndefinedRegion(Ray* ray)
{
	IFSTAT(stat("intersectRayUndefined_call"));

	if (ray->clip_insight && applyClip(ray)) {
		RaySegment& segment = ray->segment();
		if (ray->ended()) return false;
		if (ray->n==0) ray->prevZone(segment.zone);	// update previous/master zone
		if (segment.zone) return true;
	}

	RaySegment& segment = ray->segment();
	double tmin = segment.tmin;

	fillIRURegionAccel(ray);
	ArrayIterator<CBodyOrderAccel> iter(irurAccel);
#if _DEBUG>1
//	double tmin = segment.tmin;
#endif
	while (iter) {
		CBodyOrderAccel& accel = iter++;
		if (accel.t < segment.tmin) continue;
		segment.tmin = accel.t;
		segment.body = accel.body;

		for (int iz=segment.body->zones.top(); iz>=0; iz--) {
			VZone* zone = segment.body->zones[iz];
			double t = segment.tmin*(1.0+SMALL3D1);
			IFSTAT(stat("whereRay_zoneIndex_sum"));
			if (zone->inside(this, segment.pos, segment.dir, t)) {
				IFSTAT(stat("whereRay_errorHit"));
				// Lattices are trickier than this
				//if (zone->inside(ray->segment(0).pos, ray->segment(0).dir, ray->getT())) {
				segment.tmin = t;
				segment.zone = zone;
#if _DEBUG>1
//				std::cerr << "Saved ray moving by " << segment.tmin - tmin << std::endl;
#endif
				if (!Eq0(t-tmin, SMALL3D1*(ray->pos().max()+1)))
					ray->error = true;
				return true;
			}
		}
	}
#if _DEBUG>1
//	std::cerr << "Cannot save ray" << std::endl;
#endif
	ray->error = true;
	return false;
} // intersectRayUndefinedRegion

/** check4clip
 * First find if there is a clipping body in sight...
 */
void GeometryEngine::check4clip(Ray* ray)
{
	ray->clip_insight     = false;

	const RaySegment& segment = ray->segment(0);
	for (size_t i=0; i<nClipBodies; i++) {
		int bid = clipBodiesId[i];
		assert(bid>=0 and bid<(int)bodies.size());
		CBody* cbody = getBody(bid);

		bool intersect = cbody->distance(segment.pos, segment.dir);
		if (intersect || clipNegative[i]) {
			bool tinv;
			if (clipNegative[i])
				tinv = !cbody->tinverse;
			else
				tinv = cbody->tinverse;

			if (!tinv) {
				double a = Max(cbody->tmin, segment.tmin);
				double b = Min(cbody->tmax, segment.tmax);
				ray->clip_insight |= a<b;
			} else {
				ray->clip_insight |= segment.tmin < Min(cbody->tmin, segment.tmax);
				ray->clip_insight |= Max(cbody->tmax, segment.tmin) < segment.tmax;
			}
		}
	}
} // check4clip

/** @return true if the clip body was applied */
inline bool GeometryEngine::applyClip(Ray* ray)
{
	assert(ray->clip_insight);
	RaySegment &segment = ray->segment();
	double tpos = ray->T();
	ray->clip_hit = false;
	//if (ray->clip_hit)
	//	ray->clip_hit = segment.tmin > SMALL3D;

	// loop until we are out of all clips
	bool more;
	do {
		more = false;
		for (size_t i=0; i<nClipBodies; i++) {
			int bid = clipBodiesId[i];
			assert(bid>=0 and bid<(int)bodies.size());
			CBody* cbody = getBody(bid);
			bool tnormal = clipNegative[i]? cbody->tinverse : !cbody->tinverse;
			if (tnormal) {	// normal clip
				if (InRange(cbody->tmin, tpos, cbody->tmax)) {
					// Hit inside the clip body, move to clip end and continue
					segment.tmin = (cbody->tmax - ray->Tsum())*(1.0+SMALL3D1);
					if (segment.tmax <= segment.tmin) {
						segment.tmin = segment.tmax;
						ray->clip_hit = true;
						break;
					}
					ray->clip_hit = more = true;
					tpos = ray->T();
					segment.zone = whereRay(segment.pos,
							segment.dir,
							segment.tmin);
					if (segment.zone) segment.body = cbody;
				}
			} else	// clip with concave or outside of body
			if (tpos < cbody->tmin) {
				// Hit inside cbody (first segment),
				// move to boundary and continue
				segment.tmin = (cbody->tmin - ray->Tsum())*(1.0+SMALL3D1);
				if (segment.tmax <= segment.tmin) {
					segment.tmin = segment.tmax;
					ray->clip_hit = true;
					break;
				}
				ray->clip_hit = more = true;
				tpos = ray->T();
				segment.zone = whereRay(segment.pos,
							segment.dir,
							segment.tmin);
				if (segment.zone)
					segment.body = cbody;
			} else	// clip with concave or outside of body
			if (tpos >= cbody->tmax) {
				// Hit inside cbody (second segment)
				// therefore it will never leave
				segment.tmin = segment.tmax;
				ray->clip_hit = true;
				break;
			}
		}
	} while (more && !segment.ended());
	check4clip(ray);	// check again if we need to use clip
	return ray->clip_hit;
} // applyClip

/** check4project
 * First find if there is a projecting body in sight...
 */
void GeometryEngine::check4project(Ray* ray)
{
	ray->project_insight = false;
	const RaySegment& segment = ray->segment(0);
	for (auto pid : projectBodiesId) {
		assert(pid>=0 and pid<(int)bodies.size());
		CBody* pbody = getBody(pid);
		if (pbody->distance(segment.pos, segment.dir)) {
			ray->project_insight |=
					InRange(segment.tmin, pbody->tmin, segment.tmax) ||
					InRange(segment.tmin, pbody->tmax, segment.tmax);
		}
	}
} // check4project

/** applyProject */
inline bool GeometryEngine::applyProject(Ray* ray, const double prevT)
{
	assert(ray->project_insight);
	RaySegment &segment = ray->segment();
	double T = ray->T();
	ray->project_hit = false;
	for (auto pid : projectBodiesId) {
		assert(pid>=0 and pid<(int)bodies.size());
		CBody* pbody = getBody(pid);
		if (InRange(prevT, pbody->tmin, T)) {
			ray->project_hit = true;
			T = pbody->tmin;
			segment.tmin = pbody->tmin - ray->Tsum();
			segment.body = pbody;
		}
		if (InRange(prevT, pbody->tmax, T)) {
			ray->project_hit = true;
			T = pbody->tmax;
			segment.tmin = pbody->tmax - ray->Tsum();
			segment.body = pbody;
		}
	}
	check4project(ray);	// check again if we need to use project
	return ray->project_hit;
} // applyProject

/** intersectLattice
 * @return	true  if lattice was applied
 *		false if an error occurred
 */
inline bool GeometryEngine::intersectLattice(Ray* ray)
{
	RaySegment& segment = ray->segment();
	VZone* zone = segment.zone;
	VRegion* region = zone->region();
	assert(region->region()->hasMatrix());
	RaySegment new_segment(region->region());

	// Find exit distance location from lattice
	double texit = segment.tmin;

	do {
		// find exit from zone
		CBody* body = zone->distance(this,
					 segment.pos,
					 segment.dir,
					&texit,
					 segment.tmax);

		// end of the ray?
		if (body==NULL) break;

		// check if we exited the region
		zone = region->inside(this,
					segment.pos,
					segment.dir,
					texit);
	} while (zone!=NULL);

	// find new limits in the lattice
	// from 0 to Min(texit,tmax) - tmin
	// with a small shift
	new_segment.tmin = SMALL3D2;
	new_segment.tmax = Min(texit, segment.tmax) - segment.tmin;

	// Transform ray at the location that enters the lattice
	// to the prototype and come back to screen coordinates
	// Transformation: M^(-1) * R * M
	// M = screen transformation
	// R = rotdefi transformation
	new_segment.pos = new_segment.matrix() * (Point)(segment.pos
				+ segment.tmin * segment.dir);
	new_segment.acc = 8.0*SMALL3D * ( Abs(segment.pos.x)
					+ Abs(segment.pos.y)
					+ Abs(segment.pos.z)
					+ Abs(new_segment.matrix()(0,3))
					+ Abs(new_segment.matrix()(1,3))
					+ Abs(new_segment.matrix()(2,3)));
	new_segment.dir = new_segment.matrix() * segment.dir;
	new_segment.dir.renormalize();

	// Find the new zone in the prototype
	segment.bodyCheckId = incBodyCheckId();
	new_segment.zone = whereRay(new_segment.pos,
				    new_segment.dir,
				    new_segment.tmin);

	if (new_segment.zone == NULL) {
		// error in geometry, move to texit
		segment.zone = NULL;
		return false;
	}

	// start with the original hit body
	//if (ray->clip_hit)
	//	new_segment.body = ray->segment().body;

	if (ray->push(new_segment)) { // segment is no longer the last segment
		_kernel->error("GeometryEngine::intersectLattice n>=MAXLEVEL. Possibly LATTICE infinite loop");
		bodyCheckId(segment.bodyCheckId);
		segment.zone = NULL;
		return false;
	}
	return true;
} // intersectLattice

/** intersectVoxel
 * @return	 1	hit found or ended
 *		-1	skip this zone
 *		-2	exit from voxel
 */
inline int GeometryEngine::intersectVoxel(Ray* ray)
{
	if (_kernel->voxel.invalid()) return -2;

	// Normal tracking, find next intersection
	RaySegment& segment = ray->segment();
	VZone* zone = segment.zone;

	double tentry = segment.tmin;
VOX:
	double tmin = segment.tmin;
	double T = ray->T();
	segment.tmin -= segment.tmin * SMALL3D;	// move backwards a bit

	if (_kernel->voxel.distance(
			segment.pos,
			segment.dir,
			&segment.tmin)) {

		// if previous hit was clip
		if (ray->clip_hit && segment.tmin - tmin < SMALL3D4) return 1;

		// Apply clip
		if (ray->clip_insight && applyClip(ray)) {
			if (ray->project_insight && applyProject(ray, T)) return 1;
			if (ray->ended()) return 1;
			if (segment.zone != zone) return -1;
			// continue to check if we hit on a filled voxel
			goto VOX;
		}

		if (ray->project_insight && applyProject(ray, T)) return 1;
		if (segment.tmin - tmin > SMALL3D4) segment.body = NULL;
		// either we hit the clipbody or the voxel, update master zone
		if (ray->n==0) ray->prevZone(segment.zone);

		// we are outside the region
		ray->moveby(SMALL3D);	// move a bit to ensure we are in the new voxel
		return 1;
	}

	// restore tmin and exit voxel zone
	segment.tmin = tentry;
	return -2;
} // intersectVoxel

/** intersectRay
 * find the next intersection of the ray with any zone and return.
 * It takes care of multiple-lattices as well as voxels
 * @param ray	structure contains the ray/depth
 * @param step	if true, the function returns after every step otherwise only
 *		when hitting a non-transparent object
 * @return true	 if we arrived to the end of the ray or error (and
 *		 ray->segment().zone will be NULL)
 *	   false the ray hit a solid object or next step was computed and there
 *		 is still way to go (if step=true)
 */
bool GeometryEngine::intersectRay(Ray* ray, bool step)
{
	IFSTAT(stat("intersectRay_call"));

	if (!step) {
		if (ray->use_clip && nClipBodies) {
			ray->clip_hit = false;
			check4clip(ray);
		}
		if (ray->use_project && projectBodyCount()) {
			ray->project_hit = false;
			check4project(ray);
		}
	}

	// Loop until we exit the region (zone->region())
	while (true) {
		// Check for ending condition
		// XXX it doesn't work for negative rays!!!
		RaySegment& segment = ray->segment();

		// Segment has ended
		if (segment.ended()) {
			// Initial segment and ended, we arrived to the end
			if (ray->n == 0) return true;

			ray->pop();
			ray->segment().tmin += segment.tmax;

			RaySegment &s = ray->segment();
			if (segment.region) {
				// find new zone
				bodyCheckId(s.bodyCheckId);
				s.zone = whereRay(s.pos,
						  s.dir,
						  s.tmin,
						  s.zone,
						  NULL);
			}

			if (step) return false;
			continue;
		}

		VZone* zone = segment.zone;

		// Check for error
		if (zone == NULL) {
			if (!intersectRayUndefinedRegion(ray))
				return true;
			else
				continue;
		}

		// Check for clipping body
		if (!step) {
			if (zone->region()->type() == GRegion::Type::BLACKHOLE &&
			    (!ray->skip_1stblack || ray->prevZone() != NULL))
				return false;

			// Inside a solid object
			// TODO: Review this condition
			if (!ray->skip_current && zone->region()->type() == GRegion::Type::NORMAL &&
					(!ray->skip_transparent || !zone->region()->transparent() )) {

				double T = ray->T();
				// Check for clipping body
				if (ray->clip_insight && applyClip(ray)) {
					if (ray->project_insight && applyProject(ray, T))
						return ray->ended();
					if (ray->ended()) return true;
					if (ray->n==0) ray->prevZone(zone);	// update master zone
					continue;
				} else
					return false;
			}
			ray->skip_current = false;
		}

		// Only increment stats when considering a new intersection or segment
		IFSTAT(stat("intersectRay_retry"));

		// ------------------- LATTICE -------------------
		// Could be that a region is marked as LATTICE but there is
		// no transformation associated with it
		if (zone->gregion()->hasMatrix()) {
			if (intersectLattice(ray) && step) return false;
			continue;
		} else	// end of LATTICE
		// ------------------- VOXEL -------------------
		if (zone->region()->type() == GRegion::Type::VOXEL) {
			int ret = intersectVoxel(ray);
			if (ret ==  1) return segment.ended();	// Intersection in voxel
			if (ret == -1) continue;		// Skip this zone
			// if (ret == -2) let it continue...
		} // End of VOXEL

		// ------------------- NORMAL tracking -------------------
		if (ray->n==0) ray->prevZone(segment.zone);	// update master zone

		double T = ray->T();				// Move to next intersection

		segment.body = segment.zone->distance(this,
							 segment.pos,
							 segment.dir,
							&segment.tmin,
							 segment.tmax);
		if (ray->project_insight && applyProject(ray, T)) return ray->ended();

		if (segment.ended()) continue;
		if (segment.body==NULL) return false;

		// Check for clipping body
		if (ray->clip_insight && applyClip(ray)) {
			if (ray->project_insight && applyProject(ray, T)) return ray->ended();
			if (ray->ended()) return true;
			if (ray->n==0) ray->prevZone(segment.zone);	// update master zone
		}

		// Find the zone we are moving to
		segment.zone = whereRay(segment.pos,
					segment.dir,
					segment.tmin,
					NULL,
					segment.body);

		if (step) return (segment.zone==NULL);
	}
} // intersectRay

/** bodiesMemory */
size_t GeometryEngine::bodiesMemory() const
{
	size_t bmem = bodies.capacity() * sizeof(CBody);
	for (auto body : bodies)
		bmem += body.memory();
	return bmem;
} // bodiesMemory

/** memory */
size_t GeometryEngine::memory() const
{
	return sizeof(GeometryEngine) + bodiesMemory() + zonesSorted.memory() + irurAccel.memory();
} // memory

/** printMemory */
void GeometryEngine::printMemory() const
{
	cout << endl << "GeometryEngine:" << endl;
	cout << "Memory:"     << endl;
	cout << "\tSelf:\t"   << sizeof(GeometryEngine) << endl;
	cout << "\tBodies:\t" << bodiesMemory() << endl;
	cout << "\tZones:\t"  << zonesSorted.memory() << endl;
	cout << "\tTotal:\t"  << memory() << endl;
} // printMemory

/** Build bounding volume hierarchy */
// Not yet ready!
void GeometryEngine::buildBVH()
{
#if defined( USE_BBOX) && defined(USE_BBOX_HIERARCHIES)
	// if enabled we create a search tree of hierarchical bounding volumes
	// CT 20/8/2019
	if (boundingVolTree && boundingVolTree->GetNrOfRegions () != _geometry->nRegions())
		delete boundingVolTree;

	if (!boundingVolTree) {
		_geometry->lockRead();
		boundingVolTree = new CBoundingVolHierarchy(this, _geometry->regions,
							_geometry->nRegions(), true, 1);
		_geometry->unlockRead();
	}
#endif
} // buildBVH
