/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 *
 */

#include "geometry.h"

#include <stdarg.h>

#include <sstream>

#include "gbody.h"
#include "tool/array.h"
#include "tool/token.h"
#include "tool/timer.h"
#include "math/vector.h"
#include "math/matrix4.h"
#include "graph/painter.h"

using std::cout;
using std::cerr;
using std::endl;

bool Geometry::developer = false;

/* ============================== Geometry ================================= */

/** Geometry */
Geometry::Geometry() :
	editBody(NULL),
	editRegion("<editReg>")
{
	gridStep     =  3;
	lighterLevel =  0;
	axisLen      = 25;
	_msg[0]      =  0;

	// nlights
	deflights    =  0;
	nlights      =  0;
	_generation  =  0;

	_backgroundColor = 0x707070;
	_transparentColor= 0x707070 | FLAG_TRANSPARENT;
	errorColor       = 0xFF0000;
	gridTextColor    = 0xA0A0A0 | FLAG_FINAL;
	labelColor       = 0xA05010 | FLAG_FINAL;
	latticeColor     = 0xFFFFFF;
	messageColor     = 0xFF9000;
	paletteTextColor = 0x000000;
	regionColor      = 0x000000;
	regionErrorColor = 0xFFB0B0;
	selectColor      = 0xFF00FF;
	titleColor       = 0xD0D000;
	vertexColor      = 0xFFFF00;
	visibleColor     = 0x7F007F;
	voxelColor       = 0xFFFFFF;
	zoneColor        = 0x606060;
	makeLatticeColor();
	makeDefaultLights();

	pthread_rwlock_init(&rwlock, NULL);
	pthread_mutex_init(&mutexEdit, NULL);
} // Geometry

/** ~Geometry */
Geometry::~Geometry()
{
	// clean up local structures
	cleanup();
	voxel.cleanup();
	pthread_rwlock_destroy(&rwlock);
	pthread_mutex_destroy(&mutexEdit);
} /* ~Geometry */

/** cleanup everything. All threads must be stopped */
void Geometry::cleanup()
{
	delRegions();
	delBodies();
	delMaterials();

	lockEdit();
	editBody = NULL;
	editRegion.clear();
	unlockEdit();
	// Do not cleanup the voxel maybe we needed latter
} // cleanup

/** makeLatticeColor */
void Geometry::makeLatticeColor()
{
	latticeColor |= FLAG_LATTICE;
	voxelColor   |= FLAG_VOXEL;
} // makeLatticeColor

/** addBody */
GBody* Geometry::addBody(const std::string& name, const std::string& type)
{
	GBody* body = GBody::newBody(name, type);
	body->id(bodies.size());
	body->generation(nextGeneration());

	bodies.push_back(body);

	if (bodiesMap.find(body->name()) != bodiesMap.end())
		throw sformat("Warning: Duplicate body '%s'", name.c_str());
	bodiesMap[body->name()] = body;
	return body;
} // addBody

/** newBody */
GBody* Geometry::newBody(const std::string& name, const std::string type, ...)
{
	GBody* body = addBody(name, type);

	va_list args;
	va_start(args, type);
	double what[30];
	for (int i=0; i<body->nwhats(); i++)
		what[i] = va_arg(args, double);
	va_end(args);
	body->set(what);
	body->create();

	return body;
} // newBody

/** delBodies */
void Geometry::delBodies()
{
	for (size_t i=0; i<bodies.size(); i++)
		delete bodies[i];
	bodies.clear();
	bodiesMap.clear();
} // delBodies

/** invalidateBody */
void Geometry::invalidateBody(GBody* body)
{
	assert(body);
	body->generation(nextGeneration());
	ArrayIterator<GZone*> iter(body->zones);
	while (iter) {
		GZone* zone = iter++;
		zone->generation(nextGeneration());
		zone->region->generation(nextGeneration());
		zone->resetBBox();	// force recalculation of bboxes
		zone->bbox();
		//zone->obbox();
	}
} // invalidateBody

/** addRegion */
GRegion* Geometry::addRegion(const std::string& name)
{
	GRegion* region = new GRegion(name);
	region->id(regions.size());
	region->generation(nextGeneration());

	// XXX we could use the region->id to check if the region
	// is already defined
	regions.push_back(region);
	if (regionsMap.find(region->name()) != regionsMap.end())
		cerr << "Warning: Duplicate region " << region->name() << endl;
	regionsMap[region->name()] = region;
	return region;
} // addRegion

/** delRegions */
void Geometry::delRegions(bool fast)
{
	for (size_t i=0; i<regions.size(); i++) {
		regions[i]->clear(fast);
		delete regions[i];
	}
	regions.clear();
	regionsMap.clear();
} // delRegions

/** addZone to region */
GZone* Geometry::addZone(GRegion* region, GZone::Format aformat)
{
	GZone* zone = region->addZone(aformat);
	zone->generation(nextGeneration());
	return zone;
} // addZone

/** add2exp add token to regions expression
 * @param region	region to add token to
 * @param token		token to be added either body name or + - | ()
 * @param err		returned error string in case of problem
 * @return true on success, false otherwise
 */
bool Geometry::add2exp(GRegion* region, const char* token)
{
	assert(region);
	assert(token);
	GBody* body = getBody(token);
	return region->add2exp(token, body);
} // add2exp

/** addMaterial */
Material* Geometry::addMaterial(const std::string& name)
{
	Material* material = new Material(name);
	material->id(materials.size());
	materials.push_back(material);
	if (materialsMap.find(material->name()) != materialsMap.end())
		cerr << "Warning: Duplicate material " << material->name() << endl;
	materialsMap[material->name()] = material;
	return material;
} // addMaterial

/** delMaterials */
void Geometry::delMaterials()
{
	for (size_t i=0; i<materials.size(); i++)
		delete materials[i];
	materials.clear();
	materialsMap.clear();
	//materialsTree.destroyTree();
} // delMaterials

/** addLight to the light list
 * @param l	light structure to add
 * @return true on error, false otherwise
 */
bool Geometry::addLight(const Light& l)
{
	if (nlights >= MAXLIGHT) return true;
	light[nlights++] = l;
	return false;
} // addLight

/** make default lights. The default lights are created at initialization
 * and they have indices from 0..deflights-1 and will remain unchanged
 * The user lights will be indexed as deflights .. lights
 */
void Geometry::makeDefaultLights()
{
	// http://www.secondpicture.com/tutorials/3d/three-point_lighting_in_3ds_max_01.html
	nlights = 0;
	light[nlights].type     = Light::Type::Sun;		// sun light
	light[nlights].relative = true;
	light[nlights].dir      = Vector( 5.,-8.,-10.);
	light[nlights].distance = INFINITE;
	light[nlights].power    = 0.8;
	light[nlights].spec     = true;
	light[nlights].falloff  = Light::FallOff::Constant;
	light[nlights].shadow   = true;
	light[nlights].nsamples = 1;
	light[nlights].radius   = 0;
	nlights++;

	light[nlights].type     = Light::Type::Sun;		// fill light
	light[nlights].relative = true;
	light[nlights].dir      = Vector(-1.,1.,-1.);
	light[nlights].distance = INFINITE;
	light[nlights].power    = 0.25;
	light[nlights].spec     = false;
	light[nlights].falloff  = Light::FallOff::Constant;
	light[nlights].shadow   = false;
	light[nlights].nsamples = 1;
	light[nlights].radius   = 0;
	nlights++;

	light[nlights].type     = Light::Type::Sun;		// back light
	light[nlights].relative = true;
	light[nlights].dir      = Vector(-4.,-2.,6.);
	light[nlights].distance = INFINITE;
	light[nlights].power    = 0.8;
	light[nlights].spec     = false;
	light[nlights].falloff  = Light::FallOff::Constant;
	light[nlights].shadow   = false;
	light[nlights].nsamples = 1;
	light[nlights].radius   = 0;
	nlights++;

	for (int i=0; i<nlights; i++)
		light[i].dir.normalize();

	deflights = nlights;
} // makeDefaultLights

/** parse zone from an expression string
 * FIXME: merge with the GeometryObject._region_AddExpr
 */
void Geometry::parseZone(GZone& zone, const std::string& expstr)
{
	zone.clear();
	zone.format(GZone::Format::NORMAL);

	std::stringstream stream(expstr);
	Token token(stream);
	do {
		token.next();
		DUMP(if (token() != Token::Type::ENDOFILE)
			cout	<< "Token " << (int)token()
				<< " '" << token.string() << "'" << endl);
		switch (token()) {
			case Token::Type::ENDOFILE:
				return;

			case Token::Type::IDENT: {
				GBody* body = getBody(token.ident());
				if (body==NULL)
					throw sformat("Body '%s' not found", token.ident());
				zone << body;
				}
				break;

			case Token::Type::PLUS:
				zone << &GBody::tplus;
				break;

			case Token::Type::MINUS:
				zone << &GBody::tminus;
				break;

			case Token::Type::OR:
				zone << &GBody::tunion;
				break;

			case Token::Type::LEFT_PAREN:
				zone << &GBody::tleft;
				break;

			case Token::Type::RIGHT_PAREN:
				zone << &GBody::tright;
				break;

			default: {
				// FIXME better throw an exception
				throw sformat("Invalid token '%s' found", token.string());
			}
		}
	} while (token);
} // parseZone

/** derive */
void Geometry::derive()
{
	lockWrite();
	for (auto region : regions)
		// setup region (bbox, calcLength,...)
		region->setup();
	unlockWrite();
} // derive

/** printInfo */
void Geometry::printInfo() const
{
	cout << "Geometry Info" << endl;
	cout << "Number of bodies: " << bodies.size() << endl;
	cout << "Number of regions: " << regions.size() << endl;
	int nzones = 0;
	for (auto region : regions)
		nzones += region->size();
	cout << "Number of zones: " << nzones << endl;
} // printInfo

/** bodiesMemory */
size_t Geometry::bodiesMemory() const
{
	size_t bmem = bodies.capacity()*sizeof(GBody*) + sizeof(bodiesMap);
	for (auto b : bodies)
		bmem += b->memory();
	return bmem;
} // bodiesMemory

/** regionsMemory */
size_t Geometry::regionsMemory() const
{
	size_t rmem = regions.capacity()*sizeof(GRegion*) + sizeof(regionsMap);
	for (auto r : regions)
		rmem += r->memory();
	return rmem;
} // regionsMemory

/** memory */
size_t Geometry::memory() const
{
	return   sizeof(Geometry)
	       + bodiesMemory()
	       + regionsMemory();
//	       + voxel.memory();
} // memory

/** printMemory */
void Geometry::printMemory() const
{
	cout << endl << "Geometry:" << endl;
	cout << "Bodies:\t\t" << bodies.size()   <<  endl;
	cout << "Regions:\t"  << regions.size()  <<  endl;
	cout << "Memory:"     << endl;
	cout << "\tSelf:\t"   << sizeof(Geometry) << endl;
	cout << "\tBodies:\t" << bodiesMemory()   << endl;
	cout << "\tRegion:\t" << regionsMemory()  << endl;
	cout << "\tTotal:\t"  << memory() << endl;
} // printMemory
