/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 */

#ifndef __GEOMETRY_RAY_H
#define __GEOMETRY_RAY_H

#include "geo.h"
#include "vzone.h"

#include "math/point.h"
#include "math/vector.h"

class Matrix4;

/* ============================= RaySegment =========================== */
class RaySegment {
public:
	Point	pos;			// current local system position
	Vector	dir;			// current local system direction
	VZone*	zone;			// Current location
	CBody*	body;			// entering body
const	GRegion	*region;		// lattice region
	double	tmin;			// range of segment
	double	tmax;			// [tmin..tmax)
	double	acc;			// range accuracy
	int	bodyCheckId;		// bodyCheckId, used for lattices

public:
	RaySegment(GRegion* reg=NULL) :
			pos(0.,0.,0.),
			dir(0.,0.,0.),
			zone(NULL),
			body(NULL),
			region(reg),
			tmin(SMALL3D),
			tmax(INFINITE),
			acc(SMALL3D),
			bodyCheckId(0) {}
	RaySegment(const double x, const double y, const double z,
		   const double dx, const double dy, const double dz,
		   VZone* zz) :
			pos(x,y,z),
			dir(dx,dy,dz),
			zone(zz),
			body(NULL),
			region(NULL),
			tmin(SMALL3D),
			tmax(INFINITE),
			acc(SMALL3D),
			bodyCheckId(0) {}
	RaySegment(const Point& p, const Vector& d, VZone* zz) :
			pos(p),
			dir(d),
			zone(zz),
			body(NULL),
			region(NULL),
			tmin(SMALL3D),
			tmax(INFINITE),
			acc(SMALL3D),
			bodyCheckId(0) {}

	bool	ended()	const	{ return tmin*(1.0+acc) >= tmax; }

	// Relative hit position inside the segment-frame
	// LATTICE: are screen coordinates transformed by lattice rot-translation
	// VOXEL: are absolute coordinates
	Point	hit()	const	{ return pos + tmin*dir; }

	Point	hit(const double boost) const
				{ return pos + tmin*(1.0+boost)*dir; }

const	Matrix4& matrix()	{ assert(region); return region->matrix(); }
const	Matrix4& invMatrix()	{ assert(region); return region->invMatrix(); }
}; // RaySegment

/* ================================= Ray ============================== */
/** Intersect Ray structure
 * This structure contains a list of RaySegments,
 *   the last segment of the list, the one pointed by n (segments[n]) is
 *   considered the working one and the rest in the list are the previous
 *   segments.
 */
class Ray {
public:
	bool	error;			// If ray passed through an undefined region
	bool	shadow;			// calculate shadows for ray
	int	nlights;		// number of lights to use
	bool	skip_current;		// Skip current zone as if it was transparent
	bool	skip_transparent;	// Skip transparent zones
	bool	skip_1stblack;		// Skip the starting blackhole if any
	int	voxelreg;		// voxel region if any

	Vector	normal;			// keep normal at hit position

	/* Clipping body */
	bool	use_clip;		// flag if we have to use clipping
	bool	clip_insight;		// possible clip along the ray
	bool	clip_hit;		// clipping body was hit

	/* Project body */
	bool	use_project;		// flag if we have to use projection
	bool	project_insight;	// possible projection along the ray
	bool	project_hit;		// project was hit
	int	project_alpha;		// projection transparency

	int	depth;			// segment depth
	int	max_depth;		// maximum segment depth

private:
	VZone	*_prevzone;		// current upstream zone (NOT the hit zone)
	double	tsum;			// tmin running length of previous levels
public:
	int	n;			// lattice depth level
private:
	RaySegment segments[MAXLEVEL];

public:
	Ray()	: shadow(false),
		  nlights(0),
		  skip_1stblack(false),
		  use_clip(false),
		  use_project(false),
		  depth(0),
		  max_depth(0)
			{init();}

	Ray(const Ray& ray) = default;
	Ray& operator = (const Ray&) = default;

	// clone
	// NOTE: Normally I'd implement this in terms of assignment op.
	//       However, I saw that the copy ctor is not really creating a deep copy
	//       but increases the ray's depth and does a fresh init().
	//       To avoid side effects I keep the compiler's default assignment
	//       op and rather create a specific clone function.
	//       CT 16/8/2019
	//
	void clone(Ray& destRay) {
			if (&destRay != this) {
				destRay.shadow          = shadow;
				destRay.nlights         = nlights;
				destRay.use_clip        = use_clip;
				destRay.use_project     = use_project;
				destRay.skip_1stblack   = skip_1stblack;
				destRay.depth           = depth;
				destRay.max_depth       = max_depth;

				destRay.voxelreg        = voxelreg;
				destRay.error           = error;
				destRay.skip_current    = skip_current;
				destRay.skip_transparent= skip_transparent;

				destRay.clip_insight    = clip_insight;
				destRay.clip_hit        = clip_hit;

				destRay.project_insight = project_insight;
				destRay.project_hit     = project_hit;
				destRay.project_alpha   = project_alpha;

				destRay._prevzone       = _prevzone;
				destRay.tsum            = tsum;
				destRay.n               = n;

				std::copy(std::begin(segments),
					  std::end(segments),
					  std::begin(destRay.segments));
			}
		}

	void	init() {
			voxelreg         = -1;
			error            = false;
			skip_current     = false;
			skip_transparent = true;
			clip_insight     = false;
			clip_hit         = false;
			project_insight  = false;
			project_hit      = false;
			project_alpha    = 0;
			_prevzone        = NULL;
			tsum             = 0.0;
			n                = -1;
		}

	// Return true if ray segments buffer is full
	bool	isFull()	const	{ return n+1 >= MAXLEVEL; }
	bool	isEmpty()	const	{ return n == -1; }
	bool	start()		const	{ assert(n>-1); return T() <= SMALL3D3; }
	void	incDepth()		{ depth++; }

	// prefix operator to increase the depth counter
	Ray& operator++()		{ ++depth;  return *this; };
	// postfix op implemented in terms of the prefix op
	Ray  operator++(int)		{ Ray temp = *this; ++*this; return temp; };

	// Consolidate the current segment and make current a new one
	bool	push(const RaySegment& _segment) {
			if (isFull()) {
#if _DEBUG>1
//				std::cerr << "Error pushing raySegment, ray is full" << std::endl;
#endif
				error = true;
				return true;
			}
			if (n >= 0) {
				tsum += segments[n].tmin;
				if (n == 0)
					_prevzone = segments[n].zone;
			}
			n++;
			segments[n] = _segment;
			return false;
		}

	// Discards the current segment and moves as current the previous one
	bool	pop() {
			if (isEmpty()) return true;
			n--;
			if (n >= 0) {
				tsum -= segments[n].tmin;
			}
			return false;
		}

	// Return reference to any segment
	RaySegment&	segment(int idx)	{ assert(idx<=n); return segments[idx]; }
const	RaySegment&	segment(int idx) const	{ assert(idx<=n); return segments[idx]; }
	// Current segment
	RaySegment&	segment()		{ return segments[n]; }
const	RaySegment&	segment()	const	{ return segments[n]; }

const	Point&	pos()		const { return segments[0].pos; }
const	Vector&	dir()		const { return segments[0].dir; }

	// Absolute screen coordinates hit position
	Point	hit()		const { return pos() + (segments[n].tmin+tsum)*dir(); }
	Point	hit(const double boost) const {
			assert(boost!=0.0);
			if (boost>0.0)	// Forward boost
				return pos() + ((segments[n].tmin+tsum)*(1.0+boost))*dir();
			else		// Backward boost
				return pos() + ((segments[n].tmin+tsum)/(1.0-boost))*dir();
		}

	// set/get
	void	prevZone(VZone *z)    { _prevzone = z; }
	VZone*	prevZone()	const { return _prevzone; }
	CBody*	hitBody()	const { for (int i=n; i>=0 ; i--)
						if (segments[i].body)
							return segments[i].body;
					return nullptr;
				}
	VZone*	hitZone()	const { return segment().zone; }
	VRegion* hitRegion()	const { return segment().zone->region(); }

	void	moveby(double a)      { segment().tmin = T()*(1.0+a) - tsum; }

	double	T()		const { return segments[n].tmin + tsum; }
	double	Tsum()		const { return tsum; }
	bool	ended()		const { return T()>=segments[0].tmax; }
}; // Ray

#endif
