/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	16-Jul-2010
 */

#include "bfont.h"

#include <assert.h>
#include <string.h>

#include "painter.h"

/* ============================== BFont ================================== */
/** clean */
void BFont::clean()
{
	_name.clear();
	if (imageData) {
		delete [] imageData;
		imageData = NULL;
	}
} // clean

/** load a font image
 * Loads A TGA File Into Memory
 * Expects to be compression:none, TopLeft format (byte 17=0x20)
 * @return true on success
 */
bool BFont::load(const char* filename)
{
#if 0
	// Uncompressed TGA Header
	typedef struct {
	0	byte  identSize;	// size of ID field that follows 18 byte header (0 usually)
	1	byte  colorMap;		// type of color map 0=none, 1=has palette
	2	byte  imageType;	// type of image
					// 0=none,1=indexed,2=rgb,3=grey,+8=rle packed

	3	short colorMapStart;	// first color map entry in palette
	5	short colorMapLength;	// number of colors in palette
	7	byte  colorMapBits;	// number of bits per palette entry 15,16,24,32

	// Alignment issues on 64bit for the xstart is in odd position!!!
	8	short xstart;		// image x origin
	10	short ystart;		// image y origin
	12	short width;		// image width in pixels
	14	short height;		// image height in pixels
	16	byte  bits;		// image bits per pixel 8,16,24,32
	17	byte  descriptor;	// image descriptor bits (vh flip bits)
	} TGA_Header;
	TGA_Header tga;
#endif
	byte header[18];

	clean();
	assert(sizeof(short)==2);
	FILE *file = fopen(filename, "rb");
	if (file==NULL) return false;

	if (fread(&header, 1, sizeof(header), file) != sizeof(header)) {
		fclose(file);
		return false;
	}

	// It has to be grey, 8 bit
	if (header[1]  != 0 ||	// colourMap
	    header[2]  != 3 ||	// imageType
	    header[3]  != 0 ||	// colorMap start
	    header[4]  != 0 ||	//  -//-
	    header[5]  != 0 ||	// map length
	    header[6]  != 0 ||	//  -//-
	    header[7]  != 0 ||	// map bits
	    header[16] != 8) {	// xstart (should be 0)
		fclose(file);
		return false;
	}

	if (header[0]) fseek(file, header[0], SEEK_CUR);	// skip title

	imageWidth  = (int)header[12] + ((int)header[13]<<8);
	imageHeight = (int)header[14] + ((int)header[15]<<8);

	_width  = imageWidth  / 16;
	_height = imageHeight / 16;

	// Calculate The Memory Required For The TGA Data
	size_t imageSize = imageWidth * imageHeight * header[16]/8;

	// Reserve Memory To Hold The TGA Data
	imageData= new byte[imageSize];

	if (imageData==NULL || fread(imageData, 1, imageSize, file)!=imageSize) {
		if (imageData!=NULL)
			delete [] imageData;
		imageData = NULL;
		fclose(file);
		return false;
	}
	fclose(file);

	// Copy name
	_name = filename;

	build();

	return true;
} // load

/** set */
void BFont::set(const char* n, int w, int h, byte* data)
{
	clean();
	_name   = n;
	_width  = w;
	_height = h;
	imageData = new byte[_width*_height];
	memcpy(imageData, data, _width*_height);
	build();
} // set

/** set */
void BFont::set(const char* n, int w, int h, dword* data)
{
	clean();
	_name   = n;
	_width  = w;
	_height = h;

	imageData = new byte[_width*_height];

	byte *bptr = imageData;
	dword *dptr = data;

	FILE *f = fopen("font.gray","wb");
	for (int i=0; i<_width*_height; i++, bptr++, dptr++) {
		*bptr = *dptr? 0xFF : 0x00;
		fputc(*bptr,f);
	}
	fclose(f);
	build();
} // set

/** build font */
void BFont::build()
{
	if (imageData==NULL) return;

	// Convert to black/white
	byte *ptr = imageData;
	for (size_t i=0; i<(size_t)(_width*_height); i++)
		if (*ptr < 0x7F)
			*ptr = 0;
		else
			*ptr = 0xFF;

	// scan fonts to find the length of each character
	for (int ch=0; ch<256; ch++) {
		ptr = charPtr(ch);
		int maxlen = 0;
		for (int j=0; j<_height; j++) {
			for (int i=_width-2; i>=2; i--)
				if (ptr[i]) {
					maxlen = Max(maxlen, i);
					break;
			}
			ptr += imageWidth;
		}
		charLength[ch] = (byte)(maxlen+2);
		// space
		if (ch==' ') charLength[ch] = (byte)(_width/2);
	}
} // build

/** @return string length in pixels */
int BFont::measure(const char *str)
{
	int size = 0;
	for (const byte *ch=(const byte*)str; *ch; ch++)
		size += charLength[*ch];

	return size;
} // measure

/** draw a character into a viewport
 * @param painter	painter to use
 * @param x,y	location where to draw character
 * @param color	color to use
 * @param ch	character to draw
 */
int BFont::draw(Painter& painter, int x, int y, const dword color, const byte ch)
{
	byte  *src;
	dword *dst;

	if (imageData==NULL) return 0;

	int w = charLength[ch];

	src = charPtr(ch);
	dst = painter.pixelPtr(x,y);
	for (int j=0; j<_height; j++) {
		int xp = x;
		for (int i=0; i<w; i++) {
			if (painter.insideClip(xp,y) && *src != 0) *dst = color;
			xp++;
			src++;
			dst++;
		}
		y++;
		src += imageWidth - w;
		dst += painter.width() - w;
	}
	return w;
} // draw

/** draw a vertical character into a viewport
 * @param painter	painter to use
 * @param x,y	location where to draw character
 * @param color	color to use
 * @param ch	character to draw
 */
int BFont::vdraw(Painter& painter, int x, int y, const dword color, const byte ch)
{
	byte  *src;
	dword *dst;

	if (imageData==NULL) return 0;

	int w = charLength[ch];

	src = charPtr(ch);
	dst = painter.pixelPtr(x,y);
	for (int j=0; j<_height; j++) {
		int yp = y;
		for (int i=0; i<w; i++) {
			if (painter.insideClip(x,yp) && *src != 0) *dst = color;
			yp--;
			src++;
			dst -= painter.width();
		}
		x++;
		src += imageWidth - w;
		dst = painter.pixelPtr(x,y);
	}
	return w;
} // vdraw

/** draw a character outline
 * @param painter	painter to use
 * @param x,y	location where to draw character
 * @param color	color to use
 * @param outline	color to use
 * @param ch	character to draw
 */
int BFont::drawOutline(Painter& painter, int x, int y,
			const dword color, const dword outline, const byte ch)
{
	byte  *src;
	dword *dst;

	if (imageData==NULL) return 0;

	int w = charLength[ch];

	src = charPtr(ch);
	dst = painter.pixelPtr(x,y);
	int yp = y;
	for (int j=0; j<_height; j++) {
		int xp = x;
		for (int i=0; i<w; i++) {
			if (painter.insideClip(xp,yp)) {
				if (*src)
					*dst = color;
				else
				if (src[1] || src[-1])
					*dst = outline;
			}
			xp++;
			src++;
			dst++;
		}
		yp++;
		src += imageWidth - w;
		dst += painter.width() - w;
	}
	return w;
} // drawOutline
