/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	18-Jun-2018
 */

#ifndef __MATH_BIT_H
#define __MATH_BIT_H

#include <stdint.h>
#include "os.h"

const int KB = 1024;
const int MB = KB*KB;
const int GB = KB*MB;

/* --- BIT definitions --- */
#define MASK1      0x0001
#define MASK2      0x0003
#define MASK3      0x0007
#define MASK4      0x000f
#define MASK5      0x001f
#define MASK6      0x003f
#define MASK7      0x007f
#define MASK8      0x00ff
#define MASK9      0x01ff
#define MASK10     0x03ff
#define MASK11     0x07ff
#define MASK12     0x0fff
#define MASK13     0x1fff
#define MASK14     0x3fff
#define MASK15     0x7fff
#define MASK16     0xffff
#define MASK17     0x1ffffL
#define MASK18     0x3ffffL
#define MASK19     0x7ffffL
#define MASK20     0xfffffL
#define MASK21     0x1fffffL
#define MASK22     0x3fffffL
#define MASK23     0x7fffffL
#define MASK24     0xffffffL
#define MASK25     0x1ffffffL
#define MASK26     0x3ffffffL
#define MASK27     0x7ffffffL
#define MASK28     0xfffffffL
#define MASK29     0x1fffffffL
#define MASK30     0x3fffffffL
#define MASK31     0x7fffffffL
#define MASK32     0xffffffffL

#define BIT0       0x0001
#define BIT1       0x0002
#define BIT2       0x0004
#define BIT3       0x0008
#define BIT4       0x0010
#define BIT5       0x0020
#define BIT6       0x0040
#define BIT7       0x0080
#define BIT8       0x0100
#define BIT9       0x0200
#define BIT10      0x0400
#define BIT11      0x0800
#define BIT12      0x1000
#define BIT13      0x2000
#define BIT14      0x4000
#define BIT15      0x8000
#define BIT16      0x10000L
#define BIT17      0x20000L
#define BIT18      0x40000L
#define BIT19      0x80000L
#define BIT20      0x100000L
#define BIT21      0x200000L
#define BIT22      0x400000L
#define BIT23      0x800000L
#define BIT24      0x1000000L
#define BIT25      0x2000000L
#define BIT26      0x4000000L
#define BIT27      0x8000000L
#define BIT28      0x10000000L
#define BIT29      0x20000000L
#define BIT30      0x40000000L
#define BIT31      0x80000000L

/* --- bytes, words --- */
#ifndef LOBYTE
	#define LOBYTE(x)	((byte)(x))
	#define HIBYTE(x)	((byte)(((word)(x)>>8) & MASK8))
	#define LOWORD(x)	((word)(x))
	#define HIWORD(x)	((word)(((dword)(x)>>16) & MASK16))
	#define MAKEWORD(l,h)	((word)(((byte)(l)) | ((word)((byte)(h))) << 8))
	#define MAKEDWORD(l,h)	((dword)(((word)(l)) | ((dword)((word)(h))) << 16))
#endif

#define BOOL(a)		((a)?1:0)
#define BIT(n)		(1L << (n))
#define GETBIT(a,n)	BOOL((a) & BIT(n))
#define SETBIT(n,i)	((n) |= BIT(i))
#define CLRBIT(n,i)	((n) &= ~BIT(i))
#define TESTBIT(n,i)	(((n) & BIT(i)) != 0)

#define SHIFT(a,x)	((x>=0) ? ((a) << (x)) : ((a) >> (x)))
#define SUBBIT(a,p,x)	((dword)(((dword) a) << (32-p-x)) >> (32-x))

/** byte multiplication assuming both bytes represent a range of 0.0 .. 1.0 decimal
 * @param a, b
 * @return a * b / 256
 */
inline byte byteMul(byte a, byte b)
{
	  return (((int)a + 1) * (int)b) >> 8;
} // byteMul

/** Check whether the passed integer is a power of 2 */
inline bool isPowerOf2(int v)
{
	return (v & (v - 1)) == 0;
} // isPowerOf2

/** Round up to the next power of 2 */
inline uint32_t roundUpPow2(uint32_t v)
{
	v--;
	v |= v >> 1;
	v |= v >> 2;
	v |= v >> 4;
	v |= v >> 8;
	v |= v >> 16;
	return v+1;
} // roundUpPow2

/** @return integer log base2 of 32bit integers from:
 * https://stackoverflow.com/questions/11376288/fast-computing-of-log2-for-64-bit-integers
 */
inline int log2_32 (uint32_t value)
{
	const static int tab32[32] = {
		     0,  9,  1, 10, 13, 21,  2, 29,
		    11, 14, 16, 18, 22, 25,  3, 30,
		     8, 12, 20, 28, 15, 17, 24,  7,
		    19, 27, 23,  6, 26,  5,  4, 31};
	value |= value >> 1;
	value |= value >> 2;
	value |= value >> 4;
	value |= value >> 8;
	value |= value >> 16;
	return tab32[(uint32_t)(value*0x07C4ACDD) >> 27];
} // log2_32

/** @return integer log base2 of 64bit integers from the above source */
inline int log2_64 (uint64_t value)
{
	const static int tab64[64] = {
		    63,  0, 58,  1, 59, 47, 53,  2,
		    60, 39, 48, 27, 54, 33, 42,  3,
		    61, 51, 37, 40, 49, 18, 28, 20,
		    55, 30, 34, 11, 43, 14, 22,  4,
		    62, 57, 46, 52, 38, 26, 32, 41,
		    50, 36, 17, 19, 29, 10, 13, 21,
		    56, 45, 25, 31, 35, 16,  9, 12,
		    44, 24, 15,  8, 23,  7,  6,  5};
	value |= value >> 1;
	value |= value >> 2;
	value |= value >> 4;
	value |= value >> 8;
	value |= value >> 16;
	value |= value >> 32;
	return tab64[((uint64_t)((value - (value >> 1))*0x07EDD5E59A4E28C2)) >> 58];
} // log2_64

/** @return log base 10 of an integer value
 * @param value // non-zero 32-bit integer value to compute the log base 10 of
 */
inline int log10_32(uint32_t value)
{
	static unsigned const PowersOf10[] =
		{1, 10, 100, 1000, 10000, 100000,
		1000000, 10000000, 100000000, 1000000000};

	int t = (log2_32(value) + 1) * 1233 >> 12; // (use a log2 method from above)
	return t - (value < PowersOf10[t]);
} // log10_32

/** helper functions for branch elimination in conditions
*
* msb/lsb - most/less significant byte
*
* mask - 0xFFFFFFFF
* nz   - not zero
* zr   - is zero
*/
inline unsigned nz2msb(const unsigned x)
{
	return -(int)x | x;
}

inline unsigned msb2mask(const unsigned x)
{
	return (int)(x) >> 31;
}

inline unsigned nz2one(const unsigned x)
{
	return nz2msb(x) >> 31; // int((bool)x);
}

inline int nz2mask(const unsigned x)
{
	return (int)msb2mask(nz2msb(x)); // -(int32)(bool)x;
}

// select integer with mask (0xFFFFFFFF or 0x0 only!!!)
inline unsigned iselmask(const unsigned mask, unsigned x, const unsigned y)
{
	return (x & mask) | (y & ~mask);
}

// mask if(x != 0 && y != 0)
inline unsigned mask_nz_nz(const unsigned x, const unsigned y)
{
	return msb2mask(nz2msb(x) & nz2msb(y));
}

// mask if(x != 0 && y == 0)
inline unsigned mask_nz_zr(const unsigned x, const unsigned y)
{
	return msb2mask(nz2msb(x) & ~nz2msb(y));
}

// mask if(x == 0 && y == 0)
inline unsigned mask_zr_zr(const unsigned x, const unsigned y)
{
	return ~nz2mask(x | y);
}

#if 0
/** Fast inverse square root (https://en.wikipedia.org/wiki/Fast_inverse_square_root)
 * @param number	number to return 1/sqrt(number)
 * @return a reasonable approximation of the inverse square root
 */
inline float Q_rsqrt(float number)
{
	int i;
	float x2, y;
	const float threehalfs = 1.5F;

	x2 = number * 0.5F;
	y  = number;
	i  = *(int *)&y;			// evil floating point bit level hacking
	i  = 0x5f3759df - (i >> 1);		// what the fuck?
	y  = *(float*)(&i);
	y  = y * (threehalfs - x2*y*y);		// 1st iteration
	y  = y * (threehalfs - x2*y*y);		// 2nd iteration, this can be removed
	return y;
} // Q_rsqrt
#endif

#endif
