/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
// Authors:	Vasilis.Vlachoudis@cern.ch
//		Christian.Theis@cern.ch
 * Date:	17-Jul-2020
 */

#include "morton.h"

#include <assert.h>

#if _DEBUG>0
#	include <iostream>

#	include "os.h"

using std::cout;
using std::endl;
#endif


/** return next pair (x,y) in the Z-Morton sequence clipped by width & height
 * @param x,y		pair I/O
 * @param width		x<width
 * @param heght	y<height
 * @return false on end, true otherwise
 */
bool Morton::next2D(int& x, int& y, const int width, const int height)
{
	int bitx = 1;
	int bity = 1;

	while (true) {
		if ((x&1) == 0) {
			if (x+1 >= width) {
				if (y+1 >= height)
					return false;
				else
				if ((y & (2*bity-1)) == 2*bity-1) {
					x |= 2*bitx-1;
					bitx <<= 1;
					continue;
				} else {
					y++;
					return true;
				}
			} else {
				x++;
				return true;
			}
		}

		if ((y&1) == 0) {
			if (y+1>=height) {
				if (x+1>=width)
					return false;
				else
				if ((x & (4*bitx-1)) == 4*bitx-1) {
					y |= 4*bity-1;
					bity <<= 1;
					continue;
				} else {
					x++;
					return true;
				}
			} else {
				x--;
				y++;
				return true;
			}
		}

		if (bity >= bitx) {
			bitx <<= 1;
			if ((x&bitx) == 0) {
				if (x+1 >= width) {
					if ((y+1) >= height)
						return false;
					else
					if ((y & (2*bity-1)) == 2*bity-1)
						x |= 2*bitx-1;
					else
						assert(false);
				} else {
					x++;
					y -= 2*bity-1;
					return true;
				}
			} else
			if ((y&bity) == 0)
				assert(false);
		} else {
			bity <<= 1;
			if ((x&bitx) == 0) {
				if (x+1>=width) {
					x -= bitx-1;
					if (y+1 > height)
						assert(false);
				} else
					assert(false);
			} else
			if ((y&bity) == 0) {
				if (y+1 >= height) {
					if (x+1 >= width)
						return false;
					else
					if ((x & (2*bitx-1)) == 2*bitx-1)
						y |= 2*bity-1;
					else
						assert(false);
				} else {
					x -= 2*bitx-1;
					y++;
					return true;
				}
			}
		}
	}
} // next2D

#if _DEBUG>0
/** test */
void Morton::test(int W, int H, bool plot)
{
	if (plot) {
		int N = Max(W,H);
		cout << endl;
		cout << "set title 'Morton " << W << " x " << H << "'" << endl;
		cout << "set xrange [-0.5:" << N << "-0.5]" << endl;
		cout << "set yrange [-0.5:" << N << "-0.5]" << endl;
		cout << "plot \"-\" u 1:2 w lp pt 7" << endl;
	}
	int n=0;
	int x=0;
	int y=0;
	int sum=0;
	do {
		if (plot)
			std::cout << x << " " << y << endl;
		sum += x + y + 2;	// for testing
		n++;			// for testing
	} while (Morton::next2D(x,y,W,H));
	assert(n==H*W);
	int check = W*(H*(H+1)/2) + H*(W*(W+1)/2);
	assert(sum == check);
	if (plot) {
		cout << "e" << endl;
		cout << "pause -1" << endl;
	}
} // test
#endif
