/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	17-Jan-2019
 */

#ifndef __MATH_PLANE_H
#define __MATH_PLANE_H

#include "os.h"
#include "eps.h"
#include "surface.h"

// =================================== Plane ===================================
/** Plane surface using the formula ax + by + cz + d = 0
 */
class Plane : public Surface {
public:
	double	a, b, c, d;	/** plane coefficients */

private:
	double	norm;		/** normalization coefficient sqrt(a*a + b*b + c*c) */
	int	_processed;	/** processed flag			*/

public:
	Plane() : a(0), b(0), c(0), d(0), norm(1) {}
	Plane(const double aa, const double ab, const double ac, const double ad)
		: a(aa), b(ab), c(ac), d(ad)
		  { init(); }
	Plane(const Point& A, const Point& B, const Point& C) { make(A,B,C); }
virtual ~Plane() {}

	/** @return surface degree */
virtual	int	degree() const				{ return 1; }
virtual	void	init();

	double	operator ()(const Point& p) const	{ return a*p.x + b*p.y + c*p.z + d; }

	void	set(const double aa, const double bb, const double cc, const double dd)
			{ a = aa; b = bb; c = cc; d = dd; init(); }
	void	make(const Point& A, const Point& B, const Point& C);

	/** evaluate gradient at point p */
virtual	Vector	grad(const Point& /*p*/) const		{ return Vector(a,b,c) * norm; }

	/** @return closest distance */
virtual	double	distance(const Point& p) const		{ return (*this)(p) * norm; }
virtual	int	distance(const Point& p, const Vector v, double t[]) const;

//virtual	void	translate(const Vector& v)	{ d += a*v.x + b*v.y + c*v.z; }

virtual	void	normalize() {
			init();
			a *= norm;
			b *= norm;
			c *= norm;
			d *= norm;
			norm = 1.0;
		}

	/** set/get processed flag */
	int	processed()		 const	{ return _processed; }
	void	processed(int p)		{ _processed = p; }
}; // class Plane

// ================================== PlaneX ===================================
class PlaneX : public Plane {
public:
	PlaneX(double x=0) : Plane(1, 0, 0, -x) {}

	/** evaluate gradient at point p */
virtual	Vector	grad(const Point& /*p*/) const		{ return Vector::Xo; }

	/** @return closest distance */
virtual	double	distance(const Point& p) const		{ return p.x+d; }
virtual	int	distance(const Point& p, const Vector v, double t[]) const;

//virtual	void	translate(const Vector& v)		{ d += v.x; }
}; // class PlaneX

// ================================== PlaneY ===================================
class PlaneY : public Plane {
public:
	PlaneY(double y=0) : Plane(0, 1, 0, -y) {}

	/** evaluate gradient at point p */
virtual	Vector	grad(const Point& /*p*/) const		{ return Vector::Yo; }

	/** @return closest distance */
virtual	double	distance(const Point& p) const		{ return p.y+d; }
virtual	int	distance(const Point& p, const Vector v, double t[]) const;

//virtual	void	translate(const Vector& v)		{ d += v.y; }
}; // class PlaneY

// ================================== PlaneZ ===================================
class PlaneZ : public Plane {
public:
	PlaneZ(double z=0) : Plane(0, 0, 1, -z) {}

	/** evaluate gradient at point p */
virtual	Vector	grad(const Point& /*p*/) const		{ return Vector::Zo; }

	/** @return closest distance */
virtual	double	distance(const Point& p) const		{ return p.z+d; }
virtual	int	distance(const Point& p, const Vector v, double t[]) const;

//virtual	void	translate(const Vector& v)		{ d += v.z; }
}; // class PlaneZ

#endif
