/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#ifndef __MATH_POINT_H
#define __MATH_POINT_H

#include "vector.h"

class Matrix4;

/* =================================== Point ================================ *
 * 3D Point: Overloaded class of Vector defining a point in space
 *           in contrast of a vector that typically is used to
 *           define a direction and magnitude
 *
 * WARNING: all transformations for Point are different from a Vector
 *          in the vector they include only the rotation
 *          while for a point includes the translation as well
 */
class Point : public Vector {
public:
	// Constructor
	Point(const double ax=0.0, const double ay=0.0, const double az=0.0)
		: Vector(ax, ay, az) {}

	// Copy constructor
	Point(const Point&) = default;
	Point(const Vector& p) : Vector(p) {}

	// Assignment
	Point&	operator =(const Point&) = default;

	// Assignment of Vector and classes derived from it (Vector, Normal)
	Point&	operator =(const Vector &p) { x=p.x; y=p.y; z=p.z; return *this; }

	// Distance squared to the origin
	double	distance2() const { return length2(); }

	// Distance to the origin
	double	distance() const { return sqrt(distance2()); }

	// Distance squared to the point
	double	distance2(const Point &p) const {
			double delx = p.x-x;
			double dely = p.y-y;
			double delz = p.z-z;
			return delx*delx + dely*dely + delz*delz;
		}

	// Distance to the point
	double	distance(const Point &p) const { return sqrt(distance2(p)); }

	// Transformation
//	Point& transform(const Matrix4 &mat) { return *this = mat * (*this); }
}; // class Point

//#ifndef __MATH_MATRIX4_H
//#	include "math/matrix4.h"
//#endif
//inline Point& Point::transform(const Matrix4 &mat) {
//	return *this = mat * (*this);
//}

// Overloaded operations on Points

// point + point = point	(A bit brain twisted it should not exist)
inline Point operator + (const Point& a, const Point& b) {
	return Point(a.x+b.x, a.y+b.y, a.z+b.z);
}

// point + vector = point
inline Point operator + (const Point& a, const Vector& b) {
	return Point(a.x+b.x, a.y+b.y, a.z+b.z);
}

// vector + point = point
inline Point operator + (const Vector& a, const Point& b) {
	return Point(a.x+b.x, a.y+b.y, a.z+b.z);
}

// point - point = vector
inline Vector operator - (const Point& a, const Point& b) {
	return Vector(a.x-b.x, a.y-b.y, a.z-b.z);
}

// point - vector = point
inline Point operator - (const Point& a, const Vector& b) {
	return Point(a.x-b.x, a.y-b.y, a.z-b.z);
}

// vector - point = vector	(A bit brain twisted it should not exist)
inline Vector operator - (const Vector& a, const Point& b) {
	return Vector(a.x-b.x, a.y-b.y, a.z-b.z);
}

// f * point = point
inline Point operator * (const Point& p, double a) {
	return Point(a*p.x, a*p.y, a*p.z);
}

// point * f = point
inline Point operator * (double a, const Point& p) {
	return Point(a*p.x, a*p.y, a*p.z);
}

#endif
