/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#include "quad.h"

#include "bbox.h"
#include "obbox.h"
#include "format.h"
#include "matrix4.h"

const char *Quad::_typeStr[] = {	// XXX same order as enum
		"Plane",
		"PlaneX",
		"PlaneY",
		"PlaneZ",
		"Quad",
		"Sphere",
		"Ellipsoid",
		"Cylinder",
		"CylinderX",
		"CylinderY",
		"CylinderZ",
		"EllCylinder",
		"EllCylinderX",
		"EllCylinderY",
		"EllCylinderZ",
		"Cone",
		"ConeX",
		"ConeY",
		"ConeZ",
		"Degenerate",
	};

/** Quad PLANE constructor type from all parameters */
Quad::Quad(const double cx,  const double cy,  const double cz,
	   const double c)
{
	reset();
	_type = Type::PLANE;
	Cx  = cx;
	Cy  = cy;
	Cz  = cz;
	C   = c;
	Z.set(Cx,Cy,Cz);
	Z.normalize();
	computeAbs();
	findSubType();
	init();
} // Quad

/** Quad constructor type from all parameters */
Quad::Quad( const double cxx, const double cyy, const double czz,
	    const double cxy, const double cxz, const double cyz,
	    const double cx,  const double cy,  const double cz,
	    const double c)
{
	reset();
	_type = Type::QUAD;
	Cxx = cxx;
	Cyy = cyy;
	Czz = czz;
	Cxy = cxy;
	Cxz = cxz;
	Cyz = cyz;
	Cx  = cx;
	Cy  = cy;
	Cz  = cz;
	C   = c;
	computeAbs();
	//findType(false);
	init();
} // Quad

/** Quad constructor from location Point p with one value
 * @param p	center of quadratic
 * @param value	one floating point value like radius or plane position
 */
Quad::Quad(Type t, const Point& p, const double value)
{
	reset();
	_type = t;
	P = p;
	Z = Vector::O;
	switch (type()) {
		case Type::PLANE_X:
			P.x = value;
			Cx  = 1;
			C   = -value;
			Z   = Vector::Xo;
			break;

		case Type::PLANE_Y:
			P.y = value;
			Cy  = 1;
			C   = -value;
			Z   = Vector::Yo;
			break;

		case Type::PLANE_Z:
			P.z = value;
			Cz  = 1;
			C   = -value;
			Z   = Vector::Zo;
			break;

		case Type::SPHERE:
			R = value;
			Cxx = Cyy = Czz = 1;
			Cx = -2*P.x;
			Cy = -2*P.y;
			Cz = -2*P.z;
			C  = P.length2() - R*R;
			break;

		case Type::CYLINDER_X:
			R   = value;
			P.x = 0;
			Cyy = Czz = 1;
			Cy  = -2*P.y;
			Cz  = -2*P.z;
			C   = P.length2() - R*R;
			Z   = Vector::Xo;
			break;

		case Type::CYLINDER_Y:
			R   = value;
			P.y = 0;
			Cxx = Czz = 1;
			Cx  = -2*P.x;
			Cz  = -2*P.z;
			C   = P.length2() - R*R;
			Z   = Vector::Yo;
			break;

		case Type::CYLINDER_Z:
			R   = value;
			P.z = 0;
			Cxx = Cyy = 1;
			Cx  = -2*P.x;
			Cy  = -2*P.y;
			C   = P.length2() - R*R;
			Z   = Vector::Zo;
			break;

		default:
			assert(0);
	}
	computeAbs();
	findSubType();
	init();
} // Quad

/** Quad constructor type from location Point p with one value
 * @param aP	center of quadratic
 * @param aZ	z defining vector
 * @param value	one floating point value like radius
 */
Quad::Quad(Type t, const Point& aP, const Vector& aZ, const double value)
{
	reset();
	double norm = Max(aP.max(), aZ.max(), Abs(value));
	_type = t;
	P = aP;
	Z = aZ.unit();
	switch (type()) {
		case Type::CYLINDER:
			R   =  value;
			Cxx =  (1-Z.x) * (1+Z.x);
			Cyy =  (1-Z.y) * (1+Z.y);
			Czz =  (1-Z.z) * (1+Z.z);
			Cxy = -2*Z.x*Z.y;
			Cxz = -2*Z.x*Z.z;
			Cyz = -2*Z.y*Z.z;
			Cx  =  2*((Z*P)*Z.x - P.x);
			Cy  =  2*((Z*P)*Z.y - P.y);
			Cz  =  2*((Z*P)*Z.z - P.z);
			C   =  P.length2() - Sq(Z*P) - R*R;
			break;
		default:
			assert(0);
	}
	computeAbs(sqrt(norm));
	findSubType();
	init();
} // Quad

/** Quad constructor from type at location Point p with one value
 * @param aP		center of quadratic
 * @param aZ		z defining vector
 * @param value		first floating point value like radius
 * @param value2	second floating point value like radius
 */
Quad::Quad(Type t, const Point& aP, const Vector& aZ, const double value, const double value2)
{
	reset();
	double norm = Max(aP.max(), aZ.max(), Abs(value), Abs(value2));
	_type = t;
	P = aP;
	Z = aZ;
	double H = Z.normalize();
	switch (type()) {
		case Type::CONE:
			R   = value;
			Ra  = value2;
			slope = (Ra-R)/H;
			s21 = 1 + Sq(slope);
			Cxx = 1 - s21*Sq(Z.x);
			Cyy = 1 - s21*Sq(Z.y);
			Czz = 1 - s21*Sq(Z.z);
			Cxy =-2*s21*Z.x*Z.y;
			Cxz =-2*s21*Z.x*Z.z;
			Cyz =-2*s21*Z.y*Z.z;
			Cx  = 2*(s21*(Z*P)*Z.x - slope*R*Z.x - P.x);
			Cy  = 2*(s21*(Z*P)*Z.y - slope*R*Z.y - P.y);
			Cz  = 2*(s21*(Z*P)*Z.z - slope*R*Z.z - P.z);
			C   = P.length2() - s21*Sq(Z*P) - R*R + 2*slope*R*(Z*P);
			break;
		default:
			assert(0);
	}
	computeAbs(sqrt(norm));
	findSubType();
	init();
} // Quad

/** set quad from matrix */
Quad& Quad::operator =(const Matrix4& m)
{
	reset();
	_type = Type::QUAD;

	Cxx = m(0,0);
	Cyy = m(1,1);
	Czz = m(2,2);
	Cxy = 2 * m(1,0);
	Cxz = 2 * m(2,0);
	Cyz = 2 * m(2,1);
	Cx  = 2 * m(3,0);
	Cy  = 2 * m(3,1);
	Cz  = 2 * m(3,2);
	C   = m(3,3);
	computeAbs();
	//findType(false);
	init();
	return *this;
} // operator =

/** reset all quad parameters */
void Quad::reset()
{
	_type = Type::DEGENERATE;
	Cxx = Cyy = Czz = 0;
	Cxy = Cxz = Cyz = 0;
	Cx  = Cy  = Cz  = 0;
	C   = 0;
	P = Vector::O;
	Z = Vector::Zo;
	R = Ra = 0;
} // reset

/** initialize some private parameters of quad */
void Quad::init()
{
	_dist1D =  sqrt(Sq(Cx) + Sq(Cy) + Sq(Cz));
	_dist2a = -sqrt(Sq(Cxx) + Sq(Cyy) + Sq(Czz) + 0.5*(Sq(Cxy) + Sq(Cxz) + Sq(Cyz)));
} // init

/** compute absolute values needed for accuracy */
void Quad::computeAbs(const double norm)
{
	aCxx   = Sq(norm)*Abs(Cxx);
	aCyy   = Sq(norm)*Abs(Cyy);
	aCzz   = Sq(norm)*Abs(Czz);
	aCxy   = Sq(norm)*Abs(Cxy);
	aCxz   = Sq(norm)*Abs(Cxz);
	aCyz   = Sq(norm)*Abs(Cyz);
	aCxyz0 = norm*Abs(Cx) + Abs(Cy) + Abs(Cz) + Abs(C);
} // computeAbs

/** @return type detected from Quad parameters */
Quad::Type Quad::detectType() const
{
	Matrix3 M3;
	Matrix4 M4;

	matrix3(&M3);
	matrix(&M4);

	double D = M4.det();
	double eig[3];
	M3.eigenValuesOfSymmetric(eig);
	int e = M3.rank();
	int E = M4.rank();

	bool same = (Cmp0(eig[0],SMALL) == Cmp0(eig[0],SMALL) and
	             Cmp0(eig[0],SMALL) == Cmp0(eig[2],SMALL));

	if (e==3) {
		if (E==4) {
			if (same) {
				if (D<0)
					return Type::ELLIPSOID;			// #1
				else
					return Type::IMG_ELLIPSOID;		// #2
			} else {
				if (D>0)
					return Type::HYPERBOLOID_1_SHEET;	// #3
				else
					return Type::HYPERBOLOID_2_SHEET;	// #4
			}
		} else
		if (E==3) {
			if (!same)
				return Type::CONE;				// #5
			else
				return Type::IMG_CONE;				// #6
		}
	} else
	if (e==2) {
		if (E==4) {
			if (D<0 and same) return Type::ELLIPTIC_PARABOLOID;	// #7
			else
			if (D>0 and !same) return Type::HYPERBOLIC_PARABOLOID;	// #8
		} else
		if (E==3) {
			if (same) {
				return Type::ELLIPTIC_CYLINDER;			// #9
				// ????
				// FIXME return Type::IMG_ELLIPTIC_CYLINDER;	// #10
			} else
				return Type::HYPERBOLIC_CYLINDER;		// #11
		} else
		if (E==2) {
			if (!same)
				return Type::INTERSECTING_PLANES;		// #12
			else
				return Type::IMG_INTERSECTING_PLANES;		// #13
		}
	} else
	if (e==1) {
		if (E==3)
			return Type::PARABOLIC_CYLINDER;			// #14
		else
		if (E==2) {
			return Type::PARALLEL_PLANES;				// #15
			//or
			//return Type::IMG_PARALLEL_PLANES;			// #16
		} else
		if (E==1)
			return Type::COINCIDENT_PLANES;				// #17
	}
	return Type::DEGENERATE;
} // detectType

/** find sub-type of quadratic surface depending on the initial type
 * e.g. plane <-> planeX/YZ
 * @param eps	accuracy to check operations
 */
void Quad::findSubType(double eps)
{
	double norm = Max(Max(aCxx,     aCyy,     aCzz),
			  Max(Abs(Cxy), Abs(Cxz), Abs(Cyz)),
			  Max(Sq(Cx),   Sq(Cy),   Sq(Cz)));
	norm = sqrt(norm);
	norm = (Abs(norm)>SMALL? norm : 1.0);
	eps *= norm;

	switch (type()) {
		case Type::PLANE:
		case Type::PLANE_X:
		case Type::PLANE_Y:
		case Type::PLANE_Z:
			if (Cx>eps and Eq0(Cy,eps) and Eq0(Cz,eps)) {
				_type = Type::PLANE_X;
				P.y = P.z = 0.0;	// Correct P?
				P.x = -C/Cx;
			} else
			if (Eq0(Cx,eps) and Cy>eps and Eq0(Cz,eps)) {
				_type = Type::PLANE_Y;
				P.x = P.z = 0.0;	// Correct P
				P.y = -C/Cy;
			} else
			if (Eq0(Cx,eps) and Eq0(Cy,eps) and Cz>eps) {
				P.x = P.y = 0.0;	// Correct P
				P.z = -C/Cz;
				_type = Type::PLANE_Z;
			} else
				_type = Type::PLANE;
			break;

		case Type::CYLINDER:
		case Type::CYLINDER_X:
		case Type::CYLINDER_Y:
		case Type::CYLINDER_Z:
			switch (Z.direction(eps)) {
				case Vector::Direction::X:
				case Vector::Direction::Xn:
					_type = Type::CYLINDER_X;
					break;
				case Vector::Direction::Y:
				case Vector::Direction::Yn:
					_type = Type::CYLINDER_Y;
					break;
				case Vector::Direction::Z:
				case Vector::Direction::Zn:
					_type = Type::CYLINDER_Z;
					break;
				default:
					_type = Type::CYLINDER;
			}
			break;

//		case Type::CONE:
//		case Type::CONE_X:
//		case Type::CONE_Y:
//		case Type::CONE_Z:
//			switch (Z.direction(eps)) {
//				case Vector::Direction::X:
//				case Vector::Direction::Xn:
//					_type = Type::CONE_X;
//					break;
//				case Vector::Direction::Y:
//				case Vector::Direction::Yn:
//					_type = Type::CONE_Y;
//					break;
//				case Vector::Direction::Z:
//				case Vector::Direction::Zn:
//					_type = Type::CONE_Z;
//					break;
//				default:
//					_type = Type::CONE;
//			}
//			break;

		default:
			/* do nothing */;
	}
} // findSubType

/** operator * */
Quad& Quad::operator *= (const double d)
{
	Cxx *= d;
	Cyy *= d;
	Czz *= d;
	Cxy *= d;
	Cxz *= d;
	Cyz *= d;
	Cx  *= d;
	Cy  *= d;
	Cz  *= d;
	C   *= d;
	init();

	return *this;
} // operator *

/** normalize coefficients and calculate accuracy parameter
 * for checking operations
 */
void Quad::normalize()
{
	double norm;
	if (isplane())
		norm = Max(Sq(Cx),  Sq(Cy),  Sq(Cz));
	else
		norm = Max(Max(Abs(Cxx), Abs(Cyy), Abs(Czz)),
			   Max(Abs(Cxy), Abs(Cxz), Abs(Cyz)));
	if (Abs(norm) > SMALL)
		(*this) *= 1.0/norm;
/*
	if (Eq0(Cxx,TOOSMALL)) Cxx = 0.0;
	if (Eq0(Cyy,TOOSMALL)) Cyy = 0.0;
	if (Eq0(Czz,TOOSMALL)) Czz = 0.0;
	if (Eq0(Cxy,TOOSMALL)) Cxy = 0.0;
	if (Eq0(Cxz,TOOSMALL)) Cxz = 0.0;
	if (Eq0(Cyz,TOOSMALL)) Cyz = 0.0;
	if (Eq0(Cx,TOOSMALL))  Cx  = 0.0;
	if (Eq0(Cy,TOOSMALL))  Cy  = 0.0;
	if (Eq0(Cz,TOOSMALL))  Cz  = 0.0;
	if (Eq0(C,TOOSMALL))   C   = 0.0;
*/
} // normalize

/** _bboxLimits
 * @return a 1d-dimensional axis-aligned bounding limits along Z-axis
 * The routine should be called with rotated x-y-z arguments
 * to get for each axis the bounding limits
 *
 * The z-limits (critical points) are located at:   grad(Q) = [0,0,z]
 * Which collapses to the following equations:
 * 1. Q(x,y,z)  = 0
 * 2. Qx = diff(Q,x) = 0		first partial derivative
 * 3. Qy = diff(Q,y) = 0		first partial derivative
 * 4. Qxx = diff(diff(Q,x),x)		second partial derivative
 * 5. Qyy = diff(diff(Q,y),y)		second partial derivative
 * 6. Qxy = diff(diff(Q,x),y)		second partial derivative
 *
 * We have a min/max point if the concavity is the same in all directions
 *    D = Qxx*Qyy - Qxy^2
 *
 * 1. If D > 0, then the concavity is the same in all directions.
 *     (a) If Qyy > 0, then it is concave up in all
 *         directions. So Q(x,y,z) is a local minimum.
 *     (b) If Qyy < 0, then it is concave down in all
 *         directions. So Q(x,y,z) is a local maximum.
 * 2. If D < 0, then the concavity changes in some direction.
 *    (x,y,z) is a saddle point.
 * 3. If D = 0, the test is inconclusive (need a contour map)
 *
 * solve([Qx(x,y,z)=0, Qy(x,y,z)=0],[x,y])
 *
 * X(z) = (Cxy*Cyz*z-2*Cxz*Cyy*z  - 2*Cx*Cyy+Cxy*Cy) / (-Cxy^2+4*Cxx*Cyy)
 * Y(z) = (-2*Cxx*Cyz*z+Cxy*Cxz*z + Cx*Cxy-2*Cxx*Cy) / (-Cxy^2+4*Cxx*Cyy)
 *
 * substitute in #1 we have
 *
 *	+ Cxx*K^2	*z^2
 *	+ Cxy*K*M	*z^2
 *	+ Cxz*K		*z^2
 *	+ Cyy*M^2	*z^2
 *	+ Cyz*M		*z^2
 *	+ Czz		*z^2

 *	+ 2*Cxx*K*L	*z
 *	+ 2*Cyy*M*N	*z
 *	+ Cx*K		*z
 *	+ Cxy*K*N	*z
 *	+ Cxy*L*M	*z
 *	+ Cxz*L		*z
 *	+ Cy*M		*z
 *	+ Cyz*N		*z
 *	+ Cz		*z

 *	+ C
 *	+ Cx*L
 *	+ Cxx*L^2
 *	+ Cxy*L*N
 *	+ Cy*N
 *	+ Cyy*N^2
 */
static void _bboxLimits(double Cxx, double Cyy, double Czz,
			double Cxy, double Cxz, double Cyz,
			double Cx,  double Cy,  double Cz,
			double C,   double norm,
			double* zmin, double* zmax)
{
	*zmin = -INFINITE;
	*zmax = +INFINITE;
	double eps = SMALL*norm;

	double D = 4.0*Cxx*Cyy - Sq(Cxy);	// determinate of Qx=0, Qy=0 system

	/* check if D is zero, there is no solution e.g. tip of a cone => infinite */
	if (Eq0(D,eps)) return;

	// X(z) = K*z + L
	double K = (Cxy*Cyz - 2*Cxz*Cyy) / D;
	double L = (Cxy*Cy  - 2*Cx*Cyy ) / D;

	// Y(z) = M*z + N
	double M = (Cxy*Cxz - 2*Cxx*Cyz) / D;
	double N = (Cx*Cxy  - 2*Cxx*Cy)  / D;

	// substitute to Q(x,y,z) = 0 => Q(X(z),Y(z),z) = 0
	double AA =
		+ Cxx*Sq(K)
		+ Cxy*K*M
		+ Cxz*K
		+ Cyy*Sq(M)
		+ Cyz*M
		+ Czz;

	double BB =
		+ 2*Cxx*K*L
		+ 2*Cyy*M*N
		+ Cx*K
		+ Cxy*K*N
		+ Cxy*L*M
		+ Cxz*L
		+ Cy*M
		+ Cyz*N
		+ Cz;

	double CC =
		+C
		+Cx*L
		+Cxx*Sq(L)
		+Cxy*L*N
		+Cy*N
		+Cyy*Sq(N);

	if (Eq0(AA,eps)) {
		if (!Eq0(BB,eps)) {
			if (BB<0)
				*zmin = -CC/BB;
			else
				*zmax = -CC/BB;
		}
		return;
	}

	double z1, z2;
	if (!quadratic(BB/AA, CC/AA, &z1, &z2, eps)) return;
	if (z1>z2) Swap(z1,z2);

	// Check dQ/dz for max or minimum
	double x = K*z1 + L;
	double y = M*z1 + N;
	double dz = 2*Czz*z1 + Cxz*x + Cyz*y + Cz;
	if (dz < -norm*1e-5) *zmin = z1;	// Be careful on hyperbolic

	x = K*z2 + L;
	y = M*z2 + N;
	dz = 2*Czz*z2 + Cxz*x + Cyz*y + Cz;
	if (dz > norm*1e-5) *zmax = z2;
} // _bboxLimits

/** @return bounding box of quadric (in case of ellipsoid/cylinder/...) */
BBox Quad::bbox() const
{
	BBox bb;
	switch (type()) {
		case Type::SPHERE:
			bb.add(P.x-R, P.y-R, P.z-R);
			bb.add(P.x+R, P.y+R, P.z+R);
			break;

		default: {
			//Matrix4 QM;
			//matrix(&QM);
			//std::cout << "<M=\n"  << QM << std::endl;
			//std::cout << "Rank: " << QM.rank() << std::endl;
			//std::cout << ">M=\n"  << QM << std::endl;
			//auto t = detectType();
			//std::cout << "Type=" << (int)t << " " << typeid(t).name() << std::endl;
			switch (detectType()) {
				//case Type::ELLIPSOID:			// #1
				case Type::IMG_ELLIPSOID:		// #2
				case Type::HYPERBOLOID_1_SHEET:		// #3
				case Type::HYPERBOLOID_2_SHEET:		// #4
				case Type::CONE:			// #5
				case Type::IMG_CONE:			// #6
				//case Type::ELLIPTIC_PARABOLOID:	// #7
				case Type::HYPERBOLIC_PARABOLOID:	// #8
				//case Type::ELLIPTIC_CYLINDER:		// #9
				case Type::IMG_ELLIPTIC_CYLINDER:	// #10
				case Type::HYPERBOLIC_CYLINDER:		// #11
				//case Type::INTERSECTING_PLANES:	// #12
				case Type::IMG_INTERSECTING_PLANES:	// #13
				//case Type::PARABOLIC_CYLINDER:	// #14
				//case Type::PARALLEL_PLANES:		// #15
				case Type::IMG_PARALLEL_PLANES:		// #16
				//case Type::COINCIDENT_PLANES:		// #17
				case Type::DEGENERATE:
					bb.infinite();
					return bb;
				default:
					break;
			}

			double norm =	Max(Max(aCxx,        aCyy,     aCzz),
					    Max(Abs(Cxy), Abs(Cxz), Abs(Cyz)),
					    Max( Sq(Cx),   Sq(Cy),   Sq(Cz)),
					    Abs(C));

			/* limits on Z */
			double zmin, zmax;
			_bboxLimits(Cxx, Cyy, Czz,
				    Cxy, Cxz, Cyz,
				    Cx,  Cy,  Cz,
				    C,   norm,
				    &zmin, &zmax);


			/* limits on X (rotate as x->y,  y->z,  z->x) */
			double xmin, xmax;
			_bboxLimits(Cyy, Czz, Cxx,
				    Cyz, Cxy, Cxz,
				    Cy,  Cz,  Cx,
				    C,   norm,
				    &xmin, &xmax);

			/* limits on Y (rotate as x->z,  y->x,  z->y*/
			double ymin, ymax;
			_bboxLimits(Czz, Cxx, Cyy,
				    Cxz, Cyz, Cxy,
				    Cz,  Cx,  Cy,
				    C,   norm,
				    &ymin, &ymax);

			// FIXME Check for concave?
			bb.add(xmin, ymin, zmin);
			bb.add(xmax, ymax, zmax);
		}
	}

	if (!bb.isValid())
		bb.infinite();

	return bb;
} // bbox

#ifdef OBBOX
/** @return oriented bounding box of quadric (in case of ellipsoid/cylinder/...) */
OBBox Quad::obbox() const
{
	OBBox bb;
	double xmin, xmax;
	double ymin, ymax;
	double zmin, zmax;

	double norm =	Max(Max(aCxx, aCyy, aCzz),
			    Max(Abs(Cxy), Abs(Cxz), Abs(Cyz)),
			    Max(Sq(Cx),  Sq(Cy),  Sq(Cz)),
			    Abs(C));

	/* limits on Z */
	_bboxLimits(	Cxx, Cyy, Czz,
			Cxy, Cxz, Cyz,
			Cx,  Cy,  Cz,
			C, norm,
			&zmin, &zmax);

	/* limits on X */
	_bboxLimits(	Cyy, Czz, Cxx,
			Cyz, Cxy, Cxz,
			Cy,  Cz,  Cx,
			C, norm,
			&xmin, &xmax);

	/* limits on Y */
	_bboxLimits(	Czz, Cxx, Cyy,
			Cxz, Cyz, Cxy,
			Cz,  Cx,  Cy,
			C, norm,
			&ymin, &ymax);

	bb.add(xmin, ymin, zmin);
	bb.add(xmax, ymax, zmax);

	if (!bb.isValid())
		bb.infinite();

	return bb;
} // obbox
#endif

/** Smallest closest distance to the surface in any direction.
 * My approximation to distance of point to quadric.
 * The algorithm works well when the point is close to quad.
 * using the normal on the surface constructs a line and finds
 * the intersection of the line with the conic
 * @param x,y,z	point to find distance
 * @return returns a safe distance to the boundary
 *         (negative inside, positive outside)
 *         // FIXME verify if it is needed here!!!!
 */
double Quad::distance(const double x, const double y, const double z) const
{
	// XXX +/- depending on sense
	switch (type()) {
		case Type::PLANE:
			// First order approximation of the distance |f(x,y,z)| / ||grad(x,y,z)||
			// Calculated in init() as _dist1D
			return Q(x,y,z) / _dist1D;

		case Type::PLANE_X:
			return x-P.x;

		case Type::PLANE_Y:
			return y-P.y;

		case Type::PLANE_Z:
			return z-P.z;

		case Type::SPHERE:
			return sqrt(Sq(x-P.x) + Sq(y-P.y) + Sq(z-P.z)) - R;

		case Type::CYLINDER: {
				Vector S(x-P.x, y-P.y, z-P.z);
				double AS = S.length2() - Sq(Z*S);
				AS = AS>0? sqrt(AS) : 0;
				return AS - R;
			}

		case Type::CYLINDER_X:
			return sqrt(Sq(y-P.y) + Sq(z-P.z)) - R;

		case Type::CYLINDER_Y:
			return sqrt(Sq(x-P.x) + Sq(z-P.z)) - R;

		case Type::CYLINDER_Z:
			return sqrt(Sq(x-P.x) + Sq(y-P.y)) - R;

//		case Type::CONE_X:
//		case Type::CONE_Y:
//		case Type::CONE_Z:
		case Type::CONE: {
				Vector S(x-P.x, y-P.y, z-P.z);
				double zS = Z*S;
				double AS = S.length2() - Sq(zS);
				AS = AS>0? sqrt(AS) : 0;
				double r = zS*slope + R;
				if (r>0)
					return AS - r;
				else
					return (S-zS*Z).length();
			}

		case Type::QUAD: {
			// Taubin, second order approximation of the distance
			// a) Calculated in init() as _dist2a
			// length of gradient
			double b = -sqrt(Sq(2.0*Cxx*x +     Cxy*y +     Cxz*z + Cx) +
					 Sq(    Cxy*x + 2.0*Cyy*y +     Cyz*z + Cy) +
					 Sq(    Cxz*x +     Cyz*y + 2.0*Czz*z + Cz) );
			// value
			double c = Abs(Q(x,y,z));
			double D = sqrt(b*b - 4.0*_dist2a*c);
			return 2.0*c/(-b+D);
		}

		default:
			assert(false);
			return 0.0;
	}
} // distance

/**
 * intersect a line with the quad and return the two intersection distances
 * @param p	position vector
 * @param v	direction vector (has to be normalized)
 * @param tmin, tmax	distance of minimum, maximum intersection points
 * @return	number of intersection points,
 *		if positive then [tmin..tmax]
 *		if negative then (-inf..tmin] | [tmax..+inf)
 */
int Quad::distance(const double x, const double y, const double z,
		   const double u, const double v, const double w,
		   double* tmin, double* tmax) const
{
	assert(Eq(Sq(u) + Sq(v) + Sq(w), 1, 1e-7));
	double a, b, c;
	switch (type()) {
		case Type::PLANE:
			b = Cx*u + Cy*v + Cz*w;
			c = Q(x, y, z);
			if (Eq0(b, SMALL)) goto NO_SOLUTION;
			if (b > 0.0) {
				*tmin = -INFINITE;
				*tmax = -c/b;
			} else {
				*tmin = -c/b;
				*tmax = INFINITE;
			}
			return 1;	// plane returns 2 solution (1+inf)

		case Type::PLANE_X:
			if (Eq0(u,SMALL)) goto NO_SOLUTION;
			if (u > SMALL) {
				*tmin = -INFINITE;
				*tmax = (P.x - x) / u;
			} else {
				*tmin = (P.x - x) / u;
				*tmax = INFINITE;
			}
			return 2;

		case Type::PLANE_Y:
			if (Eq0(v,SMALL)) goto NO_SOLUTION;
			if (v > SMALL) {
				*tmin = -INFINITE;
				*tmax = (P.y - y) / v;
			} else {
				*tmin = (P.y - y) / v;
				*tmax = INFINITE;
			}
			return 1;

		case Type::PLANE_Z:
			if (Eq0(w, SMALL)) goto NO_SOLUTION;
			if (w > SMALL) {
				*tmin = -INFINITE;
				*tmax = (P.z - z) / w;
			} else {
				*tmin = (P.z - z) / w;
				*tmax = INFINITE;
			}
			return 1;

		case Type::SPHERE: {
				Vector S(x-P.x, y-P.y, z-P.z);
				b = 2*(u*S.x + v*S.y + w*S.z);
				c = S.length2() - R*R;
				if (!quadratic(b, c, tmin, tmax, SMALL)) goto NO_SOLUTION;
			}
			break;

		/*     |          S = v*t + r      S-P = v*t + (r-P) = v*t + rP
		 *    A+----+S    |AS| = R
		 *     |   /      |AP| = Z*(S-P)
		 *   ^ |  /
		 *   | | /        |AP|² + |AS|² = (S-P)²
		 *   Z |/     <=> (S-P)² - Z*(S-P) - R² = 0
		 *     +P     <=> (v*t+rP)
		 *            <=>  S² + P² - 2*P*S - Z*S + Z*P - R² = 0
		 */

//		case Type::CONE_X:
//		case Type::CONE_Y:
//		case Type::CONE_Z:
		case Type::CONE: {
				Vector S(x-P.x, y-P.y, z-P.z);
				double vS = u*S.x + v*S.y + w*S.z;
				double vZ = u*Z.x + v*Z.y + w*Z.z;
				double zS = Z*S;

				a = 1 - s21*Sq(vZ);
				if (Eq0(a,SMALL)) goto NO_SOLUTION;
				b = 2*(vS - (s21*zS + slope*R)*vZ);
				c = (S.length2() - s21*Sq(zS) - 2*slope*R*zS - Sq(R));
				b /= a;
				c /= a;
				int n = quadratic(b, c, tmin, tmax, SMALL);
				if (n>1) {
					if (*tmin>*tmax) Swap(*tmin, *tmax);
					// Concave if f''=2a is negative
					//  (-inf..tmin] | [tmax..+inf)
					if (a<0.0) n = -n;
				}
				return n;
			}
			break;

		case Type::DEGENERATE:
			assert(false);
			goto NO_SOLUTION;

		default:
			//assert(type()==Type::QUAD);
			a = (Cxx*u + Cxy*v + Cxz*w)*u +
				   (Cyy*v + Cyz*w)*v +
				   Czz*w*w;
			b = 2.0*(Cxx*u*x + Cyy*v*y + Czz*w*z)
				 + Cxy*(u*y + x*v)
				 + Cxz*(u*z + x*w)
				 + Cyz*(v*z + y*w)
				 + Cx*u + Cy*v + Cz*w;
			c = Q(x, y, z);

			if (Eq0(a, SMALL)) {
				// plane? it should never come here
				if (Eq0(b, SMALL)) goto NO_SOLUTION;
				if (b > 0.0) {
					*tmin = -INFINITE;
					*tmax = -c/b;
				} else {
					*tmin = -c/b;
					*tmax = INFINITE;
				}
				return 1;
			} else {
				b /= a;
				c /= a;
				int n = quadratic(b, c, tmin, tmax, SMALL);
				if (n>1) {
					if (*tmin>*tmax) Swap(*tmin, *tmax);
					// Concave if f''=2a is negative
					//  (-inf..tmin] | [tmax..+inf)
					if (a<0.0) n = -n;
				}
				return n;
			}
	}

	// Standard exit with two solutions
	if (*tmin > *tmax) Swap(*tmin, *tmax);
	return 2;

NO_SOLUTION:
	*tmin =  INFINITE;	// no solution
	*tmax = -INFINITE;
	return 0;
} // distance

/** return quad matrix */
void Quad::matrix3(Matrix3* m) const
{
	(*m)(0,0) = Cxx;
	(*m)(1,1) = Cyy;
	(*m)(2,2) = Czz;

	(*m)(0,1) = Cxy/2.0; (*m)(1,0) = Cxy/2.0;
	(*m)(0,2) = Cxz/2.0; (*m)(2,0) = Cxz/2.0;
	(*m)(1,2) = Cyz/2.0; (*m)(2,1) = Cyz/2.0;
} // matrix

/** return quad matrix */
void Quad::matrix(Matrix4* m) const
{
	(*m)(0,0) = Cxx;
	(*m)(1,1) = Cyy;
	(*m)(2,2) = Czz;

	(*m)(0,1) = Cxy/2.0; (*m)(1,0) = Cxy/2.0;
	(*m)(0,2) = Cxz/2.0; (*m)(2,0) = Cxz/2.0;
	(*m)(1,2) = Cyz/2.0; (*m)(2,1) = Cyz/2.0;

	(*m)(0,3) = Cx/2.0;  (*m)(3,0) = Cx/2.0;
	(*m)(1,3) = Cy/2.0;  (*m)(3,1) = Cy/2.0;
	(*m)(2,3) = Cz/2.0;  (*m)(3,2) = Cz/2.0;

	(*m)(3,3) = C;
} // matrix

/**
 * Translate axes of quad by dx, dy, dz
 */
void Quad::translate(const double dx, const double dy, const double dz)
{
	C += (Cxx*dx + Cxy*dy + Cxz*dz + Cx)*dx +
	     (Cyy*dy + Cyz*dz + Cy)*dy +
	     (Czz*dz + Cz)*dz;
	Cx = 2.0*Cxx*dx + Cxy*dy + Cxz*dz + Cx;
	Cy = 2.0*Cyy*dy + Cxy*dx + Cyz*dz + Cy;
	Cz = 2.0*Czz*dz + Cxz*dx + Cyz*dy + Cz;

	P.x -=  dx;
	P.y -= dy;
	P.z -= dz;
	init();
} // translate

/** transform the quadric using matrix M     (Mt*Q*M)
 * @param M	matrix to transform quadric
 */
void Quad::transform(const Matrix4& M, const Matrix4& Minv)
{
	DEBUG(fpetrap(true));
	P = Minv * P;
	Z = Minv * Z;

	double det = Minv.det3();
	if (!Eq(Abs(det), 1.0, SMALL5)) {
		Z.normalize();
		double scale = cbrt(Abs(det));
		R  *= scale;
		Ra *= scale;
	}

	if (isplane()) {
		// For planes only
		double a = Cx;
		double b = Cy;
		double c = Cz;
		Cx = a*M(0,0) + b*M(1,0) + c*M(2,0);
		Cy = a*M(0,1) + b*M(1,1) + c*M(2,1);
		Cz = a*M(0,2) + b*M(1,2) + c*M(2,2);
		C += a*M(0,3) + b*M(1,3) + c*M(2,3);
		_type = Type::PLANE;
	} else {
		double m00 = Cxx*M(0,0) + (Cxy*M(1,0) + Cxz*M(2,0) + Cx*M(3,0))/2.0;
		double m01 = Cxx*M(0,1) + (Cxy*M(1,1) + Cxz*M(2,1) + Cx*M(3,1))/2.0;
		double m02 = Cxx*M(0,2) + (Cxy*M(1,2) + Cxz*M(2,2) + Cx*M(3,2))/2.0;
		double m03 = Cxx*M(0,3) + (Cxy*M(1,3) + Cxz*M(2,3) + Cx*M(3,3))/2.0;

		double m10 = Cyy*M(1,0) + (Cxy*M(0,0) + Cyz*M(2,0) + Cy*M(3,0))/2.0;
		double m11 = Cyy*M(1,1) + (Cxy*M(0,1) + Cyz*M(2,1) + Cy*M(3,1))/2.0;
		double m12 = Cyy*M(1,2) + (Cxy*M(0,2) + Cyz*M(2,2) + Cy*M(3,2))/2.0;
		double m13 = Cyy*M(1,3) + (Cxy*M(0,3) + Cyz*M(2,3) + Cy*M(3,3))/2.0;

		double m20 = Czz*M(2,0) + (Cxz*M(0,0) + Cyz*M(1,0) + Cz*M(3,0))/2.0;
		double m21 = Czz*M(2,1) + (Cxz*M(0,1) + Cyz*M(1,1) + Cz*M(3,1))/2.0;
		double m22 = Czz*M(2,2) + (Cxz*M(0,2) + Cyz*M(1,2) + Cz*M(3,2))/2.0;
		double m23 = Czz*M(2,3) + (Cxz*M(0,3) + Cyz*M(1,3) + Cz*M(3,3))/2.0;

		double m30 = (Cx*M(0,0) + Cy*M(1,0) + Cz*M(2,0))/2.0 + C*M(3,0);
		double m31 = (Cx*M(0,1) + Cy*M(1,1) + Cz*M(2,1))/2.0 + C*M(3,1);
		double m32 = (Cx*M(0,2) + Cy*M(1,2) + Cz*M(2,2))/2.0 + C*M(3,2);
		double m33 = (Cx*M(0,3) + Cy*M(1,3) + Cz*M(2,3))/2.0 + C*M(3,3);

		Cxx =      M(0,0)*m00 + M(1,0)*m10 + M(2,0)*m20 + M(3,0)*m30;
		Cxy = 2.0*(M(0,0)*m01 + M(1,0)*m11 + M(2,0)*m21 + M(3,0)*m31);
		Cxz = 2.0*(M(0,0)*m02 + M(1,0)*m12 + M(2,0)*m22 + M(3,0)*m32);
		Cx  = 2.0*(M(0,0)*m03 + M(1,0)*m13 + M(2,0)*m23 + M(3,0)*m33);

		Cyy =      M(0,1)*m01 + M(1,1)*m11 + M(2,1)*m21 + M(3,1)*m31;
		Cyz = 2.0*(M(0,1)*m02 + M(1,1)*m12 + M(2,1)*m22 + M(3,1)*m32);
		Cy  = 2.0*(M(0,1)*m03 + M(1,1)*m13 + M(2,1)*m23 + M(3,1)*m33);

		Czz =      M(0,2)*m02 + M(1,2)*m12 + M(2,2)*m22 + M(3,2)*m32;
		Cz  = 2.0*(M(0,2)*m03 + M(1,2)*m13 + M(2,2)*m23 + M(3,2)*m33);

		double tmp = M(0,3)*m03 + M(1,3)*m13 + M(2,3)*m23 + M(3,3)*m33;
		C = (Abs(tmp) < Abs(C)*SMALL)? 0.0 : tmp;
	}

	findSubType();
	init();
	DEBUG(fpetrap(false));
} // transform

/** negate quad coefficients */
void Quad::negate()
{
	Cxx = -Cxx;
	Cyy = -Cyy;
	Czz = -Czz;
	Cxy = -Cxy;
	Cxz = -Cxz;
	Cyz = -Cyz;
	Cx  = -Cx;
	Cy  = -Cy;
	Cz  = -Cz;
	C   = -C;
	init();
} // negate

/** operator << */
std::ostream& operator << (std::ostream& s, const Quad& q)
{
	if (!q.isplane()) {
		fmt(s, q.Cxx, "x^2 ");
		fmt(s, q.Cyy, "y^2 ");
		fmt(s, q.Czz, "z^2 ");
		fmt(s, q.Cxy, "xy ");
		fmt(s, q.Cxz, "xz ");
		fmt(s, q.Cyz, "yz ");
	}
	fmt(s, q.Cx, "x ");
	fmt(s, q.Cy, "y ");
	fmt(s, q.Cz, "z ");
	fmt(s, q.C);
	s << "=0";
	s << " [" << q.typeStr() << "]";
	return s;
} /* operator << */
