/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Set of low discrepancy sampling functions as well as several geometric samplers
 *
 * Author: Christian.Theis@cern.ch
 * Date:   02-Apr-2020
 */

#ifndef RAND_SAMPLER_H
#define RAND_SAMPLER_H

#include <vector>

#include "math/eps.h"
#include "math/vector.h"
#include "math/matrix3.h"

// General helpers

void Shuffle(float* pSamp, unsigned long count, int dims, unsigned long randNr );

class CBaseSampler
{
public:

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	static float VanDerCorput( unsigned int n, unsigned int Scramble )

		// 1/7/2019 [theis]
		////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		n = (n << 16) | (n >> 16);
		n = ((n & 0x00ff00ff) << 8) | ((n & 0xff00ff00) >> 8);
		n = ((n & 0x0f0f0f0f) << 4) | ((n & 0xf0f0f0f0) >> 4);
		n = ((n & 0x33333333) << 2) | ((n & 0xcccccccc) >> 2);
		n = ((n & 0x55555555) << 1) | ((n & 0xaaaaaaaa) >> 1);
		n ^= Scramble;
		return (float)n / (float)0x100000000LL;
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	static float Sobol2( unsigned int n, unsigned int Scramble )

		// 1/7/2019 [theis]
		////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		for( unsigned int v = 1 << 31; n != 0; n >>= 1, v ^= v >> 1 )
			if( n & 0x1 )
			  Scramble ^= v;
		return (float)Scramble / (float)0x100000000LL;
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	static float LarcherPillichshammer2( unsigned int n, unsigned int Scramble )

		// 1/7/2019 [theis]
		////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		for( unsigned int v = 1 << 31; n != 0; n >>= 1, v |= v >> 1 )
			if( n & 0x1 )
			  Scramble ^= v;
		return (float)Scramble / (float)0x100000000LL;
	}


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	static void Sample02( unsigned long n, unsigned int pScramble[2], float pSample[2] )

		// 1/7/2019 [theis]
		////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		pSample[0] = VanDerCorput( n, pScramble[0] );
		pSample[1] = Sobol2( n, pScramble[1] );
	}


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	static Vector uniformSampleHemisphere( const Vector& zAxis, double rand1 /*= rand()*/, double rand2 /*= rand()*/ )

		//
		// 1/7/2019 [theis]
		////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		bool bVerbose = false;
		Matrix3 uvwBase;
		Vector randVec, vX, vY, vZ( zAxis );
		Vector zDir( 0.0, 0.0, 1.0 );
		double cosTheta = sqrt( 1.0 - rand1 );
		double sinTheta = sqrt( 1.0 - cosTheta );
		double phi = 2.0 * PI * rand2;

		// create a 3D basis aligned with the passed zAxis vector
		vZ.normalize();

		bVerbose && std::cerr << "Normal --> X: " << vZ[0] << " Y: " << vZ[1] << " Z: " << vZ[2] << std::endl << std::flush;

		if( fabs( vZ.x ) >= fabs( vZ.y ) ) {
			double invLength = 1.0 / sqrt( vZ.x * vZ.x + vZ.z * vZ.z );
			vX.x = +vZ.z * invLength;
			vX.y = 0.0;
			vX.z = -vZ.x * invLength;
		}
		else {
			double invLength = 1.0 / sqrt( vZ.y * vZ.y + vZ.z * vZ.z );
			vX.x = 0.0;
			vX.y = +vZ.z * invLength;
			vX.z = -vZ.y * invLength;
		}

		vY = vZ.cross( vX );

		uvwBase.make( vX, vY, vZ );
		uvwBase.transpose();

		// uniform sampling on hemisphere with the Z-axis (0,0,1) equal to the delimiting plane's normal
		randVec.set( cos( phi ) * sinTheta, sin( phi ) * sinTheta, cosTheta );
		bVerbose && std::cerr << "Canonical sampleDir --> X: " << randVec[0] << " Y: " << randVec[1] << " Z: " << randVec[2] << std::endl << std::flush;

		// transform to the base defined by the passed vector
		randVec = uvwBase * randVec;

		bVerbose && std::cerr << "Transformed sampleDir --> X: " << randVec[0] << " Y: " << randVec[1] << " Z: " << randVec[2] << std::endl << std::flush;
		bVerbose && std::cerr << "cos(N,R) = " << vZ.dot( randVec ) << std::endl << std::flush;

		if( bVerbose && vZ.dot( randVec ) < 0.0 ) {
			std::cerr << "ERROR: cos(N,R) < 0.0!!!" << std::endl << std::flush;
			exit( -1 );
		}

		randVec.normalize();            // safety catch
		return randVec;
	}

	// check if two vectors
	static bool  sameHemisphere( const Vector& Vec1, const Vector& Vec2 ) { return Vec1.dot( Vec2 ) > 0.f; }

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	static Vector uniformSampleSphere( double rand1, double rand2 )

		// [7/8/2019 theis]
		////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		double z = 1.0 - 2.0 * rand1;
		double r = sqrt( std::max( 0.0, 1.0 - z * z ) );
		double phi = 2.0 * PI * rand2;
		double x = r * cos( phi );
		double y = r * sin( phi );

		Vector randVec( x, y, z );
		randVec.normalize();                 // safety catch

		return randVec;
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	static void uniformSampleDisk( double& x, double& y, double rand1, double rand2 )

	// [7/8/2019 theis]
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		double r = sqrt( rand1 );
		double theta = 2.0 * PI * rand2;
		x = r * cos( theta );
		y = r * sin( theta );
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	static void concentricSampleDisk( double& dx, double& dy, double u1, double u2 )

	// [7/8/2019 theis]
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		double r, theta;
		// Map uniform random numbers to $[-1,1]^2$
		double sx = 2.0 * u1 - 1.0;
		double sy = 2.0 * u2 - 1.0;
		// Map square to $(r,\theta)$
		// Handle degeneracy at the origin
		if( sx == 0.0 && sy == 0.0 ) {
			dx = 0.0;
			dy = 0.0;
			return;
		}
		if( sx >= -sy ) {
			if( sx > sy ) {
			  // Handle first region of disk
			  r = sx;
			  if( sy > 0.0 )
			    theta = sy / r;
			  else
			    theta = 8.0 + sy / r;
			}
			else {
			  // Handle second region of disk
			  r = sy;
			  theta = 2.0 - sx / r;
			}
		}
		else {
			if( sx <= sy ) {
			  // Handle third region of disk
			  r = -sx;
			  theta = 4.0 - sy / r;
			}
			else {
			  // Handle fourth region of disk
			  r = -sy;
			  theta = 6.0 + sx / r;
			}
		}
		theta *= PI / 4.0;
		dx = r * cos( theta );
		dy = r * sin( theta );
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	static Vector uniformSampleDisk( const Vector& Normal, double BaseRadius, double Rand1, double Rand2 )

	// return position on a disk oriented in space w.r.t. passed normal vector
	// [5/22/2020 theis]
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		Vector Position;
		Vector ReflNormal1 = Vector( Normal.z, Normal.y, -Normal.x );
		Vector ReflNormal2 = Normal.cross( ReflNormal1 );

		double XOffset, YOffset;

		CBaseSampler::uniformSampleDisk( XOffset, YOffset, Rand1, Rand2 );
		XOffset *= BaseRadius;
		YOffset *= BaseRadius;

		Position = ReflNormal1 * XOffset + ReflNormal2 * YOffset;

		// return position on a disk oriented in space w.r.t. passed normal vector
		return Position;
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	static Vector concentricSampleDisk( const Vector& Normal, double BaseRadius, double Rand1, double Rand2 )

	// return position on a disk oriented in space w.r.t. passed normal vector
	// [5/22/2020 theis]
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		Vector Position;
		Vector ReflNormal1 = Vector( Normal.z, Normal.y, -Normal.x );
		Vector ReflNormal2 = Normal.cross( ReflNormal1 );

		double XOffset, YOffset;

		CBaseSampler::concentricSampleDisk( XOffset, YOffset, Rand1, Rand2 );
		XOffset *= BaseRadius;
		YOffset *= BaseRadius;

		Position = ReflNormal1 * XOffset + ReflNormal2 * YOffset;

		// return position on a disk oriented in space w.r.t. passed normal vector
		return Position;
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	static Vector cosineSampleHemisphere( const Vector& zAxis, double u1, double u2 )

	// Cosine weighted sampling of a hemisphere with a specific orientation
	// [7/8/2019 theis]
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		Matrix3 uvwBase;
		Vector  randVec, vX, vY, vZ( zAxis) ;
		double  X, Y;
		bool    bVerbose = false;

		concentricSampleDisk( X, Y, u1, u2 );

		randVec.x = X; randVec.y = Y;
		randVec.z = sqrt( std::max( SMALL3D, 1.0 - randVec.x * randVec.x - randVec.y * randVec.y ) );

		// transform from canonical system to a basis with the passed Z-axis
		// create a 3D basis aligned with the passed zAxis vector
		vZ.normalize();

		bVerbose && std::cerr << "Normal --> X: " << vZ[0] << " Y: " << vZ[1] << " Z: " << vZ[2] << std::endl << std::flush;

		if( fabs( vZ.x ) >= fabs( vZ.y ) ) {
			double invLength = 1.0 / sqrt( vZ.x * vZ.x + vZ.z * vZ.z );
			vX.x = +vZ.z * invLength;
			vX.y = 0.0;
			vX.z = -vZ.x * invLength;
		}
		else {
			double invLength = 1.0 / sqrt( vZ.y * vZ.y + vZ.z * vZ.z );
			vX.x = 0.0;
			vX.y = +vZ.z * invLength;
			vX.z = -vZ.y * invLength;
		}

		vY = vZ.cross( vX );

		uvwBase.make( vX, vY, vZ );
		uvwBase.transpose();

		// transform to the base defined by the passed vector
		randVec = uvwBase * randVec;
		bVerbose && std::cerr << "Transformed sampleDir --> X: " << randVec[0] << " Y: " << randVec[1] << " Z: " << randVec[2] << std::endl << std::flush;
		bVerbose && std::cerr << "cos(N,R) = " << vZ.dot( randVec ) << std::endl << std::flush;

		if( bVerbose && vZ.dot( randVec ) < 0.0 ) {
			std::cerr << "ERROR: cos(N,R) < 0.0!!!" << std::endl << std::flush;
			exit( -1 );
		}

		randVec.normalize();      // safety catch
		return randVec;
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	static void uniformSampleTriangle( double& u, double& v, double rand1, double rand2 )

	// [7/8/2019 theis]
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		double su1 = sqrt( rand1 );
		u = 1.0 - su1;
		v = rand2 * su1;
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	static double Lerp( double t, double v1, double v2 )

	// [7/8/2019 theis]
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		return (1.0 - t) * v1 + t * v2;
	};

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	static Vector uniformSampleCone( double costhetamax, const Vector &x, const Vector &y, const Vector &z, double rand1 , double rand2 )

	// [7/8/2019 theis]
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		double costheta = CBaseSampler::Lerp( rand1, costhetamax, 1.0 );
		double sintheta = sqrt( 1.0 - costheta * costheta );
		double phi = rand2 * 2.0 * PI;

		Vector randVec( x * cos( phi ) * sintheta + y * sin( phi ) * sintheta + z * costheta);
		randVec.normalize();

		return randVec;
	}

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	static Vector uniformSampleCone( const Vector& Direction, double BaseRadius, double Rand1, double Rand2 )

	// sample directions on a cone oriented towards the passed Direction vector,
	// with a BaseRadius at unit height
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	{
		Vector DirVec = CBaseSampler::uniformSampleDisk( Direction, BaseRadius, Rand1, Rand2 );
		DirVec = Direction + DirVec;
		DirVec.normalize();

		return DirVec;
	}

};

#endif
//vim:ts=2,sw=2
