/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	6-Feb-2013
 */

#ifndef __MATH_SPLINE_H
#define __MATH_SPLINE_H

#include <vector>

#include "vector.h"

// ============================== BaseSplineNode ===============================
/** Abstract class of SplineNodes */
class BaseSplineNode {
public:
	bool	ignore;
public:
	BaseSplineNode() : ignore(false) {}
virtual ~BaseSplineNode() {}

	/* set/get */
virtual	int	size()			const	= 0;
virtual	double	operator()(int i)	const	= 0;
virtual	void	operator()(int i, double v)	= 0;
}; // BaseSplineNode

// ================================ SplineNode =================================
class SplineNode : public BaseSplineNode {
private:
	double V[2];

public:
	SplineNode(double ax=0.0, double ay=0.0) : BaseSplineNode() {V[0]=ax; V[1]=ay;}
virtual ~SplineNode() {}

virtual	int	size()			const	{ return 2; }
virtual	double	operator()(int i)	const	{ return V[i]; }
virtual	void	operator()(int i, double v)	{ V[i] = v; }

	double	x()	const { return V[0]; }
	double	y()	const { return V[1]; }
}; // SplineNode

// ============================= VectorSplineNode ==============================
class VectorSplineNode : public BaseSplineNode {
private:
	Vector	V;

public:
	VectorSplineNode() : BaseSplineNode() {}
	VectorSplineNode(const Vector& av) : V(av) {}
virtual ~VectorSplineNode() {}

virtual	int	size()			const	{ return 3; }
virtual	double	operator()(int i)	const	{ return V[i]; }
virtual	void	operator()(int i, double v)	{ V[i] = v; }

	operator Vector()		const	{ return V; }
	VectorSplineNode& operator=(const Vector& av)
						{ V = av; return *this; }
}; // VectorSplineNode

// ============================== CardinalSpline ===============================
/** Cardinal Spline class */
template <typename T>
class CardinalSpline {
private:
	double	_tension;
	double	matrix[4][4];
	std::vector<T>	nodes;

public:
	bool	closed;		// closed or open spline

public:
	CardinalSpline(const double a=0.5) : closed(false) { tension(a); }
	~CardinalSpline() {}

	void	add(T n)		{ nodes.push_back(n); }
	void	insert(int i, T n)	{ nodes.insert(nodes.begin()+i, n); }
	T&	node(int i)		{ return nodes[i]; }
const	T&	operator[](int i) const	{ return nodes[i]; }
	T&	operator[](int i)	{ return nodes[i]; }
	int	size() const		{ return (int)nodes.size(); }
	double	tension() const		{ return _tension; }
	void	del(int i)		{ nodes.erase(nodes.begin()+i); }

	/** initialize spline matrix with coefficient a */
	void	tension(const double t) {
			_tension = t;
			matrix[0][0] =    -t; matrix[0][1] = 2.0-t; matrix[0][2] = t-2.0;     matrix[0][3] =    t;
			matrix[1][0] = 2.0*t; matrix[1][1] = t-3.0; matrix[1][2] = 3.0-2.0*t; matrix[1][3] =   -t;
			matrix[2][0] =    -t; matrix[2][1] =   0.0; matrix[2][2] = t;         matrix[2][3] =  0.0;
			matrix[3][0] =   0.0; matrix[3][1] =   1.0; matrix[3][2] = 0.0;       matrix[3][3] =  0.0;
		}

	/** evaluate spline for interval k..k+1 with fraction t [0..1] */
	T	evaluate(int k, const double t) const {
			assert(size()>2);

			double TT[4], R[4];
			TT[0] = t*t*t;    TT[1] = t*t;    TT[2] = t;   TT[3] = 1.0;

			for (int i=0; i<4; i++) {
				R[i] = 0.0;
				for (int j=0; j<4; j++)
					R[i] += TT[j] * matrix[j][i];
			}

			int idx[4];			// four indices to use
			idx[0] = prev(k);
			idx[1] = k;
			// assert(not nodes[k].ignore);	// to avoid infinite loops
			idx[2] = next(k);
			idx[3] = next(idx[2]);

			T result;
			for (int j=0; j<result.size(); j++) {
				double y = 0.0;
				for (int i=0; i<4; i++)
					y += R[i] * nodes[idx[i]](j);
				result(j, y);
			}
			return result;
		}
	T	operator()(int k, const double t) const { return evaluate(k,t); }

private:
	/** @return previous non-ignore node */
	int	prev(int idx) const {
			do {
				if (--idx<0)
					return closed? size()-1 : 0;
			} while (nodes[idx].ignore);
			return idx;
		}

	/** @return next non-ignore node */
	int	next(int idx) const {
			do {
				if (++idx>=size())
					return closed? 0 : size()-1;
			} while (nodes[idx].ignore);
			return idx;
		}
}; // CardinalSpline

#endif
