/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Please look at the supplied documentation for the user's
 * license
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2010
 */

#include "pyutils.h"

static const char *trueStr[] = {
	"on", "ON", "On",
	"true", "TRUE", "True" };

/** Py_GetBool
 * Get boolean value from string or integer
 */
bool Py_GetBool(PyObject *obj)
{
	if (PyLong_Check(obj))
		return (bool)PyLong_AsLong(obj);
	else
	if (PyUnicode_Check(obj)) {
		const char* s = PyUnicode_AsUTF8(obj);
		for (unsigned i=0; i<sizeof(trueStr)/sizeof(char*); i++)
			if (!strcmp(s, trueStr[i]))
				return true;
	}
	return false;
} /* Py_GetBool */

/** Py_GetInt
 * @param obj	python object to return integer accepts float/int/string
 * @return	integer point number
 */
int Py_GetInt(PyObject *obj)
{
	if (PyFloat_Check(obj))
		return (int)PyFloat_AsDouble(obj);
	else
	if (PyLong_Check(obj))
		return PyLong_AsLong(obj);
	else
		return atoi(PyUnicode_AsUTF8(obj));
} /* Py_GetInt */

/** Py_GetFloat
 * @param obj	python object to return float accepts float/int/string
 * @return	floating point number
 */
double Py_GetFloat(PyObject *obj)
{
	if (PyFloat_Check(obj))
		return PyFloat_AsDouble(obj);
	else
	if (PyLong_Check(obj))
		return (double)PyLong_AsLong(obj);
	else
		return atof(PyUnicode_AsUTF8(obj));
} /* Py_GetFloat */

/** Py_GetUV
 * @param obj	python object tuple/list of size=2
 * @param u,v	return variables
 * @return	false on error
 */
bool Py_GetUV(PyObject *obj, double *u, double *v)
{
	if (PyTuple_Check(obj) && PyTuple_GET_SIZE(obj)==2) {
		*u = PyFloat_AsDouble(PyTuple_GetItem(obj,0));
		*v = PyFloat_AsDouble(PyTuple_GetItem(obj,1));
		return PyErr_Occurred()==NULL;
	} else
	if (PyList_Check(obj) && PyList_GET_SIZE(obj)==2) {
		*u = PyFloat_AsDouble(PyList_GetItem(obj,0));
		*v = PyFloat_AsDouble(PyList_GetItem(obj,1));
		return PyErr_Occurred()==NULL;
	} else
		PyErr_SetString(PyExc_TypeError, "UV: tuple or list of size 2 expected");
	return true;
} /* Py_GetUV */

/** @return true if obj is a vector (tuple or list) with size 3 and
 * having 3 floating point numbers
 */
bool Py_IsVector(PyObject* obj)
{
	if (PyTuple_Check(obj) && PyTuple_GET_SIZE(obj)==3) {
		for (int i=0; i<3; i++)
			if (!PyFloat_Check(PyTuple_GetItem(obj,i)))
				return false;
		return true;
	} else
	if (PyList_Check(obj) && PyList_GET_SIZE(obj)==3) {
		for (int i=0; i<3; i++)
			if (!PyFloat_Check(PyList_GetItem(obj,i)))
				return false;
		return true;
	}
	return false;
} // Py_IsVector

/** Py_GetVector
 * @param obj	python object tuple/list of size=3
 * @return	vector
 */
Vector Py_GetVector(PyObject *obj)
{
	Vector v;
	if (PyTuple_Check(obj) && PyTuple_GET_SIZE(obj)==3)
		v.set( PyFloat_AsDouble(PyTuple_GetItem(obj,0)),
			PyFloat_AsDouble(PyTuple_GetItem(obj,1)),
			PyFloat_AsDouble(PyTuple_GetItem(obj,2)));
	else
	if (PyList_Check(obj) && PyList_GET_SIZE(obj)==3)
		v.set( PyFloat_AsDouble(PyList_GetItem(obj,0)),
			PyFloat_AsDouble(PyList_GetItem(obj,1)),
			PyFloat_AsDouble(PyList_GetItem(obj,2)));
	else
		PyErr_SetString(PyExc_TypeError, "Vector: tuple or list of size 3 expected");
	return v;
} /* Py_GetVector */

/** Py_GetPoint
 * @param obj	python object tuple/list of size=3
 * @return	point
 */
Point Py_GetPoint(PyObject *obj)
{
	Point p;
	if (PyTuple_Check(obj) && PyTuple_GET_SIZE(obj)==3)
		p.set( PyFloat_AsDouble(PyTuple_GetItem(obj,0)),
			PyFloat_AsDouble(PyTuple_GetItem(obj,1)),
			PyFloat_AsDouble(PyTuple_GetItem(obj,2)));
	else
	if (PyList_Check(obj) && PyList_GET_SIZE(obj)==3)
		p.set( PyFloat_AsDouble(PyList_GetItem(obj,0)),
			PyFloat_AsDouble(PyList_GetItem(obj,1)),
			PyFloat_AsDouble(PyList_GetItem(obj,2)));
	else
		PyErr_SetString(PyExc_TypeError, "Point: tuple or list of size 3 expected");
	return p;
} /* Py_GetPoint */

/* --- PyList_AsMatrix3 --- */
bool PyList_AsMatrix3(PyObject *obj, Matrix3& matrix)
{
	if (!PyList_Check(obj) || PyList_GET_SIZE(obj)!=3) goto ERROR;
	for (int j=0; j<3; j++) {
		PyObject *row = PyList_GetItem(obj, j);
		if (!PyList_Check(row) || PyList_GET_SIZE(row)!=3) goto ERROR;
		for (int i=0; i<3; i++)
			matrix(j,i) = PyFloat_AsDouble(PyList_GetItem(row,i));
	}
	return true;
ERROR:
	PyErr_SetString(PyExc_TypeError,
			"Invalid Matrix3 list of lists [3x3] expected");
	return false;
} /* PyList_AsMatrix3 */

/* --- PyList_AsMatrix4 --- */
bool PyList_AsMatrix4(PyObject *obj, Matrix4& matrix)
{
	if (!PyList_Check(obj) || PyList_GET_SIZE(obj)!=4) goto ERROR;
	for (int j=0; j<4; j++) {
		PyObject *row = PyList_GetItem(obj, j);
		if (!PyList_Check(row) || PyList_GET_SIZE(row)!=4) goto ERROR;
		for (int i=0; i<4; i++)
			matrix(j,i) = PyFloat_AsDouble(PyList_GetItem(row,i));
	}
	return true;
ERROR:
	PyErr_SetString(PyExc_TypeError,
			"Invalid Matrix4 list of lists [4x4] expected");
	return false;
} /* PyList_AsMatrix4 */

/* --- PyList_FromMatrix3 from a 3x3 matrix --- */
PyObject *PyList_FromMatrix3(const Matrix3& matrix)
{
	PyObject *obj = PyList_New(3);
	for (int j=0; j<3; j++) {
		PyObject *row = Py_BuildValue("[ddd]",
						matrix(j,0),
						matrix(j,1),
						matrix(j,2));
		PyList_SET_ITEM(obj, j, row);
	}
	return obj;
} /* PyList_FromMatrix3 */

/* --- PyList_FromMatrix3 from a 4x4 matrix --- */
PyObject *PyList_FromMatrix3(const Matrix4& matrix)
{
	PyObject *obj = PyList_New(3);
	for (int j=0; j<3; j++) {
		PyObject *row = Py_BuildValue("[ddd]",
						matrix(j,0),
						matrix(j,1),
						matrix(j,2));
		PyList_SET_ITEM(obj, j, row);
	}
	return obj;
} /* PyList_FromMatrix3 */

/* --- PyList_FromMatrix4 --- */
PyObject *PyList_FromMatrix4(const Matrix4& matrix)
{
	PyObject *obj = PyList_New(4);
	for (int j=0; j<4; j++) {
		PyObject *row = Py_BuildValue("[dddd]",
						matrix(j,0),
						matrix(j,1),
						matrix(j,2),
						matrix(j,3));
		PyList_SET_ITEM(obj, j, row);
	}
	return obj;
} /* PyList_FromMatrix4 */

/* Py_Check4Pattern
 * @param obj	object to check if it holds a string pattern
 * @return true if it has punctuation characters inside
 */
bool Py_Check4Pattern(PyObject *obj)
{
	if (!PyUnicode_Check(obj)) return false;
	ssize_t len;
	// fixme can search directly in unicode
	const char* str = PyUnicode_AsUTF8AndSize(obj, &len);
	if (memchr(str, '*', len)) return true;
	if (memchr(str, '?', len)) return true;
	if (memchr(str, '[', len)) return true;

	return false;
} /* Py_Check4Pattern */
