/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	04-Feb-2002
 */

#ifndef __TOOL_TIMER_H
#define __TOOL_TIMER_H

#include <time.h>
#include <iostream>

#include "os.h"

/* extern functions */
double btime();

// =================================== Timer ===================================
/** A chronometer class for both CPU and real time */
class Timer {
protected:
	bool	active;
	int	_laps;
	double	startRealTime,
		stopRealTime;
	double	startUserTime,
		stopUserTime;
public:
	Timer(bool on=false) : active(false), _laps(0),
			startRealTime(0.0),
			stopRealTime(0.0),
			startUserTime(0.0),
			stopUserTime(0.0)
		{ if (on) start(); }
	~Timer() {}

	/** reset timer */
	void	reset()	{ active = false; }

	/** start measuring */
	void	start();

	/** restart timer */
	void	restart()	{ start(); }

	/** mark a lap */
	void	lap();

	/** stop timer */
	void	stop();

	/** resume timer */
	void	resume();

	/** @return true if timer is active */
	bool	working()	const	{ return active; }

	/** @return number of laps */
	int	laps()		const	{ return _laps; }

	/** @return real time in seconds */
	double	elapsed()	{ lap(); return realTime(); }

	/** @return real time in seconds */
	double	realTime() const
		{ return (stopRealTime - startRealTime); }

	/** @return CPU time in seconds */
	double	userTime() const
		{ return (stopUserTime - startUserTime); }

	/** @return average real time per lap */
	double	averageLapRealTime()	{ return realTime() / (double)_laps; }

	/** @return average user time per lap */
	double	averageLapUserTime()	{ return userTime() / (double)_laps; }

	/** wait until real time t seconds is passed
	 * @param t time to wait
	 */
	void	delayRealTime(double t);

	/** wait until CPU time t seconds is passed
	 * @param t time to wait
	 */
	void	delayUserTime(double t);
}; // Timer

/** CPU process time with nsec resolution */
struct CPUTime : public timespec {
private:
const	long	BILLION = 1000000000;
public:
	CPUTime() : timespec{0,0} {}
	CPUTime(bool)	{ get(); }
	CPUTime(const CPUTime& t) : timespec{t.tv_sec, t.tv_nsec} {}
	CPUTime& operator =(const CPUTime& t) {
			tv_sec  = t.tv_sec;
			tv_nsec = t.tv_nsec;
			return *this;
		}

	void	get()	{ clock_gettime(CLOCK_PROCESS_CPUTIME_ID, this); }
	void	reset()	{ tv_sec = 0; tv_nsec = 0; }

	/** subtract another CPUTime */
	void	diff(const CPUTime& t) {
			if (tv_nsec < t.tv_nsec) {
				tv_sec  -= t.tv_sec+1;
				tv_nsec += BILLION;
				tv_nsec -= t.tv_nsec;
			} else {
				tv_sec  -= t.tv_sec;
				tv_nsec -= t.tv_nsec;
			}
		}
	CPUTime& operator -=(const CPUTime& t) { diff(t); return *this; }

	/** add another CPUTime */
	void	add(const CPUTime& t) {
			tv_sec  += t.tv_sec;
			tv_nsec += t.tv_nsec;
			if (tv_nsec > BILLION) {
				tv_sec++;
				tv_nsec -= BILLION;
			}
		}
	CPUTime& operator +=(const CPUTime& t) { add(t); return *this; }
}; // CPUTime

inline CPUTime operator +(const CPUTime& A, const CPUTime& B) {
		CPUTime res(A);
		res += B;
		return res;
	}

inline CPUTime operator -(const CPUTime& A, const CPUTime& B) {
		CPUTime res(A);
		res -= B;
		return res;
	}

/** Output to a stream the timer */
inline std::ostream& operator << (std::ostream& os, Timer& tm)
{
	if (tm.working()) tm.elapsed();
	os << "real: " << tm.realTime();
	os << "  user: " << tm.userTime() << " s";
	return os;
}

#endif
