/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Please look at the supplied documentation for the user's
 * license
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	01-Apr-2010
 */

#include <Python.h>	// it has to be the first include!

#include <time.h>
#include <math.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <fnmatch.h>
#include <X11/Xutil.h>

#include <iomanip>

#ifdef OPENGL
#define GLX_GLXEXT_PROTOTYPES
#	include <GL/gl.h>
#	include <GL/glx.h>
#	include <GL/glu.h>
#endif

#include <string>
#include <fstream>

#include "os.h"
#include "gobject.h"
#include "pyutils.h"
#include "tool/array.h"
#include "tool/timer.h"
#include "viewerobject.h"
#include "geometry/geo.h"
#include "geometry/gzone.h"
#include "geometry/vzone.h"
#include "geometry/planebody.h"

#ifdef OPENGL
#	include "gldraw/glmesh.h"
#endif

using std::cout;
using std::endl;

#define VPMARGIN	1	// Margin from the border of viewport lines

static const char dashedPattern[2] = {5,5};

#ifdef OPENGL
static int glBuffer[] = {
		GLX_RGBA,
		GLX_RED_SIZE,    8,
		GLX_GREEN_SIZE,  8,
		GLX_BLUE_SIZE,   8,
		GLX_DEPTH_SIZE, 12,
//		GLX_DOUBLEBUFFER,
		None };
#endif

/* --- Clear viewport errors --- */
static void _clearErrors(ViewerObject* self)
{
	if (self->errors.empty()) return;
	for (auto object : self->errors)
		delete object;
	self->errors.clear();
} // _clearErrors

/** sendMessage: send message to tcl interpreter */
static void _sendMessage(ViewerObject* self, const char* msg)
{
	//if (!self->threaded) return;
	if (!Tk_IsMapped(self->tkWin)) return;
	union {XEvent gen; XVirtualEvent virt;} event;
	memset(&event, 0, sizeof(event));
	event.gen.xany.type       = VirtualEvent;
	event.gen.xany.serial     = NextRequest(self->display);
	event.gen.xany.send_event = False;
	event.gen.xany.window     = self->window;
	event.gen.xany.display    = self->display;
	event.virt.name           = Tk_GetUid(msg);
	Tk_QueueWindowEvent(&event.gen, TCL_QUEUE_TAIL);
	//TkWindowEvent* wev = (TkWindowEvent*)ckalloc(sizeof(TkWindowEvent));
	//wev->header.proc = WindowEventProc;
	//wev->event = event;
	//Tcl_ThreadQueueEvent(self->threadId, &wev->header, TCL_QUEUE_TAIL);
} // _sendMessage

/* --- loadFont, tga or system --- */
static bool loadFont(BFont& font, const char* fontpath)
{
	if (font.load(fontpath)) return true;
	PyErr_Format(PyExc_TypeError, "Invalid font \'%s\' specified", fontpath);
	return false;

#if 0
	// Cause a seg-fault after few font reloads with bad malloc!

	int screen = DefaultScreen(display);

	// Load bitmap system font
	XFontStruct* fontinfo = XLoadQueryFont(display, fontpath);
	if (fontinfo == NULL) {
		PyErr_Format(PyExc_TypeError,
			"Invalid system font \'%s\' specified", fontpath);
		return false;
	}
	int width  = fontinfo->max_bounds.width;
	int height = fontinfo->max_bounds.ascent + fontinfo->max_bounds.descent;

	Pixmap pixmap = XCreatePixmap(display,
				XRootWindow(display,screen),
				16*width, 16*height,
				DefaultDepth(display,screen));

	XGCValues values;
	values.font       = fontinfo->fid;
	values.function   = GXcopy;
	values.plane_mask = AllPlanes;
	values.foreground = BlackPixel(display,screen);
	values.background = BlackPixel(display,screen);

	GC gc = XCreateGC(display,
		pixmap,
		GCFont |
		GCFunction | GCPlaneMask | GCForeground | GCBackground,
		&values);

	XFillRectangle(display,pixmap,gc,0,0,16*width,16*height);
	XSetForeground(display, gc, WhitePixel(display,screen));

	for (int row=0; row<16; row++) {
		int y = row*height;
		for (int col=0; col<16; col++) {
			int x = col*width;
			char str[2];
			str[0] = row<<4 | col;
			str[1] = 0;
			XDrawImageString(display,pixmap,gc,x,y,str,1);
		}
	}
	XFreeFont(display, fontinfo);

	XImage* image = XGetImage(display, pixmap,
		0, 0, 16*width, 16*height, AllPlanes,ZPixmap);

	if (image) {
		font.setMap(fontpath, 16*width, 16*height, (dword*)image->data);
		XDestroyImage(image);
	} else {
		PyErr_Format(PyExc_TypeError,
			"Invalid system font \'%s\' specified", fontpath);
		return false;
	}
	XFreeGC(display, gc);
	XFreePixmap(display, pixmap);
	return true;
#endif
} // loadFont

/* Py_VBody
 * @param self	Viewer object
 * @param obj	object to convert to body can be string or id
 * @return body
 */
static VBody* Py_VBody(ViewerObject* self, PyObject* obj)
{
	VBody* body;
	if (obj==NULL) return NULL;
	if (PyUnicode_Check(obj)) {
		const char* name = PyUnicode_AsUTF8(obj);
		body = self->kernel->getBody(name);
		if (body==NULL) {
			PyErr_Format(PyExc_KeyError, "Body \'%s\' not found", name);
			return NULL;
		}
		return body;
	} else
	if (PyLong_Check(obj)) {
		int id = PyLong_AsLong(obj);
		if (id>=0 && id<(int)self->kernel->bodies.size()) {
			body = self->kernel->getBody(id);
			if (body==NULL) {
				PyErr_Format(PyExc_IndexError, "Body #%d not found",id);
				return NULL;
			}
			return body;
		}
		PyErr_Format(PyExc_IndexError, "Body #%d not found",id);
		return NULL;
	} else {
		PyErr_SetString(PyExc_TypeError, "Invalid body type, string or integer expected");
		return NULL;
	}
} // Py_VBody

/* Py_VRegion
 * @param self	Viewer object
 * @param obj	object to convert to region can be string or id
 * @return region
 */
static VRegion* Py_VRegion(ViewerObject* self, PyObject* obj)
{
	VRegion* region;
	if (obj==NULL) return NULL;
	if (PyUnicode_Check(obj)) {
		const char* name = PyUnicode_AsUTF8(obj);
		// FIXME: potential source of troubles
		region = self->kernel->getRegion(name);
		if (region==NULL) {
			PyErr_Format(PyExc_KeyError, "Region \'%s\' not found", name);
			return NULL;
		}
		return region;
	} else
	if (PyLong_Check(obj)) {
		int id = PyLong_AsLong(obj);
		// FIXME: potential source of troubles
		region = self->kernel->getRegion(id);
		if (region==NULL) {
			PyErr_Format(PyExc_IndexError, "Region #%d not found",id);
			return NULL;
		}
		return region;
	} else {
		PyErr_SetString(PyExc_TypeError, "Invalid object type, string or integer expected");
		return NULL;
	}
} // Py_VRegion

// ================================== Viewer ===================================
// Static prototypes
static void Viewer_calcViewport(ViewerObject* self, int id);

/* --- Viewer_new --- */
static PyObject* Viewer_new(PyTypeObject* type, PyObject* /*args*/,
				PyObject* /*kwds*/)
{
	ViewerObject* self;

	self = (ViewerObject*)type->tp_alloc(type, 0);
	if (self == NULL) return NULL;

	self->display   = NULL;
	self->tkWin     = NULL;
	self->window    = 0;
	self->depth     = 0;
	self->fixedSize = false;
	self->gc        = NULL;
	self->ximage    = NULL;
	self->pixmap    = 0;
#ifdef OPENGL
	self->visinfo   = NULL;
#endif
	self->viewer    = NULL;
	self->geometry  = NULL;
	self->ocube     = NULL;
	self->showViewport  = true;
	self->borderWidth   = 0;
	self->borderColor   = 0x000000;
	self->crosshair     = 0;
	self->showTrackball = false;
	self->projectionChanged = true;

	for (int i=0; i<ViewerObject::nviews; i++) {
		self->viewport[i].viewer      = NULL;
		self->viewport[i].visible     = false;
		self->viewport[i].lineWidth   = 1;
		self->viewport[i].originWidth = 2;
		self->viewport[i].wLength     = 15.0;
		self->viewport[i].color       = 0xFFFF00;
	}

	return (PyObject*)self;
} // Viewer_new

/* --- Viewer_init --- */
static int Viewer_init(ViewerObject* self, PyObject* args, PyObject* /*kwds*/)
{
	PyObject* interpaddr;
	char  *winname;
	GeometryObject* geometry;

	if (!PyArg_ParseTuple(args, "OsO", &geometry, &winname, &interpaddr)) return -1;
	if (Py_TYPE(geometry) != &GeometryType) {
		PyErr_SetString(PyExc_TypeError, "Invalid type, Geometry type expected");
		return -1;
	}

	self->geometry = geometry;
	self->kernel = new GeometryKernel(*geometry->geometry);
	self->viewer = new GeometryViewer(*geometry->geometry, *self->kernel);
	self->ocube  = new XOrientationCube;

	Tcl_Interp* interp = (Tcl_Interp*)PyLong_AsVoidPtr(interpaddr);

	// Find if tcl is compiled with threads enabled
	//if (Tcl_Eval(interp, "::tcl::pkgconfig get threaded")==0)
	//	self->threaded = !strcmp(Tcl_GetStringResult(interp), "1");
	//else
	//	self->threaded = false;
	//self->threadId = Tcl_GetCurrentThread();

	// Find Window
	self->tkWin = Tk_NameToWindow(interp, winname, Tk_MainWindow(interp));
	if (!self->tkWin) {
		PyErr_SetString(PyExc_ValueError, Tcl_GetStringResult(interp));
		return -1;
	}

	// setup display
	if (self->display == NULL) {
		self->display = Tk_Display(self->tkWin);
		self->xdraw.display = self->display;
		self->depth   = DefaultDepth(self->display,
				DefaultScreen(self->display));
	}

	// WARNING??? window at this location is not set correctly!!!!
	assert(self->gc==NULL);

	assert(self->ximage==NULL);
	self->ximage = XCreateImage(self->display,
				DefaultVisual(self->display,
				DefaultScreen(self->display)),
				self->depth, ZPixmap, 0,
				(char*)(self->viewer->painter.data()),
				self->viewer->width(),
				self->viewer->height(),
				32, 0);

	// Initialize variables
	self->startU = self->startV = 0.0;
	self->pivotU = self->pivotV = 0.0;
	self->pivotAngle = INFINITE;	// a large number not to show rotation

	self->rectX1 = -1;	// Don't draw
	self->rectY1 = -1;
	self->rectX2 = -1;
	self->rectY2 = -1;
	self->ocube->init();
	return 0;
} // Viewer_init

/* --- Viewer_dealloc --- */
static void Viewer_dealloc(ViewerObject* self)
{
	if (self->viewer) {
		self->viewer->stopThread();
		// will be deleted from XDestroyImage
		self->viewer->painter.dataNull();
		delete self->viewer;
		delete self->kernel;
		delete self->ocube;
	}
	_clearErrors(self);
#ifdef OPENGL
	glXDestroyGLXPixmap(self->display, self->glxpixmap);
	glXDestroyContext(self->display, self->context);
#endif
	XFreePixmap(self->display, self->pixmap);
	if (self->ximage) XDestroyImage(self->ximage);
	if (self->gc) XFreeGC(self->display, self->gc);

	Py_TYPE(self)->tp_free((PyObject*)self);
} // Viewer_dealloc

#ifdef MEM
/* --- Viewer_destroy --- */
static PyObject* Viewer_destroy(ViewerObject* self, PyObject* Py_UNUSED(ignored))
{
	// unfortunately we cannot call _dealloc due to all X, Tk deallocations
	delete self->viewer;
	delete self->kernel;
	_clearErrors(self);
	Py_RETURN_NONE;
} // Geometry_destroy
#endif

/* --- Viewer_configure --- */
static PyObject* Viewer_configure(ViewerObject* self, PyObject* args)
{
	int width, height;
	if (!PyArg_ParseTuple(args, "ii", &width, &height)) return NULL;
	if (width!=self->viewer->width() || height!=self->viewer->height()) {
		if (self->pixmap) {
			XFreePixmap(self->display, self->pixmap);
			self->pixmap = XCreatePixmap(self->display, self->window,
							width, height, self->depth);
#ifdef OPENGL
			glXDestroyGLXPixmap(self->display, self->glxpixmap);
			self->glxpixmap = glXCreateGLXPixmap(self->display,
							self->visinfo,
							self->pixmap);
#endif
		}

		// prepare XImage
		self->ximage->width  = width;
		self->ximage->height = height;
		self->ximage->bytes_per_line = 0;

		if (!self->fixedSize) {
			self->viewer->resize(width, height);
			self->ximage->data = (char*)self->viewer->painter.data();
		} else {
			self->imagePainter.init(width, height);
			self->ximage->data = (char*)self->imagePainter.data();
		}
		XInitImage(self->ximage);
	}
	Py_RETURN_NONE;
} // Viewer_configure

/* --- Viewer_expose ---
 * X11 main drawing routine.
 * Combines in a temporary pixmap
 * 1. the pixel painter that has to be already present using the GeometryViewer::draw() method
 * 2. superimpose x11 objects
 */
static PyObject* Viewer_expose(ViewerObject* self, PyObject* /*args*/)
{
	XGCValues gcValues;

	int width   = Tk_Width(self->tkWin);
	int height  = Tk_Height(self->tkWin);
	int width2  = width / 2;
	int height2 = height / 2;

	// Initialize the first time
	if (self->gc == NULL) {
		if (!Tk_IsMapped(self->tkWin)) {
			Tk_MakeWindowExist(self->tkWin);
			if (!Tk_IsMapped(self->tkWin))
				Py_RETURN_NONE;
		}
		self->window = Tk_WindowId(self->tkWin);

		// Create initial pixmap
		self->pixmap = XCreatePixmap(self->display, self->window, width, height, self->depth);

		//
		gcValues.function = GXcopy;
		gcValues.plane_mask = AllPlanes;
		//self->gc =  Tk_GetGC(self->tkWin, GCFunction|GCGraphicsExposures, &gcValues);
		self->gc = XCreateGC(self->display, self->window,
					GCFunction | GCPlaneMask | GCForeground | GCBackground,
					&gcValues);
		self->xdraw.gc = self->gc;
		self->xdraw.init(self->pixmap);	// initialise pixmaps

#if OPENGL
		self->visinfo = glXChooseVisual(self->display, 0 /*self->screen*/, glBuffer);
		// Last argument is Direct rendering through False=through X11
		// Newer X-servers disable the indirect rendering!
		self->context   = glXCreateContext(self->display, self->visinfo, None, True);
		self->glxpixmap = glXCreateGLXPixmap(self->display, self->visinfo, self->pixmap);
#endif
	}

#if OPENGL
	// -------- Initialize OpenGL pixmap --------
	glXMakeCurrent(self->display, self->glxpixmap, self->context);
//	printf("GL_RENDERER = %s\n", (char *) glGetString(GL_RENDERER));

	glViewport(0, 0, width, height);
	//glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glClear(GL_DEPTH_BUFFER_BIT);
#endif

	if (self->fixedSize) { //or self->ximage->width!=width or self->ximage->height==height)
		// rescale image
		self->imagePainter.fit(self->viewer->painter);
	}
	XPutImage(self->display, self->pixmap, self->gc, self->ximage,
			0, 0, 0, 0, width, height);
#if OPENGL
	// XSync to transfer pixmap to ximage
	XSync(self->display, False);

	// ---------- Do OpenGL stuff ---
//	glEnable(GL_DEPTH_TEST);
//	glShadeModel(GL_SMOOTH);
//	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
//	glEnable(GL_POINT_SMOOTH);
//	glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
//	glEnable(GL_BLEND);
//	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	/* light */
	static GLfloat light_pos[4] = {5.0, 20.0, 10.0, 1.0 };
	glEnable(GL_LIGHTING);
	glLightfv(GL_LIGHT0, GL_POSITION, light_pos);
	glEnable(GL_LIGHT0);
	glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);

//	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	/* Setup projection */
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	float dw = self->kernel->view.imageWidth();
	float dh = self->kernel->view.imageHeight();
	glOrtho(-dw/2., dw/2., -dh/2, dh/2, -1000.f, 1000.f);

	{
		double matrix[4][4];
		memset(matrix, 0, sizeof(matrix));
		for (int j=0; j<3; j++)
			for (int i=0; i<3; i++)
				matrix[j][i] = self->kernel->view.matrix(j,i);
		matrix[3][3] = 1.0;
		glMultMatrixd((GLdouble*)&matrix);

		double x,y,z;
		self->kernel->view.origin(&x, &y, &z);
		glTranslated(-x,-y,-z);
	}

	/* Setup geometry */
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// Scan geometry and plot all meshes
	GLMesh::drawFaces = true;
	for (auto body : self->geometry->geometry->bodies) {
		GLMesh glmesh(body->mesh());
		glmesh.draw(GL_RENDER);
	}

	glFlush();	// flush contents. Doesn't ensure that the graph is completed
//	glFinish();	// wait to finish and then flush
//	glXWaitGL();	// wait X11 to finish
//	glXSwapBuffers(self->display, self->glxpixmap);
#endif

	// ---------- Draw objects ----------
	if (self->geometry->objects.size()) {
		// Viewers expose are called in sync from the main thread
		// so it is safe to transform and later draw
		for (auto object : self->geometry->objects) {
			if (!object->show) continue;
			object->transform(self);
			if (object->visible())
				object->draw(self, self->pixmap);
		}
	}

	// ---------- Draw errors ----------
	if (self->errors.size() && self->viewer->d2.show) {
		for (auto object : self->errors) {
			object->transform(self);
			if (object->visible())
				object->draw(self, self->pixmap);
		}
	}

	// ---------- Show crosshair (could be an object) FIXME XXX ----------
	if (self->crosshair) {
		gcValues.function   = GXxor;
		gcValues.line_width = 0;
		gcValues.line_style = LineSolid;
		self->xdraw.changeGC(GCFunction|GCLineWidth|GCLineStyle, &gcValues);
		self->xdraw.foreground(0xFFFFFF);

		self->xdraw.line(self->pixmap,
			width2-self->crosshair, height2, width2+self->crosshair+1, height2);
		self->xdraw.line(self->pixmap,
			width2, height2-self->crosshair, width2, height2+self->crosshair+1);
	}

	// ---------- Show trackball ----------
	if (self->showTrackball) {
		int radius = Round(Min(width, height) * self->geometry->trackballSize)/2;
		gcValues.function   = GXcopy;
		gcValues.line_width = 2;
		gcValues.line_style = LineSolid;
		self->xdraw.changeGC(GCFunction|GCLineWidth|GCLineStyle, &gcValues);
		self->xdraw.trackball(self->pixmap,
					width2, height2, radius,
					self->kernel->view.matrix());
	}

	// ---------- Orientation cube ----------
	if (self->ocube->length>0) {
		int lenscaled = self->ocube->length*2 + (self->ocube->length>>1);	// 2.5*length
		// Top-Right corner
#ifdef OCUBE_TOPRIGHT
		int x = width - lenscaled;
		int y = lenscaled;
#else
		// Bottom-Left corner
		int x = lenscaled;
		int y = height-lenscaled;
#endif
		self->ocube->draw(self->xdraw, self->pixmap,
				x, y, self->kernel->view.matrix());
	}

	// ---------- Show editing body ----------
	const GBody* editBody = self->geometry->geometry->editBody;
	if (self->geometry->cursorShow) {
		int x,y;
		self->kernel->view.xyz2ij3D(self->geometry->cursor, &x, &y);
		int m = self->geometry->cursorMoveSize;
		int r = self->geometry->cursorSize;
		if (x>=-r && x<width+r && y>=-r && y<height+r) {
			// Draw the handlers
			gcValues.function   = GXxor;
			gcValues.line_width = 2;
			gcValues.line_style = LineSolid;
			self->xdraw.changeGC(GCFunction|GCLineWidth|GCLineStyle, &gcValues);

			self->xdraw.foreground(0xFFFFFF);
			self->xdraw.arc(self->pixmap, x-m, y-m, 2*m, 2*m, 0, 360*64);

			self->xdraw.foreground(0xAF00AF);
			self->xdraw.arc(self->pixmap, x-r, y-r, 2*r, 2*r, 0, 360*64);

			gcValues.function   = GXcopy;
			self->xdraw.changeGC(GCFunction, &gcValues);
			self->xdraw.axes(self->pixmap, x, y, r, self->kernel->view.matrix());
		}
	}

	if (editBody && self->kernel->getBody(editBody)->visible) {
		// draw editing rulers if any
		if (editBody->show & BIT_SELECT && self->geometry->rulers) {
			// FIXME set position???? should be done when editing the body!
			for (int i=0; i<self->geometry->rulers; i++) {
				GRuler* ruler = self->geometry->ruler[i];
				ruler->position(editBody->position());
				const char *name;
				switch (i) {
					case 0:
						name = editBody->showX();
						if (!name) continue;
						ruler->name(name);
						ruler->direction(editBody->vectorXlen() * editBody->vectorX());
						break;
					case 1:
						name = editBody->showY();
						if (!name) continue;
						ruler->name(name);
						ruler->direction(editBody->vectorYlen() * editBody->vectorY());
						break;
					case 2:
						name = editBody->showZ();
						if (!name) continue;
						ruler->name(name);
						ruler->direction(editBody->vectorZlen() * editBody->vectorZ());
						break;
					default:
						continue;
				}
				ruler->transform(self);
				if (ruler->visible())
					ruler->draw(self, self->pixmap);
			}
		}
	}

	// ---------- Show viewports ----------
	if (self->showViewport) {
		self->xdraw.setDashes(0, dashedPattern, SIZE(dashedPattern));

		for (int i=0; i<ViewerObject::nviews; i++) {
			ViewportLine& vp = self->viewport[i];
			Viewer_calcViewport(self, i);

			gcValues.function   = GXcopy;
			gcValues.line_width = vp.lineWidth;
			gcValues.line_style = LineOnOffDash;
			self->xdraw.changeGC(GCFunction|GCLineStyle|GCLineWidth, &gcValues);

			if (vp.visible) {
				self->xdraw.foreground(vp.color);
				// Draw line
				self->xdraw.line(self->pixmap,
					vp.x[0], vp.y[0], vp.x[1], vp.y[1]);

				gcValues.line_style = LineSolid;
				self->xdraw.changeGC(GCLineStyle, &gcValues);

				// Draw Cursor
				self->xdraw.rectangle(self->pixmap,
					vp.xc-vp.originWidth, vp.yc-vp.originWidth,
					2*vp.originWidth, 2*vp.originWidth);

				// Draw w-vector
				self->xdraw.line(self->pixmap,
					vp.xc, vp.yc, vp.xw, vp.yw);
			}
		}
	}

	// ---------- Show selection rectangle ----------
	if (self->rectX1 >= 0) {
		gcValues.function   = GXxor;
		gcValues.line_width = 0;
		gcValues.line_style = LineSolid;
		self->xdraw.changeGC(GCFunction | GCLineWidth | GCLineStyle, &gcValues);
		self->xdraw.foreground(0xFFFFFF);

		int x = Min(self->rectX1, self->rectX2);
		int y = Min(self->rectY1, self->rectY2);
		int w = Abs(self->rectX2 - self->rectX1);
		int h = Abs(self->rectY2 - self->rectY1);
		self->xdraw.rectangle(self->pixmap,
			x, y, w, h);
	}

	// ---------- Show hand drawing ----------
	if (self->pen.size() > 0) {
		gcValues.function   = GXxor;
		gcValues.line_width = 2;
		gcValues.line_style = LineSolid;
		self->xdraw.changeGC(GCFunction | GCLineWidth | GCLineStyle, &gcValues);
		self->xdraw.foreground(0xFFFFFF);

		for (size_t i=0; i<self->pen.size()-1; i++) {
			IPoint& a = self->pen[i];
			IPoint& b = self->pen[i+1];
			self->xdraw.line(self->pixmap, a.x, a.y, b.x, b.y);
		}
	}

	// ---------- Draw border ----------
	if (self->borderWidth > 0) {
		self->xdraw.setFunction(GXcopy);
		self->xdraw.foreground(self->borderColor);
		const int bw = self->borderWidth;
		self->xdraw.fillRectangle(self->pixmap, 0, 0, bw, height);		// left
		self->xdraw.fillRectangle(self->pixmap, width-bw, 0, bw, height);	// right
		self->xdraw.fillRectangle(self->pixmap, 0, 0, width, bw);		// top
		self->xdraw.fillRectangle(self->pixmap, 0, height-bw, width, bw);	// bottom
	}

	// ---------- Display the self->pixmap ----------
	gcValues.function = GXcopy;
	self->xdraw.changeGC(GCFunction, &gcValues);
	XCopyArea(self->display, self->pixmap, self->window, self->gc,
		0, 0, width, height, 0, 0);

	Py_RETURN_NONE;
} // Viewer_expose

/* --- Viewer_export --- */
static PyObject* Viewer_export(ViewerObject* self, PyObject* args)
{
	char* filename;

	if (!PyArg_ParseTuple(args, "s", &filename)) return NULL;
	self->viewer->exporter(filename);
	Py_RETURN_NONE;
} /* Viewer_export*/

/* _snapUV - snap the (u,v) position with d as snapping distance
 * @param u,v	position on the viewer plane
 * @param d	snapping distance
 * @return	Point(x,y,z) with absolute coordinates of the snapped position
 */
static Point _snapUV(ViewerObject* self, double u, double v, double d)
{
	double ds = d / self->kernel->view.Sx();

	// 1st Scan near-by objects
	for (auto object : self->geometry->objects) {
		if (!object->show) continue;
		if (object->selectNode>0) continue;
		object->transform(self);
		if (!object->visible()) continue;

		// check the w?
		Point O = object->viewerPosition();

		if (Abs(u-O.x)<=ds && Abs(v-O.y)<=ds)
			// Use only the ou,ov coordinates of object!
			return self->kernel->view.uv2xyz(O.x, O.y);

		// Check END of GArrow and GRuler
		if (object->isA()==GArrowClass || object->isA()==GRulerClass) {
			double um = O.x;	// Remember for mid position checking
			double vm = O.y;

			// check the end position
			O = ((GArrow*)object)->viewerEnd();

			if (Abs(u-O.x)<=ds && Abs(v-O.y)<=ds)
				// Use only the ou,ov coordinates of object!
				return self->kernel->view.uv2xyz(O.x, O.y);

			um += O.x; um /= 2.0;
			vm += O.y; vm /= 2.0;
			if (Abs(u-um)<=ds && Abs(v-vm)<=ds)
				// Use only the ou,ov coordinates of object!
				return self->kernel->view.uv2xyz(um, vm);

			// check then angle handle
			if (object->isA()==GRulerClass && object->type==Ruler_Angle) {
				O = ((GRuler*)object)->viewerAngle();
				if (Abs(u-O.x)<=ds && Abs(v-O.y)<=ds)
					// Use only the ou,ov coordinates of object!
					return self->kernel->view.uv2xyz(O.x, O.y);
			}
		}
	}

	// 2nd scan viewports
	if (self->showViewport) {
		for (int ii=0; ii<ViewerObject::nviews; ii++) {
			ViewportLine& vp = self->viewport[ii];
			// FIXME (VISIBLE CENTER) Check center point
			double vx, vy, vz;
			double vu, vv;
			vp.viewer->origin(&vx, &vy, &vz);
			// Project on our viewport
			self->kernel->view.xyz2uv(vx, vy, vz, &vu, &vv);
			if (Abs(u-vu)<=ds && Abs(v-vv)<=ds)
				// Get x,y,z from projected center on our viewport!
				return self->kernel->view.uv2xyz(vu, vv);
		}
	}

	// 3rd scan for a close intersection if any
	double vu,vv;
	if (self->viewer->d2.showVertex && self->viewer->d2.closestVertex(u,v,ds, &vu,&vv))
		return self->kernel->view.uv2xyz(vu, vv);

	// 4th snap to 1/20th of the grid
	double dx = self->viewer->decoration.grid_dx / 20.0;
	double dy = self->viewer->decoration.grid_dy / 20.0;
	double dz = self->viewer->decoration.grid_dz / 20.0;
	double du = self->viewer->decoration.grid_du / 20.0;
	double dv = self->viewer->decoration.grid_dv / 20.0;

	// make it min least 2 pixels
	if (!Eq0(du,SMALL))
		while (!InRange(d, du*self->kernel->view.Sx(), d*10.0)) {
			double f = du*self->kernel->view.Sx()<d? 2.0 : 0.5;
			du *= f;
			switch (self->viewer->decoration.gridU) {
				case 'X':
					dx *= f;
					break;
				case 'Y':
					dy *= f;
					break;
				case 'Z':
					dz *= f;
					break;
			}
		}

	if (!Eq0(dv,SMALL))
		while (!InRange(d, dv*self->kernel->view.Sy(), d*10.0)) {
			double f = dv*self->kernel->view.Sy()<d? 2.0 : 0.5;
			dv *= f;
			switch (self->viewer->decoration.gridV) {
				case 'X':
					dx *= f;
					break;
				case 'Y':
					dy *= f;
					break;
				case 'Z':
					dz *= f;
					break;
			}
		}

	// Round to the closest of the grid
	double x = self->kernel->view.uv2x(u,v);
	double y = self->kernel->view.uv2y(u,v);
	double z = self->kernel->view.uv2z(u,v);

	bool calcXYZ = false;
	switch (self->viewer->decoration.gridU) {
		case 'X':
			x = (double)Round(x / dx) * dx;
			break;
		case 'Y':
			y = (double)Round(y / dy) * dy;
			break;
		case 'Z':
			z = (double)Round(z / dz) * dz;
			break;
		default:
			u = (double)Round(u / du) * du;
			calcXYZ = true;
	}
	switch (self->viewer->decoration.gridV) {
		case 'X':
			x = (double)Round(x / dx) * dx;
			break;
		case 'Y':
			y = (double)Round(y / dy) * dy;
			break;
		case 'Z':
			z = (double)Round(z / dz) * dz;
			break;
		default:
			v = (double)Round(v / dv) * dv;
			calcXYZ = true;
	}
	if (calcXYZ) {
		x = self->kernel->view.uv2x(u,v);
		y = self->kernel->view.uv2y(u,v);
		z = self->kernel->view.uv2z(u,v);
	}
	return Point(x,y,z);
} // _snapUV

/* --- Viewer_aspect --- */
static PyObject* Viewer_aspect(ViewerObject* self, PyObject* args)
{
	double aspect=-1.0;

	if (!PyArg_ParseTuple(args, "|d", &aspect)) return NULL;
	if (aspect<0.0)
		return PyFloat_FromDouble(self->kernel->view.aspect);
	else
		self->kernel->view.aspect = aspect;
	Py_RETURN_NONE;
} // Viewer_aspect

/* --- Viewer_basis ---
 * @return	return basis
 */
static PyObject* Viewer_basis(ViewerObject* self, PyObject* args)
{
	char axis;
	PyObject* basis=NULL;
	if (!PyArg_ParseTuple(args, "C|O", &axis, &basis)) return NULL;

	if (basis != NULL) {
		Py_RETURN_NONE;
	} else
		// return the normal matrix
		switch (axis) {
			case 'u':
			case 'U':
				return Py_Vector(self->kernel->view.axisu());
			case 'v':
			case 'V':
				return Py_Vector(self->kernel->view.axisv());
			case 'w':
			case 'W':
				return Py_Vector(self->kernel->view.axisw());
			default:
				Py_RETURN_NONE;
		}
} // Viewer_basis

/* --- Viewer_bbox ---
 * The same routine as the Geometry_bbox, but then transforms the bounding box
 * to the viewport coordinate system
 */
static PyObject* Viewer_bbox(ViewerObject* self, PyObject* args)
{
	char* type;
	PyObject* obj = NULL;
	BBox bbox;

	if (!PyArg_ParseTuple(args, "sO", &type, &obj)) return NULL;
	if (type[0]==0) {
		PyErr_SetString(PyExc_TypeError, "Invalid object type body, zone, region expected");
		return NULL;
	}

	if (type[0]=='b' || type[0]=='B') {
		GBody* body = Py_GBody(self->geometry,obj);
		if (body==NULL) return NULL;
		bbox = body->bbox();
	} else
	if (type[0]=='r' || type[0]=='R') {
		VRegion* region = Py_VRegion(self,obj);
		if (region==NULL) return NULL;
		bbox = region->region()->bbox();
	} else
//	if (type[0]=='z' || type[0]=='Z') {
//		// FIXME not implemented yet
//	} else
	if (type[0]=='o' || type[0]=='O') {
		GObject* object = Py_Object(self->geometry,obj);
		bbox = object->bbox();
	}

	if (bbox.isValid()) {
		bbox.transform(self->kernel->view.invMatrix());
		return Py_BuildValue("[dddddd]",
				bbox.low().x,  bbox.low().y,  bbox.low().z,
				bbox.high().x, bbox.high().y, bbox.high().z);
	} else
		Py_RETURN_NONE;
} // Viewer_bbox

/* --- Viewer_bbox2D --- */
static PyObject* Viewer_bbox2D(ViewerObject* self, PyObject* args)
{
	char* type;
	PyObject* obj = NULL;
	BBox bbox;

	if (!PyArg_ParseTuple(args, "sO", &type, &obj)) return NULL;
	if (type[0]==0) {
		PyErr_SetString(PyExc_TypeError, "Invalid object type body, zone, region expected");
		return NULL;
	}

	// FIXME: removed in refactor: there is no bbox2D in GBody
	if (type[0]=='b' || type[0]=='B') {
		VBody* body = Py_VBody(self,obj);
		if (body==NULL) return NULL;
		bbox = body->bbox2D();

	} else
	//if (type[0]=='r' || type[0]=='R') {
	//	VRegion* region = Py_VRegion(self,obj);
	//	if (region==NULL) return NULL;
	//	//bbox = region->region()->bbox2D();
	//	assert(false);
	//} else
	//if (type[0]=='z' || type[0]=='Z') {
	//	// FIXME not implemented yet
	//	assert(false);
	//} else
	if (type[0]=='o' || type[0]=='O') {
		GObject* object = Py_Object(self->geometry,obj);
		bbox = object->bboxView(self);
	}

	if (bbox.isValid())
		return Py_BuildValue("[dddd]",
				bbox.low().x,  bbox.low().y,
				bbox.high().x, bbox.high().y);
	else
		Py_RETURN_NONE;
} // Viewer_bbox2D

/* --- Viewer_bodyVar --- */
static PyObject* Viewer_bodyVar(ViewerObject* self, GBody* body,
				const char* var, PyObject* value)
{
	if (!strcmp(var, "move")) {	// Move relative from the last save position
		if (value==NULL) {
			PyErr_SetString(PyExc_TypeError, "body move doesn't return anything.");
			return NULL;
		}
		// What to move
		int opt = PyLong_AsLong(value);
		self->geometry->geometry->lockWrite();
		body->move(opt, body->savedPosition()+self->move, self->kernel->view.axisw());
		body->create(true);
		self->geometry->geometry->invalidateBody(body);
		self->geometry->geometry->unlockWrite();
	} else
	if (!strcmp(var, "rotate")) {	// rotate around an axis (axis,angle)
		// Calculate new position
		Point pos = body->savedPosition();
		// convert to relative
		double du = self->kernel->view.xyz2u(pos) - self->pivotU;
		double dv = self->kernel->view.xyz2v(pos) - self->pivotV;
		// Rotate around pivot point
		double u = self->rotateCos*du - self->rotateSin*dv + self->pivotU;
		double v = self->rotateSin*du + self->rotateCos*dv + self->pivotV;
		double w = self->kernel->view.xyz2w(pos);
		// back to absolute
		pos = self->kernel->view.uvw2xyz(u,v,w);

		self->geometry->geometry->lockWrite();

		// restore everything apart show
		dword show = body->show;
		body->restore();
		body->show = show;

		// 1st move. RPP can change position if aligned to axis
		body->move(0, pos, self->kernel->view.axisw());
		// 2nd Rotate in place
		body->rotate(self->rotateAngle, self->rotateAxis);
		body->create(true);
		self->geometry->geometry->invalidateBody(body);
		self->geometry->geometry->unlockWrite();
	} else {
		PyErr_Format(PyExc_TypeError, "Invalid type \'%s\' specified", var);
		return NULL;
	}
	Py_RETURN_NONE;
} // Viewer_bodyVar

/* --- Viewer_body --- */
static PyObject* Viewer_body(ViewerObject* self, PyObject* args)
{
	PyObject* obj;
	char* var;
	PyObject* value=NULL;
	if (!PyArg_ParseTuple(args, "Os|O", &obj, &var, &value)) return NULL;

	if (Py_Check4Pattern(obj)) {
		const char* pattern = PyUnicode_AsUTF8(obj);
		for (auto body : self->geometry->geometry->bodies) {
			if (!fnmatch(pattern, body->cname(), 0)) {
				PyObject* ret = Viewer_bodyVar(self, body, var, value);
				if (ret==NULL) return NULL;
				Py_XDECREF(ret);	// ret can be NULL use XDECREF
			}
		}
		Py_RETURN_NONE;
	} else
	if (PyList_Check(obj)) {
		for (ssize_t i=0; i<PyList_GET_SIZE(obj); i++) {
			GBody* body = Py_GBody(self->geometry, PyList_GetItem(obj,i));
			if (body==NULL) return NULL;
			PyObject* ret = Viewer_bodyVar(self, body, var, value);
			if (ret==NULL) return NULL;
			Py_XDECREF(ret);	// ret can be NULL use XDECREF
		}
	} else
	if (PyTuple_Check(obj)) {
		for (ssize_t i=0; i<PyTuple_GET_SIZE(obj); i++) {
			GBody* body = Py_GBody(self->geometry, PyTuple_GetItem(obj,i));
			if (body==NULL) return NULL;
			PyObject* ret = Viewer_bodyVar(self, body, var, value);
			if (ret==NULL) return NULL;
			Py_XDECREF(ret);	// ret can be NULL use XDECREF
		}
	} else  {
		GBody* body = Py_GBody(self->geometry, obj);
		if (body==NULL) return NULL;
		return Viewer_bodyVar(self, body, var, value);
	}
	Py_RETURN_NONE;
} // Viewer_body

/* --- Viewer_camera --- */
static PyObject* Viewer_camera(ViewerObject* self, PyObject* args)
{
	char*	name;
	double	u=0.0, v=0.0;

	if (!PyArg_ParseTuple(args, "s|dd", &name, &u, &v)) return NULL;

	if (!strcmp(name,"position")) {
		double x,y,z;
		self->kernel->view.cameraPosition(&x,&y,&z);
		return Py_BuildValue("ddd", x, y, z);
	} else
	if (!strcmp(name,"direction")) {
		double dx,dy,dz;
		self->kernel->view.rayDirection(u,v, &dx,&dy,&dz);
		return Py_BuildValue("ddd", dx, dy, dz);
	} else
	if (!strcmp(name,"focal")) {
		if (u>0.0)
			self->kernel->view.focalLength(u);
		else
			return PyFloat_FromDouble(self->kernel->view.focalLength());
	} else
	if (!strcmp(name,"fov")) {
		if (u>0.0)
			self->kernel->view.fov(Rad(u));
		else
			return PyFloat_FromDouble(Deg(self->kernel->view.fov()));
	} else {
		PyErr_Format(PyExc_SyntaxError, "'%s' is not a valid type option", name);
		return NULL;
	}
	Py_RETURN_NONE;
} // Viewer_camera

/* --- Viewer_calcViewport --- */
static void Viewer_calcViewport(ViewerObject* self, int id)
{
	ViewportLine& vp = self->viewport[id];
	const Matrix4& vpmatrix = vp.viewer->view().matrix();

	double w = (double)(self->viewer->width()  -VPMARGIN-4);
	double h = (double)(self->viewer->height() -VPMARGIN-2);

	// Find center of the viewport screen
	double vx, vy, vz;
	double vu, vv;
	vp.viewer->origin(&vx, &vy, &vz);
	self->kernel->view.xyz2uv(vx, vy, vz, &vu, &vv);

	// convert to screen coordinates
	vp.uc = self->kernel->view.u2id(vu);
	vp.vc = self->kernel->view.v2jd(vv);

	// w vector of other viewport
	double wx = vpmatrix(0,2);
	double wy = vpmatrix(1,2);
	double wz = vpmatrix(2,2);

	// The projected viewport LINE is NOT the intersection of
	// the two planes, rather a line parallel to this but
	// passing from the PROJECTED CENTER of the other viewport

	// project on current viewport (screen coordinates)
	// calculate the projection of the w-vector and get the
	// perpendicular (g,f) -> (f,-g) swapped
	const Matrix4& matrix = self->kernel->view.matrix();
	vp.g =   matrix(0,0)*wx + matrix(1,0)*wy + matrix(2,0)*wz;
	vp.f = -(matrix(0,1)*wx + matrix(1,1)*wy + matrix(2,1)*wz);
	vp.f *= self->kernel->view.aspect;
	// create perpendicular conic
	if (ISSMALL(vp.g) && ISSMALL(vp.f)) {
		vp.visible = false;
		return;
	}

	// The conic is projected on a viewport parallel to the current
	// one and passing from the center of the other viewport
	// line equation is g*x + f*y + c = 0
	vp.c = -(vp.uc*vp.g + vp.vc*vp.f);

	// normalize to be g^2+f^2=1.0
	double s = Sq(vp.g) + Sq(vp.f);
	if (!ISZERO(s)) {
		s = 1.0 / sqrt(s);
		vp.g *= s;
		vp.f *= s;
		vp.c *= s;
	}

	// find intersections with viewport visible window
	int n = 0;	// intersection counter
	if (!ISZERO(vp.f)) {
		// x = 0
		double y = -(vp.c + vp.g*VPMARGIN) / vp.f;
		if (InRange((double)VPMARGIN, y, h)) {
			vp.x[n] = VPMARGIN;
			vp.y[n] = Round(y);
			n++;
		}
		// x = w
		y = -(vp.c + vp.g*w) / vp.f;
		if (InRange((double)VPMARGIN, y, h)) {
			vp.x[n] = (int)w;
			vp.y[n] = Round(y);
			n++;
		}
	}
	if (!ISZERO(vp.g)) {
		// y = 0
		double x = -(vp.c + vp.f*VPMARGIN) / vp.g;
		if (n<2 && InRange((double)VPMARGIN, x, w)) {
			vp.x[n] = Round(x);
			vp.y[n] = VPMARGIN;
			n++;

		}
		// y = h
		x = -(vp.c + vp.f*h) / vp.g;
		if (n<2 && InRange((double)VPMARGIN, x, w)) {
			vp.x[n] = Round(x);
			vp.y[n] = (int)h;
			n++;
		}
	}

	// after all calculation move the center in the visible area
	if (vp.uc<VPMARGIN || vp.uc>=w || vp.vc<VPMARGIN || vp.vc>=h) {
		if (ISZERO(vp.g) || ISZERO(vp.f)) {
			// vertical or horizontal line
			vp.xc = Range(VPMARGIN, Round(vp.uc), (int)w);
			vp.yc = Range(VPMARGIN, Round(vp.vc), (int)h);
		} else
		if (n==0) {
			if (vp.g * vp.f > 0.0) {	// same sign
				// can only be the 0,0 or the w,h
				if (Abs(vp.c) < Abs(vp.g*w + vp.f*h + vp.c)) {
					vp.xc = VPMARGIN;
					vp.yc = VPMARGIN;
				} else {
					vp.xc = (int)w;
					vp.yc = (int)h;
				}
			} else {
				// can only be the w,0 or the 0,h
				if (Abs(vp.g*w+vp.c) < Abs(vp.f*h + vp.c)) {
					vp.xc = (int)w;
					vp.yc = VPMARGIN;
				} else {
					vp.xc = VPMARGIN;
					vp.yc = (int)h;
				}
			}
		} else
		if (n==1) {
			vp.xc = Round(vp.x[0]);
			vp.yc = Round(vp.y[0]);
		} else {
			// diagonal line
			double d1 = Sq(vp.x[0]-vp.uc) + Sq(vp.y[0]-vp.vc);
			double d2 = Sq(vp.x[1]-vp.uc) + Sq(vp.y[1]-vp.vc);
			if (d1<d2) {
				vp.xc = Round(vp.x[0]);
				vp.yc = Round(vp.y[0]);
			} else {
				vp.xc = Round(vp.x[1]);
				vp.yc = Round(vp.y[1]);
			}
		}
	} else {
		vp.xc = Round(vp.uc);
		vp.yc = Round(vp.vc);
	}

	// End position of the w vector
	if (vp.viewer->view().projection==ViewPort::Projection::Perspective) {
		double x,y,z;
		vp.viewer->view().cameraPosition(&x,&y,&z);
		double u = self->kernel->view.xyz2u(x,y,z);
		double v = self->kernel->view.xyz2v(x,y,z);
		if (self->kernel->view.inside(u,v)) {
			vp.xw = self->kernel->view.u2i(u);
			vp.yw = self->kernel->view.v2j(v);
		} else {
			vp.xw = vp.xc + Round(vp.g*vp.wLength);
			vp.yw = vp.yc + Round(vp.f*vp.wLength);
		}
	} else {
		vp.xw = vp.xc + Round(vp.g*vp.wLength);
		vp.yw = vp.yc + Round(vp.f*vp.wLength);
	}

	// corrected c for visual center
	vp.cc = -((double)vp.xc*vp.g + (double)vp.yc*vp.f);

	// ViewportLine is outside
	if (n==0) {
		// find closest line to draw
		if (ISZERO(vp.g)) {
			// horizontal
			vp.x[0] = VPMARGIN;
			vp.x[1] = (int)w;
			vp.y[0] = vp.y[1] = vp.yc;
		} else
		if (ISZERO(vp.f)) {
			// vertical
			vp.x[0] = vp.x[1] = vp.xc;
			vp.y[0] = VPMARGIN;
			vp.y[1] = (int)h;
		} else {
			vp.x[0] = vp.x[1] = vp.xc;
			vp.y[0] = vp.y[1] = vp.yc;
		}
	}

	vp.visible = true;
} // Viewer_calcViewport

/* --- Viewer_closest --- */
static PyObject* Viewer_closest(ViewerObject* self, PyObject* args)
{
	int i,j,k;
	double pi, pj, d=1.0;
	PyObject* obj=NULL;
	const char* defoptions = "vCVOBZR";
	const char* options = defoptions;

	if (!PyArg_ParseTuple(args, "ii|dzO", &i, &j, &d, &options, &obj)) return NULL;
	pi = (double)i;
	pj = (double)j;
	double u = self->kernel->view.i2u((int)pi);
	double v = self->kernel->view.j2v((int)pj);

	if (options == NULL) options = defoptions;

	int modifier = 0;	// modifier in the search
	self->viewer->stopThread();

	// Do 2 passes
	// first ipass=0 only selected items and not viewports
	// second ipass=1 everything else and viewports
	for (int ipass=0; ipass<2; ipass++)
		// Loop over the items to search
		for (const char *o=options; *o; o++) {
			switch (*o) {
				case 's':	// check only spline objects must be followed by O
					modifier = GSplineClass;
					break;

				// ------ Search Viewports (ORIGIN ONLY) ------
				case 'v':
					if (self->showViewport and ipass) {
						int min = -1;
						double mind = INFINITE;
						double d1=d+(double)self->viewport[0].originWidth;
						// check the cursor
						for (k=0; k<ViewerObject::nviews; k++) {
							ViewportLine& vp = self->viewport[k];
							if (vp.visible) {
								double c = Sq(pi-(double)vp.xc)
									 + Sq(pj-(double)vp.yc);
								if (c <= mind) {
									mind = c;
									min  = k;
								}
							}
						}
						if (min>=0 && mind<=Sq(d1))
							return Py_BuildValue("sis","V",min,NULL);
					}
					break;

				// ------ Search Viewports (2nd) (Line) ------
				case 'V':
					if (self->showViewport and ipass) {
						int min = -1;
						double mind = INFINITE;
						double d1=d+(double)self->viewport[0].originWidth;
						// check the line
						for (k=0; k<ViewerObject::nviews; k++) {
							ViewportLine& vp = self->viewport[k];
							if (vp.visible) {
								double c = Abs(vp.g*pi + vp.f*pj + vp.cc);
								if (c <= mind) {
									mind = c;
									min  = k;
								}
							}
						}
						if (min>=0 && mind<=d1)
							return Py_BuildValue("sis","V",min,NULL);
					}
					break;

				// ------ Search Objects ------
				case 'O':
					for (auto object : self->geometry->objects) {
						if (modifier and object->isA() != modifier) continue;
						if (!object->show) continue;
						if (ipass==0) {
							if (!object->select) continue;
						} else
							if (object->select) continue;
						object->transform(self);
						if (!object->visible()) continue;
						int close = object->closest(self,i,j,Round(d));
						if (close != NodeId_NONE)
							return Py_BuildValue("sii","O",object->id(),close);
					}
					break;

				// ------ Search Editing body ------
				case 'C':
					if (self->geometry->cursorShow and ipass==0) {
						int ci, cj;
						self->kernel->view.xyz2ij3D(self->geometry->cursor, &ci, &cj);
						double ici = (double)(i-ci);
						double jcj = (double)(j-cj);
						double dist = sqrt(Sq(ici) + Sq(jcj));
						// Free move
						if (dist <= (double)self->geometry->cursorMoveSize + d)
							return Py_BuildValue("sii","C", 0, 0);

						double r = (double)self->geometry->cursorSize;
						int   d2 = Round(Sq(d));

						// Axis X
						const Matrix4& matrix = self->kernel->view.matrix();
						double Dx =  r * matrix(0,0);
						double Dy = -r * matrix(0,1);
						double Dlen2 = Sq(Dx) + Sq(Dy);
						//if (Dlen2 < 2) return -1;
						// Check distance, calculate the dot product (to the square)
						// of the two vs length
						// Calculate the projection distance using the dot product
						if ( Dlen2>3.0 &&
						    (Sq(Dy*ici - Dx*jcj) < d2*Dlen2) &&
						    (InRange(0.0, Dx*ici + Dy*jcj, Dlen2)))
							return Py_BuildValue("sii","C", 0, NodeId_axisX);

						// Axis Y
						Dx =  r * matrix(1,0);
						Dy = -r * matrix(1,1);
						Dlen2 = Sq(Dx) + Sq(Dy);
						if ( Dlen2>3.0 &&
						    (Sq(Dy*ici - Dx*jcj) < d2*Dlen2) &&
						    (InRange(0.0, Dx*ici + Dy*jcj, Dlen2)))
							return Py_BuildValue("sii","C", 0, NodeId_axisY);

						// Axis Z
						Dx =  r * matrix(2,0);
						Dy = -r * matrix(2,1);
						Dlen2 = Sq(Dx) + Sq(Dy);
						if ( Dlen2>3.0 &&
						    (Sq(Dy*ici - Dx*jcj) <= d2 * Dlen2) &&
						    (InRange(0.0, Dx*ici + Dy*jcj, Dlen2)))
							return Py_BuildValue("sii","C", 0, NodeId_axisZ);

						// Rotate around w
						if (Abs(dist-r) <= d)
							return Py_BuildValue("sii","C", 0, NodeId_axisR);
					}
					break;

				// ------ Search Bodies ------
				case 'B': {
					int dd = Round(d);
					bool ebvisi = false;	// edit body visibility
					// remember visibility
					if (self->geometry->geometry->editBody)
						ebvisi = self->kernel->getBody(
							self->geometry->geometry->editBody)->visible;

					// Draw in the clipping region to mark the visible ones
					self->kernel->clearVisibleBodies();
					self->viewer->painter.clip(Max(0,i-dd),Max(0,j-dd), i+dd,j+dd);
					self->viewer->draw(DRAW_SEGMENTS|DRAW_WIREFRAME);
					self->viewer->painter.resetClip();

					// First check editBody if any
					if (ipass==0 and
					    self->geometry->geometry->editBody and
					   (self->geometry->geometry->editBody->show & BIT_SELECT) and
					    self->kernel->getBody(self->geometry->geometry->editBody)->visible) {
						Point r = self->kernel->view.uv2xyz(u,v);
						int close = self->geometry->geometry->editBody->closest(r,
										d/self->kernel->view.Sx(),
										self->kernel->view.axisw());

						// Special case for PLA and position is visible
						assert(	GBody::Type::YZP<GBody::Type::XZP &&
							GBody::Type::XZP<GBody::Type::XYP &&
							GBody::Type::XYP< GBody::Type::PLA);
						if (InRange(GBody::Type::YZP, self->geometry->geometry->editBody->type(), GBody::Type::PLA)) {
							// Check if we are close to the borders like in the viewports
							int ww = self->viewer->width()  / 20;
							int hh = self->viewer->height() / 20;
							Point p = self->geometry->geometry->editBody->position();
							int ei = self->kernel->view.u2i(self->kernel->view.xyz2u(p));
							int ej = self->kernel->view.v2j(self->kernel->view.xyz2v(p));
							if ((i<=ww && ei>ww) ||
							    (i>=self->viewer->width()-ww && ei<self->viewer->width()-ww) ||
							    (j<=hh && ej>hh) ||
							    (j>=self->viewer->height()-hh && ej<self->viewer->height()-hh)) {
								if (close == GObject::NodeId_1) {
									// Set also the sign of rotation
									((GPLABody*)self->geometry->geometry->editBody)->signMove20(r,
											self->kernel->view.axisw());
									return Py_BuildValue("sii","B",
										self->geometry->geometry->editBody->id(),NodeId_axisR);
								}
							}
						}
						return Py_BuildValue("sii", "B", self->geometry->geometry->editBody->id(), close);
					}

					// scan visible bodies to find the closest one
					// FIXME: potential source of troubles
					self->kernel->lock();
					double min = INFINITE;
					VBody* closestBody = NULL;
					for (size_t ib=0; ib<self->kernel->nBodies(); ib++) {
						VBody* body = self->kernel->bodies[ib];
						if (!body->visible) continue;
						double m = body->body()->distance(
								self->kernel->view.uv2xyz(u,v));
						if (m<min) {
							min = m;
							closestBody = body;
						}
					}
					self->kernel->unlock();

					// FIXME: removed in refactor: cannot access kernel->getBody in all
					// cases, the body may not exist
					// restore visibility of editBody
					if (self->geometry->geometry->editBody)
						self->kernel->getBody(self->geometry->geometry->editBody)->visible = ebvisi;

					if (closestBody) {
						int close = 0;
						if (closestBody->body()==self->geometry->geometry->editBody &&
						   (self->geometry->geometry->editBody->show & BIT_SELECT)) {
							// Find closest node/edge on the body
							Point r = self->kernel->view.uv2xyz(u,v);
							close = closestBody->body()->closest(r,
										d/self->kernel->view.Sx(),
										self->kernel->view.axisw());
						}
						return Py_BuildValue("sii", "B", closestBody->id(), close);
					} }
					break;

				// ------ Search Zones ------
				case 'Z':
					if (obj!=NULL and ipass) {
						VRegion* region = Py_VRegion(self, obj);

						// Check in 2D mode
						if (self->viewer->d2.fillRegions) {
							if (region != NULL) {
								// Iterate over the zones composing the region
								// find find if we are inside
								int zi=0;
								// FIXME: potential source of troubles
								self->kernel->lock();
								self->kernel->engine()->incBodyCheckId();
								PyObject* zones = NULL;
								const ViewPort& V = self->kernel->view;
								const Matrix4& M = V.matrix();
								for (int iz=0; iz<region->zones().size(); iz++) {
									const VZone *zone = region->zones()[iz];
									double x = V.uv2x(u,v);
									double y = V.uv2y(u,v);
									double z = V.uv2z(u,v);
									if (zone->inside2D(self->kernel->engine(),
									    x,y,z,-M(0,2),-M(1,2),-M(2,2))) {
										if (zones==NULL) zones = PyList_New(0);
										PyList_Append(zones, PyLong_FromLong(zi));
									}
									zi++;
								}
								self->kernel->unlock();
								if (zones)
									return Py_BuildValue("sOs","Z",zones,NULL);
							} else
								// avoid the message, especially when voxel
								// region is involved
								PyErr_Clear();
						}

						// if failed check in 3D mode
						if (self->viewer->d3.show) {
							self->kernel->lock();
							self->kernel->engine()->incBodyCheckId();
							const ViewPort& V = self->kernel->view;
							//const Matrix4& M = V.matrix();
							double x = V.uv2x(u,v);
							double y = V.uv2y(u,v);
							double z = V.uv2z(u,v);
							//VZone *zone = region->inside2D(self->kernel->engine(), x,y,z, -M(0,2),-M(1,2),-M(2,2));
							//if (zone->region()->transparent()) {

							Ray ray;
							ray.use_clip      = self->kernel->clipBodyCount() > 0;
							ray.skip_1stblack = self->viewer->d3.skip_1stblack;
							self->kernel->view.rayPosition(u,v, &x,&y,&z);
							double dx,dy,dz;
							self->kernel->view.rayDirection(u,v, &dx,&dy,&dz);
							// Find the new zone of the camera
							self->kernel->engine()->incBodyCheckId();
							VZone* zone = self->kernel->engine()->whereRay(x,y,z, dx,dy,dz, SMALL3D, NULL);
							if (zone) {
								if (zone->region() != region && zone->region()->transparent()) {
									ray.push(RaySegment(x,y,z, dx,dy,dz, zone));
									self->kernel->engine()->incBodyCheckId();
									self->viewer->d3.nextIntersection(self->kernel->engine(), &ray);
									zone = ray.hitZone();
								}
							} else
								zone = NULL;
							self->kernel->unlock();
							if (zone && zone->region() == region) {
								PyObject* zones = PyList_New(0);
								PyList_Append(zones, PyLong_FromLong(zone->id()));
								return Py_BuildValue("sOs","Z",zones,NULL);
							}
						}
					}
					break;

				// ------ Search Regions ------
				case 'R':
					if (ipass and (self->viewer->d2.fillRegions or self->viewer->d3.show)) {
						// Find region under point
						// FIXME: potential source of troubles
						self->kernel->lock();
						self->kernel->engine()->incBodyCheckId();
						PyObject* regions = NULL;
						const ViewPort& V = self->kernel->view;
						const Matrix4& M = V.matrix();
						for (size_t ir=0; ir<self->kernel->nRegions(); ir++) {
							VRegion* region = self->kernel->regions[ir];
							double x = V.uv2x(u,v);
							double y = V.uv2y(u,v);
							double z = V.uv2z(u,v);
							VZone* zone = region->inside2D(self->kernel->engine(),
										x,y,z, -M(0,2),-M(1,2),-M(2,2));
							// Check for 3D if needed
							bool d3used;
							if (self->viewer->d3.show && (zone==NULL || zone->region()->transparent())) {
								//self->kernel->engine()->incBodyCheckId();
								//zone = self->kernel->engine()->whereRay(x,y,z, dx,dy,dz, SMALL3D, zone);
								Ray ray;
								ray.use_clip      = self->kernel->clipBodyCount() > 0;
								ray.skip_1stblack = self->viewer->d3.skip_1stblack;
								self->kernel->view.rayPosition(u,v, &x,&y,&z);
								double dx,dy,dz;
								self->kernel->view.rayDirection(u,v, &dx,&dy,&dz);

								// Find the new zone of the camera
								self->kernel->engine()->incBodyCheckId();
								zone = self->kernel->engine()->whereRay(x,y,z, dx,dy,dz, SMALL3D, NULL);

								ray.push(RaySegment(x,y,z, dx,dy,dz, zone));
								self->kernel->engine()->incBodyCheckId();
								self->viewer->d3.nextIntersection(self->kernel->engine(), &ray);
								zone = ray.hitZone();
								d3used = true;
							} else
								d3used = false;
							if (zone) {
								if (regions==NULL) regions = PyList_New(0);
								PyList_Append(regions, PyLong_FromLong(zone->region()->id()));
								if (d3used) break;
							}
						}
						self->kernel->unlock();
						if (regions) return Py_BuildValue("sOs","R",regions,NULL);
					}
					break;
				default:
					PyErr_Format(PyExc_TypeError, "Invalid option '%c' specified", o);
					return NULL;
			}
		}
	Py_RETURN_NONE;
} // Viewer_closest

/* --- Viewer_derive --- */
static PyObject* Viewer_derive(ViewerObject* self, PyObject* Py_UNUSED(ignored))
{
	self->kernel->derive();
	Py_RETURN_NONE;
} // Viewer_derive

/* _endDraw: notifier function for end of projection */
static void _endDraw(ViewerObject* self)
{
	DUMP(cout << "_endDraw("<<self->viewer->title()<<")" << endl);
	_sendMessage(self, "EndDraw");
} // _endDraw

/* --- Viewer_draw --- */
static PyObject* Viewer_draw(ViewerObject* self, PyObject* args)
{
	int  asThread = false;
	int  mask     = -1;
	bool fast     = false;

	if (!PyArg_ParseTuple(args, "i|ii", &asThread, &mask, &fast)) return NULL;
	DUMP(cout << "Viewer_draw("<<self->viewer->title()<<", " << asThread << ")" << endl);
	if (asThread) {
		// Run as a thread
		self->viewer->spawnDraw((NotifyFunc)_endDraw, self, mask, fast);
		//assert(self->viewer->state() == GeometryViewer::Projection::DRAW);
		Py_RETURN_NONE;
	} else {
		// Run synchronously
		if (self->viewer->state() == GeometryViewer::Projection::DRAW) {
			// Stop any pending drawing thread, don't stop other threads
			DUMP(cout << "Viewer_draw(" << self->viewer->title()
				<< ").stopThread()" << endl);
			self->viewer->stopThread();
		}
		assert(fast);
		return PyLong_FromLong((int)self->viewer->draw(mask, fast));
	}
} // Viewer_draw

/* --- Viewer_calcWindow --- */
static PyObject* Viewer_calcWindow(ViewerObject* self, PyObject* args)
{
	double zoom;

	if (!PyArg_ParseTuple(args, "d", &zoom)) return NULL;

	self->kernel->view.calcWindow(zoom);
	Py_RETURN_NONE;
} // Viewer_calcWindow

/* --- Viewer_edit --- */
static PyObject* Viewer_edit(ViewerObject* self, PyObject* args)
{
	char* name;
	PyObject* value=NULL;
	PyObject* value2=NULL;

	if (!PyArg_ParseTuple(args, "s|OO", &name, &value, &value2)) return NULL;

	if (!strcmp(name, "move")) {	// Calculate move distance
		if (value) {
			// Expect: (u,v), (axis, snap)
			double u,v;
			if (!Py_GetUV(value, &u, &v)) return NULL;

			// Expect: (u,v), (axis, snap [, relative])
			// relative = move to current viewing plane if the object is behind
			if (!PyTuple_Check(value2) || PyTuple_GET_SIZE(value2)<2) {
				PyErr_SetString(PyExc_TypeError, "tuple of size 2 expected");
				return NULL;
			}

			// how to move
			int axis = PyLong_AsLong(PyTuple_GetItem(value2, 0));
			int snap = PyLong_AsLong(PyTuple_GetItem(value2, 1));
			bool relative = true;
			if (PyTuple_GET_SIZE(value2)>=3)
				relative = PyObject_IsTrue(PyTuple_GetItem(value2, 2));
			if (PyErr_Occurred()) return NULL;

			// Real or aligned coordinates
			Point D = snap?
					_snapUV(self, u, v, self->geometry->snapDistance) :
					self->kernel->view.uv2xyz(u,v);
			self->move = self->start;

			// Find difference of self->move from absolute coordinates of projection on viewport
			Vector dS = self->move - self->kernel->view.uv2xyz(
						 self->kernel->view.xyz2u(self->move),
						 self->kernel->view.xyz2v(self->move));

			// Axis=value2[0] Find new position
			switch (axis) {
				case 0:	// X-axis
					self->move.x = D.x;
					if (relative) self->move.x += dS.x;
					break;
				case 1: // Y-axis
					self->move.y = D.y;
					if (relative) self->move.y += dS.y;
					break;
				case 2: // Z-axis
					self->move.z = D.z;
					if (relative) self->move.z += dS.z;
					break;
				default:
					self->move = D;
					if (relative) self->move += dS;
			}
			self->move -= self->start;
		} else
			return Py_Vector(self->move);
	} else
	if (!strcmp(name, "rotate")) {	// Calculate rotate angle
		if (value) {
			// Expect: (uv,  axis, snap))
			if (!PyTuple_Check(value) || PyTuple_GET_SIZE(value)!=3) {
				PyErr_SetString(PyExc_TypeError, "tuple of size 3 expected");
				return NULL;
			}
			// What to move
			double u,v;
			if (!Py_GetUV(PyTuple_GetItem(value, 0), &u, &v)) return NULL;
			self->rotateAxis = Py_GetVector(PyTuple_GetItem(value, 1));
			int snap = PyLong_AsLong(PyTuple_GetItem(value, 2));
			if (PyErr_Occurred()) return NULL;

			// Calculate angle
			self->rotateAngle = atan2(v-self->pivotV, u-self->pivotU) - self->pivotAngle;
			if (snap) // FIXME angle should be past from python
				self->rotateAngle = (double)Round(
						self->rotateAngle / self->geometry->snapAngle)
					* self->geometry->snapAngle;
			bsincos(self->rotateAngle, &self->rotateSin, &self->rotateCos);
		} else
			return PyFloat_FromDouble(Deg(self->rotateAngle));
	} else
	if (!strcmp(name, "value")) {	// Calculate move by typing
		if (value) {
			// Move list(dx[,dy[,dz]]), int(option)
			if (PyList_Check(value)) {
				double a = PyFloat_AsDouble(PyList_GetItem(value, 0));
				double b=0.0, c=0.0;
				if (PyList_Size(value)>1) {
					b = PyFloat_AsDouble(PyList_GetItem(value, 1));
					if (PyList_Size(value)>2)
						c = PyFloat_AsDouble(PyList_GetItem(value, 2));
				}

				self->move.x = 0.0;
				self->move.y = 0.0;
				self->move.z = 0.0;
				if (value2==NULL) {
					self->move.x = a;
					self->move.y = b;
					self->move.z = c;
				} else
					switch (PyLong_AsLong(value2)) {
						case 0:	// X-axis
							self->move.x = a;
							break;
						case 1: // Y-axis
							self->move.y = a;
							break;
						case 2: // Z-axis
							self->move.z = a;
							break;
						default:
							self->move.x = a;
							self->move.y = b;
							self->move.z = c;
					}
			} else {
				// Angle   float, list(axis)
				self->rotateAngle = Rad(PyFloat_AsDouble(value));
				self->rotateAxis  = Py_GetVector(value2);
			}
		} else
			return Py_Vector(self->move);
	} else
	if (!strcmp(name, "start")) {	// Set start point
		if (value==NULL)
			return Py_BuildValue("dd",self->startU,self->startV);
		else
		if (Py_IsVector(value)) {
			self->start  = Py_GetVector(value);
			self->startU = self->kernel->view.xyz2u(self->start);
			self->startV = self->kernel->view.xyz2v(self->start);
		} else
		if (PyTuple_Check(value) and PyTuple_GET_SIZE(value)==2) {
			// Expect ( (u,v)|(x,y,z), snap )
			PyObject* pos = PyTuple_GetItem(value,0);
			int snap = PyLong_AsLong(PyTuple_GetItem(value, 1));

			// Relative if (u,v)
			if (PyTuple_Check(pos) && PyTuple_GET_SIZE(pos)==2) {
				self->startU = PyFloat_AsDouble(PyTuple_GetItem(pos,0));
				self->startV = PyFloat_AsDouble(PyTuple_GetItem(pos,1));
				if (snap) {
					Point D = _snapUV(self, self->startU, self->startV,
							self->geometry->snapDistance);
					self->startU = self->kernel->view.xyz2u(D);
					self->startV = self->kernel->view.xyz2v(D);
				}
				if (PyErr_Occurred()) return NULL;
				self->start = self->kernel->view.uv2xyz(self->startU, self->startV);
			} else {	// Otherwise absolute x,y,z
				self->start = Py_GetVector(pos);
				if (PyErr_Occurred()) return NULL;
				self->startU = self->kernel->view.xyz2u(self->start);
				self->startV = self->kernel->view.xyz2v(self->start);
			}
		} else {
			PyErr_SetString(PyExc_TypeError, "tuple of size 2 (uv|xyz, snap) expected");
			return NULL;
		}
	} else
	if (!strcmp(name, "pivot")) {	// Set pivot point
		if (value) {
			if (PyTuple_Check(value) && PyTuple_GET_SIZE(value)==2) {
				self->pivotU = PyFloat_AsDouble(PyTuple_GetItem(value,0));
				self->pivotV = PyFloat_AsDouble(PyTuple_GetItem(value,1));
				self->pivotAngle = atan2(self->startV-self->pivotV, self->startU-self->pivotU);
			}
		} else
			return Py_BuildValue("dd",self->pivotU,self->pivotV);
	} else
	if (!strcmp(name, "cursor")) {	// Set start position from cursor
			self->start  = self->geometry->cursor;
			self->startU = self->kernel->view.xyz2u(self->start);
			self->startV = self->kernel->view.xyz2v(self->start);
	} else
	if (!strcmp(name, "object")) {	// Set start position from object
		if (value) {
			GObject* object = Py_Object(self->geometry,value);
			object->selectNode = PyLong_AsLong(value2);	// What to move
			if (PyErr_Occurred()) return NULL;
			self->start  = object->node(object->selectNode);
			self->startU = self->kernel->view.xyz2u(self->start);
			self->startV = self->kernel->view.xyz2v(self->start);
		} else
			return Py_Vector(self->start);
	} else
	if (!strcmp(name, "body")) {	// Set start position from body,item pair
		if (value) {
			GBody* body  = Py_GBody(self->geometry, value);
			self->start  = body->position();
			self->startU = self->kernel->view.xyz2u(self->start);
			self->startV = self->kernel->view.xyz2v(self->start);
		} else
			return Py_Vector(self->start);
	} else {
		PyErr_Format(PyExc_TypeError, "Invalid type '%s' specified", name);
		return NULL;
	}
	Py_RETURN_NONE;
} // Viewer_edit

/* --- Viewer_enclosed --- */
static PyObject* Viewer_enclosed(ViewerObject* self, PyObject* args)
{
	int l,t,r,b;
	char* options=NULL;

	if (!PyArg_ParseTuple(args, "iiii|s", &l, &t, &r, &b, &options)) return NULL;

	if (r<l) Swap(l,r);
	if (b<t) Swap(t,b);

//	double u = self->kernel->view.realX((int)pi);
//	double v = self->kernel->view.realY((int)pj);
//	double x = self->kernel->view.uv2x(u,v);
//	double y = self->kernel->view.uv2y(u,v);
//	double z = self->kernel->view.uv2z(u,v);

	self->viewer->painter.clip(l, t, r, b);
#ifdef EXPERIMENTAL
#if 0
	self->viewer->clearVisibleRegions();
#endif
#endif
	self->kernel->clearVisibleBodies();
	self->viewer->draw(DRAW_SEGMENTS|DRAW_WIREFRAME);
	self->viewer->painter.resetClip();

	if (options==NULL || options[0]=='B') {
		PyObject* bodies=NULL;
		// scan bodies to find the visible ones
		// FIXME: potential source of troubles
		self->kernel->lock();
		bodies = PyList_New(0);
		for (size_t ib=0; ib<self->kernel->nBodies(); ib++) {
			VBody* body = self->kernel->bodies[ib];
			if (!body->visible) continue;
			PyList_Append(bodies, PyLong_FromLong(body->id()));
		}
		self->kernel->unlock();

		return bodies;
	}

	if (options==NULL || strchr(options,'O')) {
		PyObject* objects=NULL;
		// scan regions to find the visible ones
		objects = PyList_New(0);
		for (auto object : self->geometry->objects) {
			if (!object->show) continue;
			object->transform(self);
			if (!object->visible()) continue;
			if (object->enclosed(self, l,t,r,b))
				PyList_Append(objects, PyLong_FromLong(object->id()));
		}
		return objects;
	}

	Py_RETURN_NONE;
} // Viewer_enclosed

/* --- Viewer_error --- */
static PyObject* Viewer_error(ViewerObject* self, PyObject* args)
{
	char* cmd=nullptr;
	int err = 0;
	if (!PyArg_ParseTuple(args, "|si", &cmd, &err)) return NULL;

	if (cmd==nullptr or !strcmp(cmd, "n"))		// return number of errors
		return PyLong_FromLong(self->kernel->errors());
	else
	if (!strcmp(cmd, "clear")) {	// clear error points
		_clearErrors(self);
		self->kernel->clearErrors();
	} else
	if (!strcmp(cmd, "show")) {
		_clearErrors(self);
		// create temporary point objects
		if (err==0)
			err = self->kernel->errors();
		else
			err = Min(err, self->kernel->errors());
		for (int i=1; i<=err; i++) {
			char name[10];
			sprintf(name,"%d",i);
			GObject* obj = new GPoint(name, Point_X);
			obj->color = self->geometry->geometry->errorColor;
			obj->anchor = Anchor_E;

			ZoneOfPoint pIn, pOut;
			double u,v, xmin, xmax, ymin, ymax;
			self->kernel->error(i, &u, &v, &pIn, &pOut, &xmin, &xmax, &ymin, &ymax);
			obj->position(self->kernel->view.uv2xyz(u,v));
			self->errors.push_back(obj);
		}
	} else
	if (!strcmp(cmd, "get")) {
		ZoneOfPoint pIn, pOut;
		double u,v;
		double xmin, xmax, ymin, ymax;
		VBody* body = self->kernel->error(err, &u, &v, &pIn, &pOut, &xmin, &xmax, &ymin, &ymax);
		if (body==NULL) Py_RETURN_NONE;

		PyObject* inList  = PyList_New(pIn.n);
		for (int i=0; i<pIn.n; i++)
			PyList_SetItem(inList, i,
				PyUnicode_FromFormat("%s:%d",
						pIn.zone[i]->region()->cname(),
						pIn.zone[i]->id()+1));
				//PyUnicode_FromString(pIn.zone[i]->region()->cname()));

		PyObject* outList = PyList_New(pOut.n);
		for (int i=0; i<pOut.n; i++)
			PyList_SetItem(outList, i,
				PyUnicode_FromFormat("%s:%d",
						pOut.zone[i]->region()->cname(),
						pOut.zone[i]->id()+1));
				//PyUnicode_FromString(pOut.zone[i]->region()->cname()));

		return Py_BuildValue("dddddddsOO",
			self->kernel->view.uv2x(u,v),
			self->kernel->view.uv2y(u,v),
			self->kernel->view.uv2z(u,v),
			xmin, xmax, ymin, ymax,
			body->cname(),
			inList, outList);
	} else
	if (!strcmp(cmd, "hash")) {
		return PyLong_FromLong(self->kernel->errorHash()*(int)self->viewer->state());
	} else
	if (!strcmp(cmd, "kernel")) {
		if (self->kernel->error()[0])
			return PyUnicode_FromString(self->kernel->error());
	} else
	if (!strcmp(cmd, "reset")) {
		self->kernel->errorReset();
	} else {
		PyErr_Format(PyExc_TypeError, "Invalid type '%s' specified", cmd);
		return NULL;
	}
	Py_RETURN_NONE;
} // Viewer_error

/* --- Viewer_extends --- */
static PyObject* Viewer_extends(ViewerObject* self, PyObject* args)
{
	double xmin, xmax, ymin, ymax;
	if (PyTuple_Size(args)!=4)
		return Py_BuildValue("dddd",
				self->kernel->view.minu(),
				self->kernel->view.minv(),
				self->kernel->view.maxu(),
				self->kernel->view.maxv());
	else {
		if (!PyArg_ParseTuple(args, "dddd", &xmin, &ymin, &xmax, &ymax)) return NULL;
		self->viewer->window(xmin, ymin, xmax, ymax);
	}
	Py_RETURN_NONE;
} // Viewer_extends

/* --- Viewer_font --- */
static PyObject* Viewer_font(ViewerObject* self, PyObject* args)
{
	char* fontname = NULL;
	char* fontpath = NULL;
	if (!PyArg_ParseTuple(args, "s|s", &fontname, &fontpath)) return NULL;

	if (!strcmp(fontname,"general")) {
		if (fontpath==NULL) {
			if (self->viewer->font.cname())
				return PyUnicode_FromString(self->viewer->font.cname());
		} else {
			loadFont(self->viewer->font, fontpath);
			// FIXME: once we know the font path we know where flair is installed
			//        and we can read farm's config file. It's a bit of a hack
			//        but for the moment it works. Should be replaced by a clean
			//        solution to get the path where flair resides.
			self->viewer->d3.readFARMConfig(fontpath);
		}
	} else
	if (!strcmp(fontname,"grid")) {
		if (fontpath==NULL) {
			if (self->viewer->decoration.gridFont.cname())
				return PyUnicode_FromString(self->viewer->decoration.gridFont.cname());
		} else
			loadFont(self->viewer->decoration.gridFont, fontpath);
	} else
	if (!strcmp(fontname,"palette")) {
		if (fontpath==NULL) {
			if (self->viewer->palette.font.cname())
				return PyUnicode_FromString(self->viewer->palette.font.cname());
		} else
			loadFont(self->viewer->palette.font, fontpath);
	} else {
		PyErr_Format(PyExc_TypeError, "Invalid font '%s' specified", fontname);
		return NULL;
	}

	Py_RETURN_NONE;
} // Viewer_font

/* --- Viewer_grid --- */
static PyObject* Viewer_grid(ViewerObject* self, PyObject* args)
{
	char* name;
	PyObject* value=NULL;
	if (!PyArg_ParseTuple(args, "s|O", &name, &value)) return NULL;

	if (!strcmp(name,"label")) {
		if (value==NULL) {
			char ul[2], vl[2];
			ul[0] = self->viewer->decoration.gridU;
			ul[1] = 0;
			vl[0] = self->viewer->decoration.gridV;
			vl[1] = 0;
			return Py_BuildValue("ss", ul, vl);
		} else {
			if (PyTuple_Check(value) && PyTuple_GET_SIZE(value)==2) {
				const char* gu = PyUnicode_AsUTF8(PyTuple_GetItem(value,0));
				const char* gv = PyUnicode_AsUTF8(PyTuple_GetItem(value,1));
				char ua = (gu[0]=='-'? gu[1]:gu[0]);
				char va = (gv[0]=='-'? gv[1]:gv[0]);
				self->viewer->decoration.gridU = ua;
				self->viewer->decoration.gridV = va;
			} else {
				PyErr_SetString(PyExc_TypeError, "tuple expected of size 2");
				return NULL;
			}
		}
	} else
	if (!strcmp(name,"size")) {
		if (value==NULL)
			return Py_BuildValue("ff",
				self->viewer->decoration.grid_du,
				self->viewer->decoration.grid_dv);
		else {
			PyErr_SetString(PyExc_SyntaxError, "cannot set grid size");
			return NULL;
		}
	} else
	if (!strcmp(name,"low")) {
		if (value==NULL)
			return Py_BuildValue("ff",
				self->viewer->decoration.grid_u,
				self->viewer->decoration.grid_v);
		else {
			PyErr_SetString(PyExc_SyntaxError, "cannot set grid low value");
			return NULL;
		}
	} else
	if (!strcmp(name,"axes")) {
		if (value) {
			PyErr_SetString(PyExc_SyntaxError, "cannot set grid axes value");
			return NULL;
		}

		// Grid axes direction as tuple of strings ([-][XYZUV], [-][XYZUV])
		const Matrix4& matrix = self->kernel->view.matrix();
		Vector u(matrix(0,0), matrix(1,0), matrix(2,0));
		Vector v(matrix(0,1), matrix(1,1), matrix(2,1));
		const char *ustr = "U";
		const double accuracy = 1E-6;
		if (Eq(u.x, 1.0, accuracy))
			ustr = "X";
		else
		if (Eq(u.x, -1.0, accuracy))
			ustr = "-X";

		if (Eq(u.y, 1.0, accuracy))
			ustr = "Y";
		else
		if (Eq(u.y, -1.0, accuracy))
			ustr = "-Y";

		if (Eq(u.z, 1.0, accuracy))
			ustr = "Z";
		else
		if (Eq(u.z, -1.0, accuracy))
			ustr = "-Z";

		const char *vstr = "V";
		if (Eq(v.x, 1.0, accuracy))
			vstr = "X";
		else
		if (Eq(v.x, -1.0, accuracy))
			vstr = "-X";

		if (Eq(v.y, 1.0, accuracy))
			vstr = "Y";
		else
		if (Eq(v.y, -1.0, accuracy))
			vstr = "-Y";

		if (Eq(v.z, 1.0, accuracy))
			vstr = "Z";
		else
		if (Eq(v.z, -1.0, accuracy))
			vstr = "-Z";

		return Py_BuildValue("ss", ustr, vstr);
	} else {
		PyErr_Format(PyExc_SyntaxError, "'%s' is not a valid type option", name);
		return NULL;
	}
	Py_RETURN_NONE;
} // Viewer_grid

/* --- Viewer_hit ---
 * @param u,v		of ray direction
 * @param options	optional B(ody), R(egion), T(ype)
 * @returns hit position or a tuple with hit and options
 */
static PyObject* Viewer_hit(ViewerObject* self, PyObject* args)
{
	static const char *regType[] = {"NORMAL", "BLCKHOLE", "LATTICE", "VOXEL"};

	double u, v;
	char* options=NULL;
	if (!PyArg_ParseTuple(args, "dd|s", &u, &v, &options)) return NULL;

	double x, y, z;
	self->kernel->view.rayPosition(u,v, &x,&y,&z);

	double dx, dy, dz;
	self->kernel->view.rayDirection(u,v, &dx,&dy,&dz);

	self->kernel->lock();
	self->kernel->engine()->incBodyCheckId();
	VZone* zone = self->kernel->engine()->whereRay(x,y,z, dx,dy,dz, SMALL3D, NULL);

	// prepare the ray
	Ray ray;
	ray.use_clip      = self->kernel->clipBodyCount() > 0;
	ray.skip_1stblack = self->viewer->d3.skip_1stblack;
	ray.push(RaySegment(x,y,z, dx,dy,dz, zone));

	self->kernel->engine()->incBodyCheckId();
	self->viewer->d3.nextIntersection(self->kernel->engine(), &ray);
	self->kernel->unlock();

//	if (ray.hitZone() && ray.hitRegion()->type() == REGION_BLACKHOLE)
//		Py_RETURN_NONE;

	Point hit = ray.hit();
	PyObject* pyhit = Py_BuildValue("ddd", hit.x, hit.y, hit.z);
	if (options==NULL) return pyhit;

	PyObject* ret = PyTuple_New(1+strlen(options));
	PyTuple_SetItem(ret, 0, pyhit);

	// B, R, T
	for (size_t i=0; i<strlen(options); i++) {
		PyObject* item = NULL;
		switch (options[i]) {
			case 'b':
			case 'B':
				if (ray.hitBody())
					item = PyUnicode_FromString(ray.hitBody()->cname());
				break;
			case 'r':
			case 'R':
				if (ray.hitZone())
					item = PyUnicode_FromString(ray.hitRegion()->cname());
				break;
			case 't':
			case 'T':
				if (ray.hitZone())
					item = PyUnicode_FromString(regType[(int)ray.hitRegion()->type()]);
				break;
		}
		if (item)
			PyTuple_SetItem(ret, i+1, item);
		else {
			Py_INCREF(Py_None);
			PyTuple_SetItem(ret, i+1, Py_None);
		}
	}
	return ret;
} // Viewer_hit

/* --- Viewer_image --- */
static PyObject* Viewer_image(ViewerObject* self, PyObject* args)
{
	char* name;
	PyObject* value=NULL;
	const char* option=NULL;

	if (!PyArg_ParseTuple(args, "s|Oz", &name, &value, &option)) return NULL;

	if (!strcmp(name,"get")) { // Return Viewport image data
		int size = self->viewer->painter.width() * self->viewer->painter.height();
		if (option==NULL or !strcmp(option, "image")) {
			dword* pixel = self->viewer->painter.data();
			for (int i=0; i<size; i++, pixel++) {
				// Convert from ?RGB to ABGR
				Color32 p;
				p = *pixel;
				// PIL format is ABGR
				*pixel = (*pixel&FLAG_TRANSPARENT?0:0xFF000000)	// Alpha
					| ((dword)p.blue()<<16)
					| ((dword)p.green()<<8)
					|  (dword)p.red();
			}
			return PyBytes_FromStringAndSize((const char *)self->viewer->painter.data(),
					self->viewer->painter.width()*
					self->viewer->painter.height()*4);
		} else
		if (!strcmp(option, "win")) {
			XImage* ximg = XGetImage(self->display, self->pixmap,
						0, 0,
						Tk_Width(self->tkWin),
						Tk_Height(self->tkWin),
						AllPlanes,
						ZPixmap);
//			size_t red_mask   = ximg->red_mask;
//			size_t green_mask = ximg->green_mask;
//			size_t blue_mask  = ximg->blue_mask;
			size_t red_mask   = 0xFF0000;
			size_t green_mask = 0x00FF00;
			size_t blue_mask  = 0x0000FF;

			PyObject* pybytes = PyBytes_FromStringAndSize(NULL, size*4);
			byte* bytes = (byte*)PyBytes_AS_STRING(pybytes);
			for (int y = 0; y < Tk_Height(self->tkWin) ; y++)
				for (int x = 0; x < Tk_Width(self->tkWin); x++) {
					dword xpixel = XGetPixel(ximg, x, y);
					*bytes++ = (xpixel & red_mask)   >> 16;	// red
					*bytes++ = (xpixel & green_mask) >>  8;	// green
					*bytes++ =  xpixel & blue_mask;		// blue
					*bytes++ = 0xFF;			// alpha
			}
			return pybytes;
		} else {
			PyErr_SetString(PyExc_SyntaxError, "invalid option specified");
			return NULL;
		}
	} else
	if (!strcmp(name,"restore")) {	// restore viewport image from previous byte swapping
		dword* pixel = self->viewer->painter.data();
		int size = Tk_Width(self->tkWin)*Tk_Height(self->tkWin);
		for (int i=0; i<size; i++, pixel++) {
			// Convert from ?RGB to ABGR
			Color32 p;
			p = *pixel;
			// PIL format is ABGR
			*pixel =   ((dword)p.blue()<<16)
				 | ((dword)p.green()<<8)
				 |  (dword)p.red();
		}
	} else
	if (!strcmp(name,"data")) {	// set image data
		if (!PyBytes_Check(value)) {
			PyErr_SetString(PyExc_SyntaxError, "bytes type was expected for image data");
			return NULL;
		}
		dword* data = reinterpret_cast<dword*>(PyBytes_AsString(value));
		size_t size  = PyBytes_Size(value);
		int s = (int)(size/sizeof(dword));
		dword* pixel = data;
		for (int i=0; i<s; i++, pixel++) {
			// Convert from ?RGB to ABGR
			Color32 p;
			p = *pixel;
			// PIL format is ABGR
			*pixel =   ((dword)p.blue()<<16)
				 | ((dword)p.green()<<8)
				 |  (dword)p.red();
		}
		//if (!self->viewer->image.data(data, size)) {
		if (!self->viewer->image.data(data, 0)) {
			PyErr_SetString(PyExc_MemoryError, "not enough memory to load image");
			return NULL;
		}
	} else
	if (!strcmp(name,"size")) {	// set image size
		if (!PyTuple_Check(value) || PyTuple_Size(value)!=2) {
			PyErr_SetString(PyExc_SyntaxError, "tuple (w,h) was expected for image size");
			return NULL;
		}
		int w = PyLong_AsLong(PyTuple_GetItem(value,0));
		int h = PyLong_AsLong(PyTuple_GetItem(value,1));
		self->viewer->image.size(w,h);
	} else
	if (!strcmp(name,"matrix")) {	// set transformation matrix
		Matrix3 R,M;
		if (!PyTuple_Check(value) || PyTuple_Size(value)!=2) {
			PyErr_SetString(PyExc_SyntaxError, "tuple (R,M) was expected for transformation matrix");
			return NULL;
		}
		PyList_AsMatrix3(PyTuple_GetItem(value,0), R);
		PyList_AsMatrix3(PyTuple_GetItem(value,1), M);
		self->viewer->image.matrix(&R, &M);
	} else
	if (!strcmp(name,"alpha")) {	// set image transparency
		self->viewer->image.alpha(PyLong_AsLong(value));
	} else
	if (!strcmp(name,"level")) {	// adjust color level
		dword low=0x000000;
		dword high=0xffffff;

		if (!PyTuple_Check(value) || PyTuple_GET_SIZE(value)!=2) {
			PyErr_SetString(PyExc_TypeError, "Tuple expected with color limits");
			return NULL;
		}
		low  = (dword)PyLong_AsUnsignedLongMask(PyTuple_GetItem(value,0));
		high = (dword)PyLong_AsUnsignedLongMask(PyTuple_GetItem(value,1));
		self->viewer->image.colorRange(low, high);
	} else {
		PyErr_Format(PyExc_SyntaxError, "'%s' is not a valid option", name);
		return NULL;
	}

	Py_RETURN_NONE;
} // Viewer_image

/* --- Viewer_invalid --- */
static PyObject* Viewer_invalid(ViewerObject* self, PyObject* Py_UNUSED(ignored))
{
	if (self->kernel->view.invalidWindow() or
	   (self->viewer->d2.show and !self->viewer->d2.isValid()))
		Py_RETURN_TRUE;
	else
		Py_RETURN_FALSE;
} // Viewer_invalid

#if 0
/* --- Viewer_lock --- */
static PyObject* Viewer_lock(ViewerObject* self)
{
	self->viewer->lockState();
	Py_RETURN_NONE;
} // Viewer_lock

/* --- Viewer_unlock --- */
static PyObject* Viewer_unlock(ViewerObject* self)
{
	self->viewer->unlockState();
	Py_RETURN_NONE;
} // Viewer_unlock
#endif

/* --- Viewer_matrix ---
 * @param matrix	(3x3/4x4) or integer
 * @param  -4	to return the inverse matrix 4x4
 * @param  -3	to return the inverse matrix 3x3
 * @param  -1	to return the inverse matrix 4x4
 * @param   3	to return the normal matrix 3x3
 * @param   4	to return the normal matrix 4x4
 * @param 3x3, 4x4	to set the viewport matrix
 * @return	return transformation matrix or the inverse if @param=-1
 */
static PyObject* Viewer_matrix(ViewerObject* self, PyObject* args)
{
	PyObject* matrixObj=NULL;
	if (!PyArg_ParseTuple(args, "|O", &matrixObj)) return NULL;

	if (matrixObj != NULL) {
		if (PyLong_Check(matrixObj))
			switch (PyLong_AsLong(matrixObj)) {
				case 3:		// Return the normal matrix as 3x3
					return PyList_FromMatrix3(self->kernel->view.matrix());
				case -3:	// Return the inverse matrix as 3x3
					return PyList_FromMatrix3(self->kernel->view.invMatrix());
				case -1:	// Return the inverse matrix
				case -4:
					return PyList_FromMatrix4(self->kernel->view.invMatrix());
				default:	// Return the normal matrix
					return PyList_FromMatrix4(self->kernel->view.matrix());
			}
		else
		if (PyList_Check(matrixObj) and PyList_Size(matrixObj)==4) {
			// Set matrix 4x4
			Matrix4 matrix;
			PyList_AsMatrix4(matrixObj, matrix);

			self->projectionChanged = false;
			for (int j=0; j<matrix.rows() and !self->projectionChanged; j++)
				for (int i=0; i<matrix.columns(); i++)
					if (!Eq(self->kernel->view.matrix()(i,j), matrix(i,j), 1e-10)) {
						self->projectionChanged = true;
						break;
					}
			self->viewer->matrix(matrix);
		} else
		if (PyList_Check(matrixObj) and PyList_Size(matrixObj)==3) {
			// Set matrix 3x3
			Matrix3 matrix;
			PyList_AsMatrix3(matrixObj, matrix);

			self->projectionChanged = false;
			for (int j=0; j<matrix.rows() and !self->projectionChanged; j++)
				for (int i=0; i<matrix.columns(); i++)
					if (!Eq(self->kernel->view.matrix()(i,j), matrix(i,j), 1e-10)) {
						self->projectionChanged = true;
						break;
					}
			self->viewer->matrix(matrix);
		} else {
			PyErr_SetString(PyExc_TypeError, "Invalid viewer matrix");
			return NULL;
		}
	} else	// return the normal matrix
		return PyList_FromMatrix4(self->kernel->view.matrix());
	Py_RETURN_NONE;
} // Viewer_matrix

/* --- Viewer_memory --- */
static PyObject* Viewer_memory(ViewerObject* self, PyObject* args)
{
	char* dump=NULL;
	if (!PyArg_ParseTuple(args, "|s", &dump)) return NULL;

	if (dump==NULL)
		return PyLong_FromLong(self->viewer->memory());
	else {
		self->viewer->printMemory();
		self->kernel->engine()->printMemory();
	}
	Py_RETURN_NONE;
} // Viewer_memory

/* --- Viewer_moveUV --- *
 * using an old coordinate (x,y,z) move to the new (xn,yn,zn) but keeping the w
 * the same. Used when moving objects in the viewports
 */
static PyObject* Viewer_moveUV(ViewerObject* self, PyObject* args)
{
	double x,y,z;
	double xn,yn,zn;

	if (!PyArg_ParseTuple(args, "(ddd)(ddd)", &x, &y, &z, &xn, &yn, &zn)) return NULL;

	double u = self->kernel->view.xyz2u(xn,yn,zn);
	double v = self->kernel->view.xyz2v(xn,yn,zn);
	double w = self->kernel->view.xyz2w(x,y,z);

	return Py_BuildValue("ddd",	self->kernel->view.uvw2x(u,v,w),
					self->kernel->view.uvw2y(u,v,w),
					self->kernel->view.uvw2z(u,v,w));
} // Viewer_moveUV

/* --- Viewer_ocube --- */
static PyObject* Viewer_ocube(ViewerObject* self, PyObject* args)
{
	char* cmd;
	PyObject* value  = NULL;
	PyObject* value2 = NULL;
	if (!PyArg_ParseTuple(args, "s|OO", &cmd, &value, &value2)) return NULL;

	if (!strcmp(cmd, "length")) {
		if (value==NULL)
			return PyLong_FromLong(self->ocube->length);
		else
			self->ocube->length = PyLong_AsLong(value);
	} else
	if (!strcmp(cmd, "detail")) {
		if (value==NULL)
			return PyBool_FromLong(self->ocube->detail);
		else
			self->ocube->detail = PyObject_IsTrue(value);
	} else
	if (!strcmp(cmd, "mouse")) {
		if (value==NULL)
			return Py_BuildValue("ii",
					self->ocube->xmouse,
					self->ocube->ymouse);
		else {
			self->ocube->xmouse = PyLong_AsLong(PyTuple_GetItem(value,0));
			self->ocube->ymouse = PyLong_AsLong(PyTuple_GetItem(value,1));
		}
	} else
	if (!strcmp(cmd, "selection")) {
		if (value==NULL)
			return Py_BuildValue("iiii",
					self->ocube->imouse,
					self->ocube->icube[0],
					self->ocube->icube[1],
					self->ocube->icube[2]);
		else {
			PyErr_SetString(PyExc_TypeError, "Called ocube with selection and value");
			return NULL;
		}
	} else
	if (!strcmp(cmd, "off"))
		self->ocube->length = -Abs(self->ocube->length);
	else
	if (!strcmp(cmd, "on"))
		self->ocube->length = Abs(self->ocube->length);
	else
	if (!strcmp(cmd, "facename")) {
		if (PyLong_Check(value) and PyUnicode_Check(value2)) {
			int face = PyLong_AsLong(value);
			//if (face>0 and face<6)
			const char* name = PyUnicode_AsUTF8(value2);
			self->ocube->facename(face, name);
		} else {
			PyErr_SetString(PyExc_ValueError,
				"Incorrect values for ocube facename");
			return NULL;
		}
	} else {
		PyErr_Format(PyExc_TypeError, "Invalid type \'%s\' specified", cmd);
		return NULL;
	}

	Py_RETURN_NONE;
} // Viewer_ocube

/* --- Viewer_origin --- */
static PyObject* Viewer_origin(ViewerObject* self, PyObject* args)
{
	// Get current origin, we will needed to check if projection is changed
	double x, y, z;
	self->viewer->origin(&x,&y,&z);

	if (PyTuple_Size(args)==0)
		return Py_BuildValue("ddd", x, y, z);

	// set center accept a tuple/list of length 3 or 3 arguments
	if (PyTuple_Size(args)==3)
		// 3 arguments passed
		self->viewer->origin(
			PyFloat_AsDouble(PyTuple_GetItem(args,0)),
			PyFloat_AsDouble(PyTuple_GetItem(args,1)),
			PyFloat_AsDouble(PyTuple_GetItem(args,2)));
	else
	if (PyTuple_Size(args)==1) {
		// 1 argument passed
		PyObject* xyz=PyTuple_GetItem(args,0);
		if (PyTuple_Check(xyz) && PyTuple_Size(xyz)==3)
			self->viewer->origin(
				PyFloat_AsDouble(PyTuple_GetItem(xyz,0)),
				PyFloat_AsDouble(PyTuple_GetItem(xyz,1)),
				PyFloat_AsDouble(PyTuple_GetItem(xyz,2)));
		else
		if (PyList_Check(xyz) && PyList_Size(xyz)==3) {
			self->viewer->origin(
				PyFloat_AsDouble(PyList_GetItem(xyz,0)),
				PyFloat_AsDouble(PyList_GetItem(xyz,1)),
				PyFloat_AsDouble(PyList_GetItem(xyz,2)));
		} else {
			PyErr_SetString(PyExc_TypeError,
				"function takes exactly 1 tuple of size 3 or 3 arguments");
			return NULL;
		}
	} else {
		PyErr_SetString(PyExc_TypeError,
				"function takes exactly 1 tuple of size 3 or 3 arguments");
		return NULL;
	}

	double nx, ny, nz;
	self->viewer->origin(&nx, &ny, &nz);
	if (!Eq(x,nx,1e-10) or !Eq(y,ny,1e-10) or !Eq(z,nz,1e-10))
		self->projectionChanged = true;
	Py_RETURN_NONE;
} // Viewer_origin

/* --- Viewer_offset --- */
static PyObject* Viewer_offset(ViewerObject* self, PyObject* args)
{
	// Get current offset, we will needed to check if projection is changed
	double u = self->kernel->view.Uofs();
	double v = self->kernel->view.Vofs();
	if (PyTuple_Size(args)==0)
		return Py_BuildValue("dd", u, v);

	// set center accept a tuple of length 3 or 3 arguments
	PyObject* uv;

	if (PyTuple_Size(args)==2) {
		self->kernel->view.offset(
				PyFloat_AsDouble(PyTuple_GetItem(args,0)),
				PyFloat_AsDouble(PyTuple_GetItem(args,1)));
	} else
	if (PyTuple_Size(args)==1 && PyTuple_Check(uv=PyTuple_GetItem(args,0))
				  && PyTuple_Size(uv)==2) {
		self->kernel->view.offset(
				PyFloat_AsDouble(PyTuple_GetItem(uv,0)),
				PyFloat_AsDouble(PyTuple_GetItem(uv,1)));
	} else {
		PyErr_SetString(PyExc_TypeError,
				"function takes exactly 1 tuple of size 2 or 2 arguments");
		return NULL;
	}
	if (!Eq(u, self->kernel->view.Uofs(), 1e-10) or
	    !Eq(v, self->kernel->view.Vofs(), 1e-10))
		self->projectionChanged = true;
	Py_RETURN_NONE;
} // Viewer_offset

/* --- Viewer_objectVar --- */
static PyObject* Viewer_objectVar(ViewerObject* self, GObject* object,
				const char* var, PyObject* value)
{
	if (value==NULL) {
		PyErr_SetString(PyExc_TypeError, "object 'move'/'rotate' doesn't return anything.");
		return NULL;
	} else
	if (!strcmp(var, "move")) {		// Move relative from the last save position
		int opt = PyLong_AsLong(value);	// What to move
		//self->viewer->lock();
		object->node(opt, object->savedNode(opt) + self->move);
		//self->viewer->unlock();
	} else
	if (!strcmp(var, "rotate")) {	// rotate around an axis (axis,angle)
		//self._viewer.body(bodies, "rotate", (uv,  self._axis, snap))
		if (!PyTuple_Check(value) || PyTuple_GET_SIZE(value)!=3) {
			PyErr_SetString(PyExc_TypeError, "tuple of size 3 expected");
			return NULL;
		}

		// What to move
		double u,v;
		if (!Py_GetUV(PyTuple_GetItem(value, 0), &u, &v)) return NULL;
		Vector axis = Py_GetVector(PyTuple_GetItem(value, 1));
		int snap = PyLong_AsLong(PyTuple_GetItem(value, 2));
		if (PyErr_Occurred()) return NULL;

		// Calculate angle
		double angle = atan2(v-self->pivotV, u-self->pivotU) - self->pivotAngle;
		if (snap) // FIXME dangle should be past from python
			angle = (double)Round(angle / self->geometry->snapAngle)
				* self->geometry->snapAngle;

		object->rotate(angle,axis);
	} else {
		PyErr_Format(PyExc_TypeError, "Invalid type \'%s\' specified", var);
		return NULL;
	}
	Py_RETURN_NONE;
} // Viewer_objectVar

/* --- Viewer_object --- */
static PyObject* Viewer_object(ViewerObject* self, PyObject* args)
{
	PyObject* obj;
	char const *var = (char const *)"id";
	PyObject* value=NULL;
	if (!PyArg_ParseTuple(args, "O|sO", &obj, &var, &value)) return NULL;

	if (Py_Check4Pattern(obj)) {
		const char* pattern = PyUnicode_AsUTF8(obj);
		for (auto object : self->geometry->objects) {
			if (!fnmatch(pattern, object->cname(), 0)) {
				PyObject* ret = Viewer_objectVar(self, object, var, value);
				if (ret == NULL) return NULL;
				Py_XDECREF(ret);	// ret can be NULL use XDECREF
			}
		}
		Py_RETURN_NONE;
	} else
	if (PyList_Check(obj)) {
		for (ssize_t i=0; i<PyList_GET_SIZE(obj); i++) {
			GObject* object = Py_Object(self->geometry, PyList_GetItem(obj,i));
			if (object==NULL) return NULL;
			PyObject* ret = Viewer_objectVar(self, object, var, value);
			if (ret == NULL) return NULL;
			Py_XDECREF(ret);	// ret can be NULL use XDECREF
		}
		Py_RETURN_NONE;
	} else
	if (PyTuple_Check(obj)) {
		for (ssize_t i=0; i<PyTuple_GET_SIZE(obj); i++) {
			GObject* object = Py_Object(self->geometry, PyTuple_GetItem(obj,i));
			if (object==NULL) return NULL;
			PyObject* ret = Viewer_objectVar(self, object, var, value);
			if (ret == NULL) return NULL;
			Py_XDECREF(ret);	// ret can be NULL use XDECREF
		}
		Py_RETURN_NONE;
	} else {
		GObject* object = Py_Object(self->geometry, obj);
		if (object)
			return Viewer_objectVar(self, object, var, value);
	}
	return NULL;
} // Viewer_object

/* --- Viewer_pixel2uv --- */
static PyObject* Viewer_pixel2uv(ViewerObject* self, PyObject* args)
{
	int i,j;
	int centered=0;

	if (!PyArg_ParseTuple(args, "ii|i", &i, &j, &centered)) return NULL;

	if (centered)
		// WARNING: the uv of the viewer is always centered on the window
		// while on the viewer could be shifted...
		return Py_BuildValue("dd",	self->kernel->view.i2u(i)-self->kernel->view.Uofs(),
						self->kernel->view.j2v(j)-self->kernel->view.Vofs());
	else
		return Py_BuildValue("dd",	self->kernel->view.i2u(i),
						self->kernel->view.j2v(j));
} // Viewer_pixel2uv

/* --- Viewer_pixel2xyz --- */
static PyObject* Viewer_pixel2xyz(ViewerObject* self, PyObject* args)
{
	int i,j;

	if (!PyArg_ParseTuple(args, "ii", &i, &j)) return NULL;

	double x, y, z;
	self->kernel->view.ij2xyz(i,j,&x,&y,&z);

	return Py_BuildValue("ddd", x,y,z);
} // Viewer_pixel2xyz

/* --- Viewer_duv2dxyz --- */
static PyObject* Viewer_duv2dxyz(ViewerObject* self, PyObject* args)
{
	double du,dv;

	if (!PyArg_ParseTuple(args, "dd", &du, &dv)) return NULL;

	double dx = self->kernel->view.uv2dx(du,dv);
	double dy = self->kernel->view.uv2dy(du,dv);
	double dz = self->kernel->view.uv2dz(du,dv);

	return Py_BuildValue("ddd", dx,dy,dz);
} // Viewer_duv2dxyz

/* --- Viewer_palette --- */
static PyObject* Viewer_palette(ViewerObject* self, PyObject* args)
{
	int	 id;
	char*	name;
	PyObject* value=NULL;

	if (!PyArg_ParseTuple(args, "is|O", &id, &name, &value)) return NULL;

	if (id<0 || id>= MAXPALETTE) {
		PyErr_SetString(PyExc_ValueError,"Invalid palette index");
		return NULL;
	}

	if (!strcmp(name,"alphamin")) {
		if (value)
			self->viewer->palette[id].alphamin(PyObject_IsTrue(value));
		else
			return PyBool_FromLong(self->viewer->palette[id].alphamin());
	} else
	if (!strcmp(name,"alphamax")) {
		if (value)
			self->viewer->palette[id].alphamax(PyObject_IsTrue(value));
		else
			return PyBool_FromLong(self->viewer->palette[id].alphamax());
	} else
	if (!strcmp(name,"default")) {
		if (value)
			self->viewer->palette.def(id);
		else
			return PyBool_FromLong(self->viewer->palette.def());
	} else
	if (!strcmp(name,"interpolate")) {
		if (value)
			self->viewer->palette[id].interpolate(PyObject_IsTrue(value));
		else
			return PyBool_FromLong(self->viewer->palette[id].interpolate());
	} else
	if (!strcmp(name,"invert")) {
		if (value)
			self->viewer->palette[id].invert(PyObject_IsTrue(value));
		else
			return PyBool_FromLong(self->viewer->palette[id].invert());
	} else
	if (!strcmp(name,"label")) {
		if (value)
			self->viewer->palette.label(id, PyUnicode_AsUTF8(value));
		else
			return PyUnicode_FromString(self->viewer->palette.label(id).c_str());
	} else
	if (!strcmp(name,"log")) {
		if (value)
			self->viewer->palette[id].log(PyObject_IsTrue(value));
		else
			return PyBool_FromLong(self->viewer->palette[id].log());
	} else
	if (!strcmp(name,"max")) {
		if (value)
			self->viewer->palette[id].max(PyFloat_AsDouble(value));
		else
			return PyFloat_FromDouble(self->viewer->palette[id].max());
	} else
	if (!strcmp(name,"min")) {
		if (value)
			self->viewer->palette[id].min(PyFloat_AsDouble(value));
		else
			return PyFloat_FromDouble(self->viewer->palette[id].min());
	} else
	if (!strcmp(name,"palette")) {
		if (value) {
			if (!PyList_Check(value)) {
				PyErr_SetString(PyExc_TypeError, "Invalid type, list expected");
				return NULL;
			}
			size_t n = PyList_Size(value);
			if (n>_MAXCOLORS) {
				PyErr_SetString(PyExc_TypeError, "Maximum number of colors accepted is 256");
				return NULL;
			}
			self->viewer->palette[id].size((int)n);
			for (size_t i=0; i<n; i++)
				self->viewer->palette[id][i] = \
					(dword)PyLong_AsUnsignedLongMask(PyList_GetItem(value, i));
		}
	} else
	if (!strcmp(name,"reset")) {
		self->viewer->palette[id].reset();
		self->viewer->palette.display(id, false);
	} else
	if (!strcmp(name,"show")) {
		if (value)
			self->viewer->palette.display(id, PyObject_IsTrue(value));
		else
			return PyBool_FromLong(self->viewer->palette.display(id));
	} else
	if (!strcmp(name,"smooth")) {
		if (value)
			self->viewer->palette[id].interpolate(PyObject_IsTrue(value));
		else
			return PyBool_FromLong(self->viewer->palette[id].interpolate());
	} else {
		// Error
		PyErr_Format(PyExc_SyntaxError, "'%s' is not a valid type option", name);
		return NULL;
	}
	self->viewer->palette[id].init();	// calc palette

	Py_RETURN_NONE;
} // Viewer_palette

/** Viewer_pen
 * process hand drawn array 'pen'
 */
static PyObject* Viewer_pen(ViewerObject* self, PyObject* args)
{
	char*	name;
	PyObject* value=NULL;

	if (!PyArg_ParseTuple(args, "s|O", &name, &value)) return NULL;

	if (!strcmp(name,"add")) {
		// expect pixel (i,j) coordinates
		if (!PyTuple_Check(value) || PyTuple_GET_SIZE(value)!=2) {
			PyErr_SetString(PyExc_TypeError, "pixel tuple of size 2 expected");
			return NULL;
		}
		self->pen.push_back(IPoint(PyLong_AsLong(PyTuple_GetItem(value,0)),
					PyLong_AsLong(PyTuple_GetItem(value,1))));
	} else
	if (!strcmp(name,"clear")) {
		self->pen.clear();
	} else {
		// Error
		PyErr_Format(PyExc_SyntaxError, "'%s' is not a valid type option", name);
		return NULL;
	}

	Py_RETURN_NONE;
} // Viewer_pen

/** Viewer_planner
 * prepare a report on dose integral of a usrbin along a path
 */
static PyObject* Viewer_planner(ViewerObject* self, PyObject* args)
{
	char*	filename = nullptr;
	int	splineid=-1;
	int	Nsteps=0;
	double	tscale=1.0;

	if (!PyArg_ParseTuple(args, "si|di", &filename, &splineid, &tscale, &Nsteps)) return NULL;

	// find spline object
	GSpline* splineObj = nullptr;
	for (auto object : self->geometry->objects)
		if (object->id() == splineid and object->isA()==GSplineClass) {
			splineObj = static_cast<GSpline*>(object);
			break;
		}
	if (splineObj == nullptr) {
		PyErr_Format(PyExc_IndexError, "spline %d not found", splineid);
		return NULL;
	}

	double	path_total_length = 0;	// cumulative length of the whole path
	double	path_total_dose = 0;	// total dose of the whole intervention defined by the path
	double  sum  = 0.0;
	double  time = 0.0;

	struct Record {			// data structure to record data per integration step
		int	node;		// node id
		double	time;		// start time of step
		double	time_step;	// size of time step
		Point	pos;		// current position
		double	length;		// step length
		double	speed;		// instantaneous speed
		double	dose_per_step;	// cumulative dose for the current record
		double	max_dose_rate;	// maximum dose rate encountered

		Record() : node(0),
			time(0),
			time_step(0),
			length(0),
			speed(0),
			dose_per_step(0), max_dose_rate(0) {};

		void reset() {
			node = 0;
			time = time_step = length = speed = dose_per_step = max_dose_rate = 0.0;
			pos.set(0, 0, 0);
		}
	}; // struct Record

	// vector containing all data per integration step
	std::vector<Record>   integralDataPerStep;
	// map containing the collapsed data from all integration
	// sub-steps belonging to one node
	std::map<int, Record> integralDataPerNode;

	if (Nsteps<=0) Nsteps = splineObj->nsegments;
	auto&  spline  = splineObj->getSpline();
	const  Point P = splineObj->node(0);
	Usrbin& usrbin = self->viewer->usrbin[0];

	Record record;
	record.pos = P + (Vector)spline[0];
	integralDataPerStep.push_back(record);

	int n = spline.size();
	bool closed = true;
	if (splineObj->type == Spline_Open or
	    splineObj->type == Spline_Lines) {
		closed = false;
		n--;
	}

	bool lines = splineObj->type == Spline_Lines or
		     splineObj->type == Spline_LinesClosed or
		     Eq0(splineObj->getTension(),1e-10);
	for (int i=0; i<n; i++) {
		record.node = i;
		Point A = P + (Vector)spline[i];

		int in = i+1;
		double tnode = splineObj->getTime(in)*tscale;
#if _DEBUG>1
		cout << "Node: " << i << " t=" << splineObj->getTime(i)*tscale
				<< " i+1:" << in << " t=" << splineObj->getTime(in)*tscale << endl;
#endif
		double Dtime = tnode - time;
		if (Dtime <= 0.0) {
			/* CT: PyErr_Format seems to have issues with %lf formatters that it cannot resolve.
			       Thus, we replace it for the time being. - FIXME
			PyErr_Format(PyExc_ValueError,
					"Node %d has a time stamp (t = %lf) greater "
					"than the following node (t = %lf). "
					"Please assign consecutive absolute times for each node!",
					i+1, tnode, time);
			*/
			char pBuf[256];
			sprintf(pBuf, "Node %d has a time stamp (t = %lf) greater "
					"than the following node (t = %lf). "
					"Please assign consecutive absolute times for each node!",
					i+1, tnode, time );
			PyErr_Format( PyExc_ValueError, pBuf );
			return NULL;
		}
		time = tnode;

		// treat working locations without movement of the actor
		if (in<spline.size() and spline[in].ignore) {
			double curDose = usrbin.integrate(A, A, Dtime);
			if (curDose == -1)  {
				PyErr_Format(PyExc_ValueError,
					"Path outside mesh at point [%d, %d, %d] - no valid data found!",
					 A.x, A.y, A.z );
				return NULL;
			}
			sum += curDose;
			record.pos       = A;
			record.length    = 0;
			record.time     += Dtime;
			record.time_step = Dtime;
			record.speed     = 0;
			record.dose_per_step = curDose;
			record.max_dose_rate = curDose / Dtime;

			integralDataPerStep.push_back(record);
#if _DEBUG>1
			cout << "\tA=" << A << " A=" << A << endl;
#endif
		} else
		if (lines) {
			// integrate moving along lines
			Point B;
			if (closed and in==spline.size())
				B = splineObj->node(1);		// 1 based
			else
				B = splineObj->node(in+1);	// 1 based

			double curDose = usrbin.integrate(A, B, Dtime);
			if (curDose == -1)  {
				PyErr_Format(PyExc_ValueError,
					"Path outside mesh between [%d, %d, %d] -> "
					"[%d, %d, %d]- no valid data found!",
					A.x, A.y, A.z, B.x, B.y, B.z );
				return NULL;
			}
			sum += curDose;
			double dL = (B-A).length();
			record.time       += Dtime;
			record.time_step   = Dtime;
			record.pos         = B;
			record.length      = dL;
			record.speed       = dL / Dtime;
			record.dose_per_step = curDose;
			record.max_dose_rate = curDose / Dtime;

			integralDataPerStep.push_back(record);
			path_total_length += dL;
#if _DEBUG>1
			cout << "\tA=" << A << " B=" << B << endl;
#endif
			A = B;
		} else {
			// integrate moving along a curve
			const double step = 1.0/(double)Nsteps;
			double dt = Dtime / (double)Nsteps;
			double s=step;
			for (int j=0; j<Nsteps; j++, s+=step) {
				Point B = P + spline(i,s);
				double curDose = usrbin.integrate(A, B, dt);
				if (curDose == -1) {
					PyErr_Format( PyExc_ValueError,
						"Path outside mesh between [%d, %d, %d] "
						"-> [%d, %d, %d]- no valid data found!",
						A.x, A.y, A.z, B.x, B.y, B.z );
					return NULL;
				}
				sum += curDose;
				double dL = (B-A).length();
				record.time       += dt;
				record.time_step   = dt;
				record.pos         = B;
				record.length      = dL;
				record.speed       = dL / dt;
				record.dose_per_step = curDose;
				record.max_dose_rate = curDose / dt;
				integralDataPerStep.push_back(record);
				path_total_length += dL;
#if _DEBUG>1
				cout << "\tA=" << A << " B=" << B << endl;
#endif
				A = B;
			}
		}
	}
	path_total_dose = sum;

	//
	// collapse data from sub-steps to data per node
	//
	int prevNode = 0, curNode = 0;
	Record curRecord, nodeRecord;
	for (size_t i=0; i<integralDataPerStep.size(); ++i) {
		curNode = integralDataPerStep[i].node;
		if (curNode != prevNode) {
			nodeRecord.speed  = nodeRecord.length / nodeRecord.time_step;
			integralDataPerNode[nodeRecord.node] = nodeRecord;
			nodeRecord.reset();
			prevNode = curNode;
		}
		nodeRecord.node = curNode;
		if (nodeRecord.length == 0)
			nodeRecord.pos = integralDataPerStep[i].pos;

		nodeRecord.time_step     += integralDataPerStep[i].time_step;
		nodeRecord.time           = integralDataPerStep[i].time;
		nodeRecord.dose_per_step += integralDataPerStep[i].dose_per_step;
		nodeRecord.length        += integralDataPerStep[i].length;
		nodeRecord.max_dose_rate = std::max( nodeRecord.max_dose_rate, integralDataPerStep[i].max_dose_rate );
	}
	nodeRecord.speed = nodeRecord.length / nodeRecord.time_step;
	integralDataPerNode[nodeRecord.node] = nodeRecord;

	// analyze cumulative dose while moving vs. those while standing
	double doseMoving = 0.0, doseWorking = 0.0;
	for (auto iter=integralDataPerNode.begin(); iter != integralDataPerNode.end(); ++iter) {
		if (iter->second.speed == 0)
			doseWorking += iter->second.dose_per_step;
		else
			doseMoving += iter->second.dose_per_step;
	}

	std::ofstream fout(filename);
	fout << "###############################################################################" << endl;
	fout << "#" << endl;
	fout << "# /!\\ For graphical representation please load this file in GNUPLOT > 5.0 /!\\" << endl;
	fout << "#" << endl;
	fout << "###############################################################################" << endl;
	fout << "#" << endl;
	fout << "# Planner" << endl;
	auto curtime = ::time(nullptr);
	char timestr[32];
	strftime(timestr, sizeof timestr, "%Y-%m-%d %H-%M-%S", localtime(&curtime));
	fout << "# Report date:              " << timestr << endl;
	//fout << "# Report date:              " << asctime(localtime(&curtime)) << endl;
	//fout << "# Report date:              " << std::put_time(&curLocalTime, "%d-%m-%Y %H-%M-%S") << endl;
	fout << "# Workflow :                " << splineObj->name() << endl;
	fout << "# Usrbin File:              " << usrbin.filename << endl;
	fout << "# Name:                     " << usrbin.name << " [" << usrbin.detector() << "]" << endl;
	fout << "# Norm:                     " << usrbin.norm() << endl;
	fout << "# Nodes:                    " << spline.size() << endl;
	fout << "# Total time:               " << time / 60.0 <<  " min" << endl;
	fout << "# Total length:             " << path_total_length / 100.0 <<  " m" << endl;
	fout << "# Integral dose:            " << path_total_dose << endl;
	fout << "# Total dose while moving:  " << doseMoving  << " (" << doseMoving / sum * 100.0 << "%)" << endl;
	fout << "# Total dose while working: " << doseWorking << " (" << doseWorking / sum * 100.0 << "%)" << endl;
	fout << "#" << endl << "#" << endl;

	/*
	fout << "# 1 Cumulative: " << endl;
	for (auto r : integralDataPerStep) {
		fout << r.node << ' '
		     << r.time << ' '
		     << r.pos << ' '
		     << "l = " << r.length << ' '
		     << "v = " << r.speed << ' '
		     << "cD = " << path_total_dose  << ' '
		     << "D/step " << r.dose_per_step << endl;
	}
	fout << endl << endl;
	*/

	fout << "##########################################################################################################################" << endl;
	fout << "#" << endl;
	fout << "#  I N T E R V E N T I O N    R E P O R T" << endl;
	fout << "#" << endl;
	fout << "##########################################################################################################################" << endl;
	fout << "# step     abs. time (s)      Dose           duration (s)   speed  (m/s)		avg. dose rate		max. dose rate		coordinates" << endl;
	fout << "#                                                                          per hour                                       " << endl;
	fout << "#-------------------------------------------------------------------------------------------------------------------------" << endl;

	int nodeCtr = 1;
	for (auto iter = integralDataPerNode.begin(); iter != integralDataPerNode.end(); ++iter, nodeCtr++) {
		fout << "# " << std::left << std::setw(10)
		     << std::setprecision(1) << std::fixed << iter->first+1
		     << " " << std::setw(13) << std::fixed << iter->second.time << std::left
		     << " " << std::setw(20) << std::scientific
		     << std::setprecision(4) << iter->second.dose_per_step
		     << " " << std::setw(15) << std::setprecision(1)
		     << std::fixed << iter->second.time_step
		     << " " << std::right << std::scientific
		     << std::setprecision(2) << std::fixed << iter->second.speed / 100
				 << std::scientific << std::setprecision(2)
				 << std::setw(15) << iter->second.dose_per_step / iter->second.time_step
				 << std::setw(15) << iter->second.max_dose_rate * 3600
				 << "\t" << splineObj->node( nodeCtr ) << endl;
	}

	fout<< "# Detector n: " << usrbin.detector() << endl;
	fout << "reset" << endl
		<< "set tmargin 1" << endl
		<< "set bmargin 0" << endl
		<< "set lmargin 10" << endl
		<< "set rmargin 10" << endl
		<< "unset xtics" << endl
		<< "set multiplot layout 4, 1 title \"Dose planning -" << splineObj->name()
		<< "\" font \",14\"" << endl
		<< "set boxwidth 0.5" << endl
		<< "set style fill solid" << endl
		<< "set logscale y" << endl
		<< "set autoscale y" << endl
		<< "set key left autotitle column nobox samplen 1 noenhanced" << endl
		<< "# Step, Dose, % of total, duration( s ), speed( m / s )" << endl
		<< "$data << EOD" << endl;

	for( auto iter = integralDataPerNode.begin(); iter != integralDataPerNode.end(); ++iter ) {
		fout << std::setw( 5 ) << iter->first + 1
			<< std::setw( 15 ) << std::scientific
			<< std::setprecision( 4 ) << iter->second.dose_per_step
			<< " " << std::setw( 15 ) << iter->second.dose_per_step / sum * 100.0
			<< " " << std::setw( 15 ) << iter->second.time_step
			<< " " << std::setw( 15 ) << iter->second.speed / 100.0 << endl;
	}
	fout << "EOD" << endl;
	fout << "set offset graph 0, 0, 0.3, 0.3" << endl
			 << "plot \"$data\" using 1:2 : xtic( 1 ) with boxes title \"Dose at workstep\" "
			 << "lc rgb \"#87221b\",\\" << endl
		   << "\"\"  using 1:($2 * 35) : (sprintf( \"%3.2E\", $2 )) with labels notitle" << endl
			 << "unset logscale y" << endl
			 << "set yrange[0:100]" << endl
			 << "plot \"$data\" using 1 : 3 : xtic( 1 ) with boxes title \"% of total dose\" "
		   << "lc rgb \"#1b7187\"" << endl
			 << "set autoscale y" << endl
			 << "set xtics" << endl
			 << "set xlabel \"Workstep\" font \",12\"" << endl;

	fout << "set label 1 \"Integral dose: " << path_total_dose << "\\nTotal dose while moving: ";
	fout << doseMoving << " (" << doseMoving / sum * 100.0 << "%)\\n";
	fout << "Total dose while working: " << doseWorking << " (" << doseWorking / sum * 100.0 << "%)" << endl;

	fout << "set label 1 at 1.0, -25.0 font \",10\"" << endl
			 << "plot \"$data\" using 1:($4 / 60) : xtic( 1 ) with boxes title \"Duration / min\" "
		   << "lc rgb \"#4d21c4\"" << endl
			 << "unset multiplot" << endl;

	fout.close();

	Py_RETURN_NONE;
} // Viewer_planner

/* _endProjection: notifier function for end of projection */
static void _endProjection(ViewerObject* self)
{
	DUMP(cout << "_endProjection("<<self->viewer->title()<<")" << endl);
	_sendMessage(self, "EndProjection");
} // _endProjection

/* --- Viewer_project --- */
static PyObject* Viewer_project(ViewerObject* self, PyObject* args)
{
	int asThread = false;
	int all = true;
	if (!PyArg_ParseTuple(args, "|ii", &asThread, &all)) return NULL;
	self->viewer->d2.projectAll(all);
	_clearErrors(self);
	DUMP(cout << "Viewer_project("<<self->viewer->title()<<", " << asThread << ")" << endl);
	if (asThread) {
		self->viewer->spawnProject((NotifyFunc)_endProjection, self);
	} else {
		// Stop any pending thread
		DUMP(cout << "Viewer_project((" << self->viewer->title() << ").stopThread()" << endl);
		self->viewer->stopThread();
		// Run synchronously
		self->viewer->d2.project();
	}
	Py_RETURN_NONE;
} // Viewer_project

/* --- Viewer_projectionChanged --- */
static PyObject* Viewer_projectionChanged(ViewerObject* self, PyObject* args)
{
	int flag=-1;

	if (!PyArg_ParseTuple(args, "|i", &flag)) return NULL;
	if (flag == -1)
		return PyBool_FromLong(self->projectionChanged);
	else
		self->projectionChanged = (bool)flag;
	Py_RETURN_NONE;
} // Viewer_projectionChanged

/* --- Viewer_rectangle --- */
static PyObject* Viewer_rectangle(ViewerObject* self, PyObject* args)
{
	if (PyTuple_Size(args)==0) {
			self->rectX1 = -1;
			self->rectY1 = -1;
			self->rectX2 = -1;
			self->rectX2 = -1;
	} else
	if (!PyArg_ParseTuple(args, "iiii",
			&(self->rectX1), &(self->rectY1),
			&(self->rectX2), &(self->rectY2))) return NULL;
	Py_RETURN_NONE;
} // Viewer_rectangle

/* --- Viewer_Sx --- */
static PyObject* Viewer_Sx(ViewerObject* self, PyObject* Py_UNUSED(ignored))
{
	return PyFloat_FromDouble(self->kernel->view.Sx());
} // Viewer_Sx

/* --- Viewer_Sy --- */
static PyObject* Viewer_Sy(ViewerObject* self, PyObject* Py_UNUSED(ignored))
{
	return PyFloat_FromDouble(self->kernel->view.Sy());
} // Viewer_Sy

/* --- Viewer_snap --- */
static PyObject* Viewer_snap(ViewerObject* self, PyObject* args)
{
	double u, v, d=self->geometry->snapDistance;

	if (!PyArg_ParseTuple(args, "dd|d", &u, &v, &d)) return NULL;

	return Py_Vector(_snapUV(self, u, v, d));
} // Viewer_snap

/* --- Viewer_set --- */
static PyObject* Viewer_set(ViewerObject* self, PyObject* args)
{
	char* name;
	PyObject* value = nullptr;

	if (!PyArg_ParseTuple(args, "s|O", &name, &value)) return NULL;

	if (!strcmp(name,"2D")) {
		if (value==nullptr)
			return PyBool_FromLong(self->viewer->d2.show);
		else
			self->viewer->d2.show = (bool)PyLong_AsLong(value);
	} else
	if (!strcmp(name,"3D")) {
		if (value==nullptr)
			return PyBool_FromLong(self->viewer->d3.show);
		else
			self->viewer->d3.show = (bool)PyLong_AsLong(value);
	} else
	if (!strcmp(name,"axes")) {
		if (value==nullptr)
			return PyBool_FromLong(self->viewer->decoration.showAxes);
		else
			self->viewer->decoration.showAxes = PyLong_AsLong(value);
	} else
	if (!strcmp(name, "aoc") ) { // ambient occlusion enable/disable and set number of samples
		if (value == nullptr)
			return PyBool_FromLong(self->viewer->d3.ambientOcclusionSamples());
		else
			self->viewer->d3.ambientOcclusionSamples(PyLong_AsLong(value));
	} else
	if (!strcmp(name,"antialias")) {
		if (value==nullptr)
			return PyLong_FromLong(self->viewer->d3.antialias);
		else
			self->viewer->d3.antialias = Range(1,(int)PyLong_AsLong(value),16);
	} else
	if (!strcmp(name,"ambient")) {
		if (value==nullptr)
			return PyLong_FromLong(self->viewer->d3.ambient);
		else
			self->viewer->d3.ambient = PyLong_AsLong(value) & 0xFF;
	} else
	if (!strcmp(name,"backgroundcolor")) {
		if (value==nullptr)
			return PyLong_FromLong(self->geometry->geometry->backgroundColor());
		else {
			self->viewer->backgroundColor(PyLong_AsLong(value));
		}
	} else
	if (!strcmp(name,"borderwidth") || !strcmp(name,"bd")) {	// viewport border
		if (value==nullptr)
			return PyBool_FromLong(self->borderWidth);
		else
			self->borderWidth = PyLong_AsLong(value);
	} else
	if (!strcmp(name,"bordercolor")) {
		if (value==nullptr)
			return PyBool_FromLong(self->borderColor);
		else
			self->borderColor = PyLong_AsLong(value);
	} else
	if (!strcmp(name,"clipbody")) {
		if (value==nullptr) {
//			if (self->kernel->clipbody())
//				return PyLong_FromLong(self->kernel->clipbody());
//			else
//				return PyLong_FromLong(-1);
		} else {
			int bid = PyLong_AsLong(value);
			if (bid<0)
				self->kernel->clipBodyClear();
			else
				self->kernel->clipBodyAdd(bid);
		}
	} else
	if (!strcmp(name,"clipnegative")) {
//		if (value==nullptr)
//			return PyBool_FromLong(self->kernel->clipNegative());
//		else
			self->kernel->clipBodyNegative(PyLong_AsLong(value));
	} else
	if (!strcmp(name,"crosshair")) {
		if (value==nullptr)
			return PyLong_FromLong(self->crosshair);
		else
			self->crosshair = PyLong_AsLong(value);
	} else
	if (!strcmp(name,"cores")) {
		if (value==nullptr)
			return PyLong_FromLong(self->kernel->nthreads());
		else
			self->kernel->initThreads(PyLong_AsLong(value));
	}  else
	if (!strcmp(name,"deflights")) {
		if (value==nullptr)
			return PyBool_FromLong(self->viewer->d3.deflights);
		else
			self->viewer->d3.deflights = (bool)PyLong_AsLong(value);
	} else
	if (!strcmp(name,"depth"))
		return PyLong_FromLong(self->depth);
	else
	if (!strcmp(name,"drawtime")) {
		if (value==nullptr)
			return PyLong_FromLong(self->viewer->d3.drawTime());
		else
			self->viewer->d3.drawTime(PyLong_AsLong(value));
	} else
	if (!strcmp(name,"edgedetect")) {
		if (value==nullptr)
			return PyBool_FromLong(self->viewer->d3.drawEdges);
		else
			self->viewer->d3.drawEdges = (bool)PyLong_AsLong(value);
	} else
	if (!strcmp(name,"errors")) {
		if (value==nullptr)
			return PyBool_FromLong(self->viewer->showErrors);
		else
			self->viewer->showErrors = PyLong_AsLong(value);
	} else
	if (!strcmp(name,"fill")) {
		if (value==nullptr)
			return PyBool_FromLong(self->viewer->d2.fillRegions);
		else
			self->viewer->d2.fillRegions = (bool)PyLong_AsLong(value);

	} else
	if (!strcmp(name,"image")) {
		if (value==nullptr)
			return PyLong_FromLong(self->viewer->image.show);
		else
			self->viewer->image.show = PyLong_AsLong(value);
	} else
	if (!strcmp(name,"grid")) {
		if (value==nullptr)
			return PyBool_FromLong(self->viewer->decoration.showGrid);
		else
			self->viewer->decoration.showGrid = PyLong_AsLong(value);
	} else
	if (!strcmp(name,"gridlevel")) {
		// Old values was from 0..32 (5bits) now is 8bits
		if (value==nullptr)
			return PyLong_FromLong(self->viewer->decoration.gridLevel>>3);
		else
			self->viewer->decoration.gridLevel = PyLong_AsLong(value)<<3;
	} else
	if (!strcmp(name,"latticelevel")) {
		// Old values was from 0..32 (5bits) now is 8bits
		if (value==nullptr)
			return PyLong_FromLong(self->viewer->lattice.level>>3);
		else {
			self->viewer->lattice.level = PyLong_AsLong(value)<<3;
			self->viewer->lattice.makeHashColor();
		}
	} else
	if (!strcmp(name,"labels")) {
		if (value==nullptr)
			return PyBool_FromLong(self->viewer->d2.showLabels);
		else
			self->viewer->d2.showLabels = (bool)PyLong_AsLong(value);
	} else
	if (!strcmp(name,"lattice")) {
		if (value==nullptr)
			return PyLong_FromLong(self->viewer->lattice.show);
		else
			self->viewer->lattice.show = PyLong_AsLong(value);
	} else
	if (!strcmp(name,"palette")) {
		if (value==nullptr)
			return PyBool_FromLong(self->viewer->palette.show);
		else
			self->viewer->palette.show = (bool)PyLong_AsLong(value);
	} else
	if (!strcmp(name,"projection")) {
		if (value==nullptr)
			return PyLong_FromLong((int)self->kernel->view.projection);
		else
			self->kernel->view.projection = (ViewPort::Projection)PyLong_AsLong(value);
	} else
	if (!strcmp(name,"projectbody")) {
		if (value==nullptr) {
			// return a list of projection bodies
			PyObject* obj = PyTuple_New(self->kernel->projectBodyCount());
			for (int i=0; i<self->kernel->projectBodyCount(); i++)
				PyTuple_SetItem(obj, i, PyLong_FromLong(self->kernel->projectBody(i)));
			return obj;
		} else {
			int val = PyLong_AsLong(value);
			if (val<0)
				self->kernel->projectBodyClear();
			else
				self->kernel->projectBodyAdd(val);
		}
	} else
	if (!strcmp(name,"reflections")) {
		if (value==nullptr) {
			if (self->viewer->d3.maxDepth())
				return PyLong_FromLong(self->viewer->d3.maxDepth());
			else
				return PyLong_FromLong(-1);
		} else
			self->viewer->d3.maxDepth(PyLong_AsLong(value));
	} else
	if (!strcmp(name,"shadows")) {
		if (value==nullptr)
			return PyBool_FromLong(self->viewer->d3.shadows);
		else
			self->viewer->d3.shadows = (bool)PyLong_AsLong(value);
	} else
	if (!strcmp(name,"skip1stblack")) {
		if (value==nullptr)
			return PyBool_FromLong(self->viewer->d3.skip_1stblack);
		else
			self->viewer->d3.skip_1stblack = (bool)PyLong_AsLong(value);
	} else
	if (!strcmp(name,"textbackground")) {
		if (value==nullptr)
			return PyBool_FromLong(self->viewer->textBackgroundLevel);
		else
			self->viewer->textBackgroundLevel = PyLong_AsLong(value);
	} else
	if (!strcmp(name,"title")) {
		if (value==nullptr)
			return PyBool_FromLong(self->viewer->decoration.showTitle);
		else
			self->viewer->decoration.showTitle = PyLong_AsLong(value);
	} else
	if (!strcmp(name,"trackball")) {
		if (value==nullptr)
			return PyBool_FromLong(self->showTrackball);
		else
			self->showTrackball = (bool)PyLong_AsLong(value);
	} else
	if (!strcmp(name,"usrbinastexture")) {
		if (value==nullptr)
			return PyBool_FromLong(self->viewer->d3.usrbinAsTexture());
		else
			self->viewer->d3.usrbinAsTexture((bool)PyLong_AsLong(value));
	} else
	if (!strcmp(name,"userdump")) {
		if (value==nullptr)
			return PyBool_FromLong(self->viewer->userdump.show);
		else
			self->viewer->userdump.show = (bool)PyLong_AsLong(value);
	} else
	if (!strcmp(name,"usrbin")) {
		if (value==nullptr)
			return PyLong_FromLong(self->viewer->usrbin.show);
		else
			self->viewer->usrbin.show = PyLong_AsLong(value);
	} else
	if (!strcmp(name,"vertex")) {
		if (value==nullptr)
			return PyBool_FromLong(self->viewer->d2.showVertex);
		else
			self->viewer->d2.showVertex = (bool)PyLong_AsLong(value);
	} else
	if (!strcmp(name,"viewport")) {
		if (value==nullptr)
			return PyLong_FromLong(self->showViewport);
		else
			self->showViewport = (bool)PyLong_AsLong(value);
	} else
	if (!strcmp(name,"voxel")) {
		if (value==nullptr)
			return PyLong_FromLong(self->viewer->voxel.show);
		else
			self->viewer->voxel.show = PyLong_AsLong(value);
	} else
	if (!strcmp(name,"voxellevel")) {
		// Old values was from 0..32 (5bits) now is 8bits
		if (value==nullptr)
			return PyLong_FromLong(self->viewer->voxel.level>>3);
		else {
			self->viewer->voxel.level = PyLong_AsLong(value)<<3;
			self->viewer->voxel.makeHashColor();
		}
	} else
	if (!strcmp(name,"xray")) {
		if (value==nullptr)
			return PyLong_FromLong(self->viewer->d3.xray);
		else
			self->viewer->d3.xray = Range(0,(int)PyLong_AsLong(value),255);
	} else {
		PyErr_Format(PyExc_SyntaxError, "'%s' is not a valid type option", name);
		return NULL;
	}
	Py_RETURN_NONE;
} // Viewer_set

/* --- Viewer_regionVar --- */
static PyObject* Viewer_regionVar(ViewerObject* self, VRegion* region,
				const char* var, PyObject* value)
{
	if (!strcmp(var, "id")) {
		if (value==NULL)
			return PyLong_FromLong(region->id());
		else {
			PyErr_SetString(PyExc_SyntaxError, "Cannot set region id");
			return NULL;
		}
	} else
	if (!strcmp(var, "color")) {
		if (value==NULL)
			return PyLong_FromLong(region->color() & FLAG_COLORMASK);
		else {
			region->value(0.0);	// FIXME Do I need it here ????
			region->color((dword)PyLong_AsUnsignedLongMask(value));
		}
	} else
	if (!strcmp(var, "value")) {
		if (value==NULL)
			return PyFloat_FromDouble(region->value());
		else {
			int pal = self->viewer->palette.def();
			// set color from the palette using the value
			region->value(PyFloat_AsDouble(value));
			region->color(self->viewer->palette[pal].color(region->value()));
		}
	} else
	if (!strcmp(var, "label")) {
		if (value==NULL)
			return PyUnicode_FromString(region->label);
		else {
			strncpy(region->label, PyUnicode_AsUTF8(value),
				sizeof(region->label));
			region->label[sizeof(region->label)-1] = 0;
		}
	} else
	if (!strcmp(var, "alpha")) {
		if (value==NULL)
			return PyLong_FromLong(region->alpha()&0xFF);
		else
			region->alpha(PyLong_AsLong(value));
	} else
	if (!strcmp(var, "correct")) {
		if (value==NULL)
			return PyBool_FromLong(self->kernel->correctOverlaps(region));
		else if (PyList_Check(value)) {
			if (PyList_GET_SIZE(value)==0)
				return PyBool_FromLong(self->kernel->correctOverlaps(region));
			bool modified = false;
			for (ssize_t i=0; i<PyList_GET_SIZE(value); i++) {
				int zoneid =  PyLong_AsLong(PyList_GetItem(value,i));
				modified |= self->kernel->correctOverlaps(region, zoneid);
			}
			return PyBool_FromLong(modified);
		}
	} else {
		PyErr_Format(PyExc_TypeError, "Invalid type \'%s\' specified", var);
		return NULL;
	}
	Py_RETURN_NONE;
} // Viewer_regionVar

/* --- Viewer_region --- */
static PyObject* Viewer_region(ViewerObject* self, PyObject* args)
{
	PyObject* obj;
	char* var;
	PyObject* value=NULL;
	if (!PyArg_ParseTuple(args, "Os|O", &obj, &var, &value)) return NULL;

	if (Py_Check4Pattern(obj)) {
		const char* pattern = PyUnicode_AsUTF8(obj);
		for (size_t ir=0; ir<self->kernel->nRegions(); ir++) {
			VRegion* region = self->kernel->regions[ir];
			if (!fnmatch(pattern, region->cname(), 0)) {
				PyObject* ret = Viewer_regionVar(self, region, var, value);
				Py_XDECREF(ret);	// ret can be NULL use XDECREF
			}
		}
		Py_RETURN_NONE;
	} else  {
		VRegion* region = Py_VRegion(self, obj);
		if (region==NULL) return NULL;
		return Viewer_regionVar(self, region, var, value);
	}
	Py_RETURN_NONE;		// To keep compiler happy
} // Viewer_region

/** Viewer_size set or get the size of the viewer */
static PyObject* Viewer_size(ViewerObject* self, PyObject* args)
{
	char* cmd = nullptr;
	int width=-1;
	int height=-1;
	if (!PyArg_ParseTuple(args, "|sii", &cmd, &width, &height)) return NULL;

	if (cmd==NULL or !strcmp(cmd,"get")) {
		// default return width, height
		assert(width==-1 and height==-1);
		return Py_BuildValue("ii", self->viewer->width(), self->viewer->height());
	} else
	if (!strcmp(cmd,"set")) {
		assert(width>0 and height>0);
		// otherwise fixed size
		self->fixedSize = true;
		if (width!=self->viewer->width() || height!=self->viewer->height()) {
			self->viewer->resize(width, height);
			self->ximage->width  = width;
			self->ximage->height = height;
			self->ximage->bytes_per_line = 0;
			self->ximage->data   = (char*)self->viewer->painter.data();
			XInitImage(self->ximage);
		}
	} else
	if (!strcmp(cmd,"dynamic") or !strcmp(cmd,"restore")) {
		// with set to 0 make it dynamic
		self->fixedSize = false;
		width  = Tk_Width(self->tkWin);
		height = Tk_Height(self->tkWin);
		if (width!=self->viewer->width() || height!=self->viewer->height()) {
			self->viewer->resize(width, height);
			self->ximage->width  = width;
			self->ximage->height = height;
			self->ximage->bytes_per_line = 0;
			self->ximage->data = (char*)self->viewer->painter.data();
			XInitImage(self->ximage);
		}
	} else
	if (!strcmp(cmd,"isfixed"))
		return PyBool_FromLong(self->fixedSize);
	else
		assert(false);

	Py_RETURN_NONE;
} // Viewer_size

/* --- Viewer_stopThread --- */
static PyObject* Viewer_stopThread(ViewerObject* self, PyObject* Py_UNUSED(ignored))
{
	DUMP(cout << "Viewer_stopThread("<<self->viewer->title()<<")" << endl);
	self->viewer->stopThread();
	Py_RETURN_NONE;
} // Viewer_stopThread

/* --- Viewer_title --- */
static PyObject* Viewer_title(ViewerObject* self, PyObject* args)
{
	char* title = NULL;
	if (!PyArg_ParseTuple(args, "|s", &title)) return NULL;
	if (title==NULL)
		return PyUnicode_FromString(self->viewer->title().c_str());
	else
		self->viewer->title(title);
	Py_RETURN_NONE;
} // Viewer_title

/* --- Viewer_state --- */
static PyObject* Viewer_state(ViewerObject* self, PyObject* args)
{
	char* name=NULL;

	if (!PyArg_ParseTuple(args, "|s", &name))
		return NULL;

	if (name==NULL)
		return PyLong_FromLong((int)self->viewer->state());
	else
	if (!strcmp(name,"%"))	// load usrbin detector
		return PyLong_FromLong((int)self->viewer->percent());
	else {
		PyErr_Format(PyExc_SyntaxError, "'%s' is not a valid option", name);
		return NULL;
	}
} // Viewer_state

/* --- Viewer_userdump --- */
static PyObject* Viewer_userdump(ViewerObject* self, PyObject* args)
{
	char*	name;
	PyObject* value=NULL;
	PyObject* opt=NULL;

	if (!PyArg_ParseTuple(args, "s|OO", &name, &value, &opt))
		return NULL;

	if (!strcmp(name,"file")) {	// load usrbin detector
		if (value!=NULL) {
			self->viewer->userdump.open(PyUnicode_AsUTF8(value));
		}
	} else
	if (!strcmp(name,"start")) {	// start event
		if (value!=NULL)
			self->viewer->userdump.start = PyLong_AsLong(value);
		else
			return PyLong_FromLong(self->viewer->userdump.start);
	} else
	if (!strcmp(name,"n")) {	// # events
		if (value!=NULL)
			self->viewer->userdump.n = PyLong_AsLong(value);
		else
			return PyLong_FromLong(self->viewer->userdump.n);
	} else
	if (!strcmp(name,"reset")) {	// # reset all information on particles
		self->viewer->userdump.reset();
	} else
	if (!strcmp(name,"show")) {	// particle to hide
		int n = PyLong_AsLong(value);
		self->viewer->userdump.display(n);
	} else
	if (!strcmp(name,"alpha")) {	// particle to hide
		if (value!=NULL)
			self->viewer->userdump.alpha = Range(0,(int)PyLong_AsLong(value),255);
		else
			return PyLong_FromLong(self->viewer->userdump.alpha);
	} else
	if (!strcmp(name,"hide")) {	// particle to hide
		int n = PyLong_AsLong(value);
		self->viewer->userdump.hide(n);
	} else
	if (!strcmp(name,"color")) {	// color of particle
		int n = PyLong_AsLong(value);
		if (value == NULL)
			return PyLong_FromLong(self->viewer->userdump.color(n));
		else
			self->viewer->userdump.color(n, PyLong_AsLong(opt));
	} else
	if (!strcmp(name,"emin")) {	// minimum energy
		int n = PyLong_AsLong(value);
		if (value == NULL)
			return PyFloat_FromDouble(self->viewer->userdump.emin(n));
		else
			self->viewer->userdump.emin(n, PyFloat_AsDouble(opt));
	} else
	if (!strcmp(name,"emax")) {	// maximum energy
		int n = PyLong_AsLong(value);
		if (value == NULL)
			return PyFloat_FromDouble(self->viewer->userdump.emax(n));
		else
			self->viewer->userdump.emax(n, PyFloat_AsDouble(opt));
	} else {
		PyErr_Format(PyExc_SyntaxError, "'%s' is not a valid option", name);
		return NULL;
	}

	Py_RETURN_NONE;
} // Viewer_userdump

/* --- Viewer_usrbin --- */
static PyObject* Viewer_usrbin(ViewerObject* self, PyObject* args)
{
	char* name;
	int index;
	PyObject* value=NULL, *opt=NULL;

	if (!PyArg_ParseTuple(args, "is|OO", &index, &name, &value, &opt))
		return NULL;

	if (!InRange(0,index,MAXUSRBIN)) {
		PyErr_SetString(PyExc_ValueError, "Invalid usrbin index");
		return NULL;
	}

	Usrbin& usrbin = self->viewer->usrbin[index];

	if (!strcmp(name,"load")) {	// load usrbin detector
		if (value != NULL && opt != NULL) {
			const char* filename = PyUnicode_AsUTF8(value);
			int det = (int)PyLong_AsLong(opt);
			if (!usrbin.load(filename, det)) {
				PyErr_Format(PyExc_IOError, "Unable to load usrbin detector \'%s\' %d",
						filename, det);
				return NULL;
			}

//			return PyUnicode_FromFormat(
//				"Usrbin: %s\n"
//				"Title:  %s\n"
//				"Time:   %s\n"
//				"Detector: %d %s\n"
//				"\tScore Type=%d Particle=%d\n"
//				"\tX: %.15g\t%.15g\t%d\t%.15g\n"
//				"\tY: %.15g\t%.15g\t%d\t%.15g\n"
//				"\tZ: %.15g\t%.15g\t%d\t%.15g\n",
//				filename,
//				usrbin.runtitle,
//				usrbin.runtime,
//				usrbin.detector(), usrbin.titusb,
//				usrbin.type(), usrbin.score(),
//				usrbin.xlow, usrbin.xhigh, usrbin.nx, usrbin.dx,
//				usrbin.ylow, usrbin.yhigh, usrbin.ny, usrbin.dy,
//				usrbin.zlow, usrbin.zhigh, usrbin.nz, usrbin.dz);
		}
	} else
	if (!strcmp(name,"alpha")) {	// set usrbin transparency
		if (value==NULL)
			return PyLong_FromLong(self->viewer->usrbin.alpha(index));
		else
			self->viewer->usrbin.alpha(index, (int)PyLong_AsLong(value));
	} else
	if (!strcmp(name,"palette")) {
		if (value==NULL)
			return PyLong_FromLong(self->viewer->usrbin.palette(index));
		else
			self->viewer->usrbin.palette(index, (int)PyLong_AsLong(value));
	} else
	if (!strcmp(name,"cleanup"))
		usrbin.cleanup();
	else
	if (!strcmp(name,"norm")) {	// set normalization factor
		if (value==NULL)
			return PyFloat_FromDouble((double)usrbin.norm());
		else
			usrbin.norm((float)PyFloat_AsDouble(value));
	} else
	if (!strcmp(name,"log")) {	// set normalization factor
		if (value==NULL)
			return PyBool_FromLong(usrbin.logscale());
		else
			usrbin.convert(PyObject_IsTrue(value));
	} else
	if (!strcmp(name,"matrix")) {	// set transformation matrix
		if (value) {
			if (PyList_Check(value)) {
				Matrix4	matrix;
				PyList_AsMatrix4(value, matrix);
				usrbin.matrix(matrix);
			} else {
				usrbin.clearMatrix();
			}
		} //else return matrix
	} else
	if (!strcmp(name,"offset")) {	// set offset
		if (value)
			self->viewer->usrbin[index].offset(Py_GetVector(value));
	} else
	if (!strcmp(name,"regioncolor")) {
		self->viewer->usrbin.regionColorFromUsrbin(index);
	} else
	if (!strcmp(name,"checker")) {	// set checker usrbin
		if (value!=NULL) {
			if (PyTuple_Check(value) && PyTuple_GET_SIZE(value)==10) {
				Usrbin::Type it = (Usrbin::Type)PyLong_AsLong(PyTuple_GetItem(value, 0));
				double xl = PyFloat_AsDouble(PyTuple_GetItem(value, 1));
				double xh = PyFloat_AsDouble(PyTuple_GetItem(value, 2));
				int    nx = PyLong_AsLong(PyTuple_GetItem(value, 3));

				double yl = PyFloat_AsDouble(PyTuple_GetItem(value, 4));
				double yh = PyFloat_AsDouble(PyTuple_GetItem(value, 5));
				int    ny = PyLong_AsLong(PyTuple_GetItem(value, 6));

				double zl = PyFloat_AsDouble(PyTuple_GetItem(value, 7));
				double zh = PyFloat_AsDouble(PyTuple_GetItem(value, 8));
				int    nz = PyLong_AsLong(PyTuple_GetItem(value, 9));
				usrbin.checker(it, xl, xh, nx, yl, yh, ny, zl, zh, nz);
			} else
				PyErr_SetString(PyExc_SyntaxError, "for 'checker' a tuple of length 10 expected");
		}
	} else
	if (!strcmp(name,"type")) {
		if (value==NULL)
			return PyLong_FromLong((int)usrbin.type());
	} else
	if (!strcmp(name,"score")) {
		if (value==NULL)
			return PyLong_FromLong(usrbin.score());
	} else
	if (!strcmp(name,"bins")) {
		if (value==NULL)
			return Py_BuildValue("ii",
					usrbin.nx,
					usrbin.ny,
					usrbin.nz);
	} else
	if (!strcmp(name,"low")) {
		if (value==NULL)
			return Py_BuildValue("ddd",
					usrbin.xlow,
					usrbin.ylow,
					usrbin.zlow);
	} else
	if (!strcmp(name,"high")) {
		if (value==NULL)
			return Py_BuildValue("ddd",
					usrbin.xhigh,
					usrbin.yhigh,
					usrbin.zhigh);
	} else
	if (!strcmp(name,"min")) {
		if (value==NULL)
			return PyFloat_FromDouble((double)usrbin.min);
	} else
	if (!strcmp(name,"max")) {
		if (value==NULL)
			return PyFloat_FromDouble((double)usrbin.max);
	} else
	if (!strcmp(name,"value")) {
		if (value!=NULL &&
		    usrbin.hasData() &&
		    PyTuple_Check(value) &&
		    PyTuple_GET_SIZE(value)==3) {
			double x = PyFloat_AsDouble(PyTuple_GetItem(value, 0));
			double y = PyFloat_AsDouble(PyTuple_GetItem(value, 1));
			double z = PyFloat_AsDouble(PyTuple_GetItem(value, 2));
			double val;
			if (usrbin.get(x, y, z, &val))
				return PyFloat_FromDouble(val);
		}
	} else {
		PyErr_Format(PyExc_SyntaxError, "'%s' is not a valid option", name);
		return NULL;
	}

	Py_RETURN_NONE;
} // Viewer_usrbin

/* --- Viewer_uv2pixel --- */
static PyObject* Viewer_uv2pixel(ViewerObject* self, PyObject* args)
{
	double u,v;

	if (!PyArg_ParseTuple(args, "dd", &u, &v)) return NULL;
	return Py_BuildValue("ii",	self->kernel->view.u2i(u),
					self->kernel->view.v2j(v));
} // Viewer_uv2pixel

/* --- Viewer_uv2xyz --- */
static PyObject* Viewer_uv2xyz(ViewerObject* self, PyObject* args)
{
	double u,v;

	if (!PyArg_ParseTuple(args, "dd", &u, &v)) return NULL;

	// FIXME u,v should be relative to the center of the screen!!!
	double x = self->kernel->view.uv2x(u,v);
	double y = self->kernel->view.uv2y(u,v);
	double z = self->kernel->view.uv2z(u,v);

	return Py_BuildValue("ddd", x,y,z);
} // Viewer_uv2xyz

/* --- Viewer_vertex --- */
static PyObject* Viewer_vertex(ViewerObject* self, PyObject* args)
{
	char* axis;
	if (!PyArg_ParseTuple(args, "s", &axis)) return NULL;
	Array<double> vertices;

	if (('u'<=axis[0] && axis[0]<='z') || ('U'<=axis[0] && axis[0]<='Z')) {
		if (!self->viewer->d2.projectVertices(axis[0], vertices)) {
			PyErr_SetString(PyExc_SyntaxError, "Error getting vertices");
			return NULL;
		}
		double prev = -MAX_REAL;
		PyObject* vertexList = PyList_New(0);
		for (int i=0; i<vertices.size(); i++) {
			double& v = vertices[i];
			if (Abs(v-prev)>SMALL5)
				PyList_Append(vertexList, PyFloat_FromDouble(v));
			prev = v;
		}
		return vertexList;
	} else {
		PyErr_SetString(PyExc_ValueError, "Axis string expected as argument");
		return NULL;
	}
} // Viewer_vertex

/* --- Viewer_viewOto0 --- *
 * Move view origin to center of "view"
 */
static PyObject* Viewer_viewOto0(ViewerObject* self, PyObject* Py_UNUSED(ignored))
{
	self->viewer->moveViewOriginTo0();
	Py_RETURN_NONE;
} // Viewer_viewOto0

/* --- Viewer_viewport --- *
 * Manipulate viewport lines
 */
static PyObject* Viewer_viewport(ViewerObject* self, PyObject* args)
{
	int id;
	char* name;
	PyObject* value=NULL;
	if (!PyArg_ParseTuple(args, "is|O", &id, &name, &value)) return NULL;
	if (id<0 || id>=ViewerObject::nviews) {
		PyErr_SetString(PyExc_KeyError, "Viewport not found");
		return NULL;
	}
	if (!strcmp(name, "centerview")) {
		if (value != NULL) {
			PyErr_SetString(PyExc_TypeError, "centerview is readonly");
			return NULL;
		} else
			return Py_BuildValue("dd",
					self->viewport[id].uc,
					self->viewport[id].vc);
	} else
	if (!strcmp(name, "centerpixel")) {
		if (value != NULL) {
			PyErr_SetString(PyExc_TypeError, "centerpixel is readonly");
			return NULL;
		} else
			return Py_BuildValue("ii",
					Round(self->viewport[id].xc),
					Round(self->viewport[id].yc));
	} else
	if (!strcmp(name, "viewer")) {
		if (value != NULL) {
			if (Py_TYPE(value) != &ViewerType) {
				PyErr_SetString(PyExc_TypeError, "Invalid type, Viewer expected");
				return NULL;
			}
			self->viewport[id].viewer = ((ViewerObject*)value)->viewer;
		} else
			return PyList_FromMatrix4(self->viewport[id].viewer->view().matrix());
	} else if (!strcmp(name, "linewidth")) {
		if (value != NULL)
			self->viewport[id].lineWidth = (int)PyLong_AsLong(value);
		else
			return PyLong_FromLong(self->viewport[id].lineWidth);

	} else if (!strcmp(name, "originwidth")) {
		if (value != NULL)
			self->viewport[id].originWidth = (int)PyLong_AsLong(value);
		else
			return PyLong_FromLong(self->viewport[id].originWidth);

	} else if (!strcmp(name, "wlength")) {
		if (value != NULL)
			self->viewport[id].wLength = PyFloat_AsDouble(value);
		else
			return PyFloat_FromDouble(self->viewport[id].wLength);

	} else if (!strcmp(name, "color")) {
		if (value != NULL)
			self->viewport[id].color = (int)PyLong_AsLong(value);
		else
			return PyLong_FromLong(self->viewport[id].color);
	} else {
		PyErr_Format(PyExc_SyntaxError, "'%s' is not a valid type option", name);
		return NULL;
	}
	Py_RETURN_NONE;
} // Viewer_viewport

/* --- Viewer_voxel --- */
static PyObject* Viewer_voxel(ViewerObject* self, PyObject* args)
{
	char* name;
	PyObject* value=NULL, *opt=NULL;

	if (!PyArg_ParseTuple(args, "s|OO", &name, &value, &opt)) return NULL;

	// Prepare Viewer voxel if needed
	self->viewer->voxel().allocate();

	if (!strcmp(name, "color")) {	// set color of voxel
		if (value != NULL) {
			if (opt != NULL) {
				self->viewer->voxel().color((int)PyLong_AsLong(value),
						(dword)PyLong_AsUnsignedLongMask(opt));
			} else
				return PyLong_FromLong(self->viewer->voxel().color(
						(int)PyLong_AsLong(value)-1));
		}
	} else
	if (!strcmp(name, "value")) {
		if (value != NULL) {
			if (opt==NULL) {
				PyErr_SetString(PyExc_SyntaxError, "Cannot get voxel color value");
				return NULL;
			} else {
				// set color from the palette using the value
				int pal = self->viewer->palette.def();
				self->viewer->voxel().color((int)PyLong_AsLong(value),
						self->viewer->palette[pal].color(PyFloat_AsDouble(opt)));
			}
		}
	} else
	if (!strcmp(name, "roi")) {
		if (value != NULL) {
			if (PyLong_Check(value)) {
				int roi = PyLong_AsLong(value);
				self->viewer->voxel().roiShow(roi, PyLong_AsLong(opt));
			}
		}
	} else
	if (!strcmp(name, "roiclear")) {
		self->viewer->voxel().roiShowClear();
	} else
	if (!strcmp(name, "roialpha")) {
		if (value != NULL && PyLong_Check(value))
			self->viewer->voxel().roiAlpha(PyLong_AsLong(value));
		else
			return PyLong_FromLong(self->viewer->voxel().roiAlpha());
	} else {
		PyErr_Format(PyExc_SyntaxError, "'%s' is not a valid option", name);
		return NULL;
	}

	Py_RETURN_NONE;
} // Viewer_voxel

/* --- Viewer_where --- */
static PyObject* Viewer_where(ViewerObject* self, PyObject* args)
{
	double x, y, z;

	if (!PyArg_ParseTuple(args, "ddd", &x, &y, &z)) return NULL;

	Point pos(x,y,z);
	Vector dir(-self->kernel->view.invMatrix()(2,0),
		   -self->kernel->view.invMatrix()(2,1),
		   -self->kernel->view.invMatrix()(2,2));

	// FIXME MAX lattices
	VZone* zone = NULL;
	for (int i=0; i<10; i++) {
		self->kernel->lock();
		self->kernel->engine()->incBodyCheckId();
		zone = self->kernel->engine()->where(pos.x,pos.y,pos.z, dir.x,dir.y,dir.z, NULL);
		self->kernel->unlock();

		if (zone==NULL)
			Py_RETURN_NONE;
		else
		if (!zone->gregion()->hasMatrix())
			return PyUnicode_FromString(zone->gregion()->cname());
		else {
			pos = zone->gregion()->matrix() * pos;
			dir = zone->gregion()->matrix() * dir;
			dir.renormalize();
		}
	}
	return PyLong_FromLong(zone->gregion()->id());
} // Viewer_where

/* --- Viewer_xyz2uv --- */
static PyObject* Viewer_xyz2uv(ViewerObject* self, PyObject* args)
{
	double x,y,z;

	if (!PyArg_ParseTuple(args, "ddd", &x, &y, &z)) return NULL;

	double u,v;
	// FIXME should be relative to the center of the screen!!!
	self->kernel->view.xyz2uv(x,y,z,&u,&v);

	return Py_BuildValue("dd", u,v);
} // Viewer_xyz2uv

/* --- Viewer_xyz2uvw --- */
static PyObject* Viewer_xyz2uvw(ViewerObject* self, PyObject* args)
{
	double x,y,z;

	if (!PyArg_ParseTuple(args, "ddd", &x, &y, &z)) return NULL;

	double u,v,w;

	self->kernel->view.xyz2uvw(x,y,z,&u,&v,&w);

	return Py_BuildValue("ddd", u,v,w);
} // Viewer_xyz2uvw

/* --- Viewer_zoom --- */
static PyObject* Viewer_zoom(ViewerObject* self, PyObject* args)
{
	double zoom=-1.0;

	if (!PyArg_ParseTuple(args, "|d", &zoom)) return NULL;
	if (zoom<0.0)
		return PyFloat_FromDouble(self->kernel->view.zoom());
	else
		self->kernel->view.zoom(zoom);
	Py_RETURN_NONE;
} // Viewer_zoom

/** Manipulate selected zone */
static PyObject* Viewer_zone(ViewerObject* self, PyObject* args)
{
	char* cmd;
	PyObject* value=NULL;

	if (!PyArg_ParseTuple(args, "s|O", &cmd, &value)) return NULL;

	if (!strcmp(cmd, "find")) {
		if (value==NULL) {
			PyErr_SetString(PyExc_SyntaxError, "tuple expected with (u,v,mask)");
			return NULL;
		}
		if (PyTuple_Check(value) && PyTuple_GET_SIZE(value)==3) {
			double u = Py_GetFloat(PyTuple_GetItem(value,0));
			double v = Py_GetFloat(PyTuple_GetItem(value,1));
			dword mask = (dword)Py_GetInt(PyTuple_GetItem(value,2));

			GZone zone(nullptr, false);
			// Do I need a lockRead/Write?
			self->kernel->engine()->incBodyCheckId();

			const ViewPort& V = self->kernel->view;
			const Matrix4& M = V.matrix();

			// First pass add only the positive bodies
			for (auto body : self->geometry->geometry->bodies) {
				if (!(body->show & mask)) continue;
				VBody* vbody = self->kernel->getBody(body);
				double x  = V.uv2x(u,v);
				double y  = V.uv2y(u,v);
				double z  = V.uv2z(u,v);
				if (!vbody->inside2D(x,y,z,-M(0,2),-M(1,2),-M(2,2))) continue;
				zone.add(body->cname(), body);
			}
			zone.add("-");

			// Second pass add the negative terms
			for (auto body : self->geometry->geometry->bodies) {
				if (!(body->show & mask)) continue;
				VBody* vbody = self->kernel->getBody(body);
				double x  = V.uv2x(u,v);
				double y  = V.uv2y(u,v);
				double z  = V.uv2z(u,v);
				if (vbody->inside2D(x,y,z, -M(0,2),-M(1,2),-M(2,2))) continue;
				zone.add(body->name(), body);
			}

			// Optimize the zones expression
			zone.optimize();

			// empty? only $?
			if (zone.size()==0 or
			   (zone.size()==1 and zone[0]->isOperator())) Py_RETURN_NONE;
			return Py_ZoneExpr(&zone);
		} else {
			PyErr_SetString(PyExc_SyntaxError, "tuple expected with (u,v,mask)");
			return NULL;
		}
	} else {
		PyErr_Format(PyExc_TypeError, "Invalid type \'%s\' specified", cmd);
		return NULL;
	}

	Py_RETURN_NONE;
} // Viewer_zone

/* ----------------- Viewer type ------------------- */
static PyMethodDef Viewer_methods[] = {
	{"Sx",		(PyCFunction)Viewer_Sx,		METH_NOARGS,	"Return Sx"},
	{"Sy",		(PyCFunction)Viewer_Sy,		METH_NOARGS,	"Return Sy"},
	{"aspect",	(PyCFunction)Viewer_aspect,	METH_VARARGS,	"Return/Set aspect"},
	{"basis",	(PyCFunction)Viewer_basis,	METH_VARARGS,	"Return/Set basis"},
	{"bbox",	(PyCFunction)Viewer_bbox,	METH_VARARGS,	"Return the transformed(Viewer) 3D bounding box of body/zone/region in absolute coordinates"},
	{"bbox2D",	(PyCFunction)Viewer_bbox2D,	METH_VARARGS,	"Return the 2D bounding box of body/zone/region"},
	{"body",	(PyCFunction)Viewer_body,	METH_VARARGS,	"Modify body on viewer plane"},
	{"calcWindow",	(PyCFunction)Viewer_calcWindow,	METH_VARARGS,	"Calculate and set correct window limits"},
	{"camera",	(PyCFunction)Viewer_camera,	METH_VARARGS,	"Return 3D camera parameters"},
	{"closest",	(PyCFunction)Viewer_closest,	METH_VARARGS,	"Return closest objects"},
	{"configure",	(PyCFunction)Viewer_configure,	METH_VARARGS,	"Configure"},
	{"derive",	(PyCFunction)Viewer_derive,	METH_NOARGS,	"Derive viewer geometry from Geometry"},
	{"draw",	(PyCFunction)Viewer_draw,	METH_VARARGS,	"Start a drawing thread of viewer"},
	{"duv2dxyz",	(PyCFunction)Viewer_duv2dxyz,	METH_VARARGS,	"Convert d(relative) to d(absolute)"},
	{"edit",	(PyCFunction)Viewer_edit,	METH_VARARGS,	"Get/Set editing body"},
	{"enclosed",	(PyCFunction)Viewer_enclosed,	METH_VARARGS,	"Return objects enclosed inside region"},
	{"error",	(PyCFunction)Viewer_error,	METH_VARARGS,	"Return number of error"},
	{"export",	(PyCFunction)Viewer_export,	METH_VARARGS,	"Export viewport to vector file"},
	{"expose",	(PyCFunction)Viewer_expose,	METH_VARARGS,	"Expose image to X11"},
	{"extends",	(PyCFunction)Viewer_extends,	METH_VARARGS,	"Get/Set window/extend dimensions"},
	{"font",	(PyCFunction)Viewer_font,	METH_VARARGS,	"load font file"},
	{"get",		(PyCFunction)Viewer_set,	METH_VARARGS,	"Return flag/parameter"},
	{"grid",	(PyCFunction)Viewer_grid,	METH_VARARGS,	"Return/set grid settings"},
	{"hit",		(PyCFunction)Viewer_hit,	METH_VARARGS,	"Return hit position in 3D"},
	{"image",	(PyCFunction)Viewer_image,	METH_VARARGS,	"Get/Set image data"},
	{"invalid",	(PyCFunction)Viewer_invalid,	METH_NOARGS,	"Return true if window is invalid"},
	{"matrix",	(PyCFunction)Viewer_matrix,	METH_VARARGS,	"Return/set transformation matrix"},
	{"memory",	(PyCFunction)Viewer_memory,	METH_VARARGS,	"Get or print memory"},
	{"moveUV",	(PyCFunction)Viewer_moveUV,	METH_VARARGS,	"Move an object on uv plane by preserving w"},
	{"object",	(PyCFunction)Viewer_object,	METH_VARARGS,	"Modify object on viewer plane"},
	{"ocube",	(PyCFunction)Viewer_ocube,	METH_VARARGS,	"Orientation cube manipulation"},
	{"offset",	(PyCFunction)Viewer_offset,	METH_VARARGS,	"Return/Set projection offset (in UV)"},
	{"origin",	(PyCFunction)Viewer_origin,	METH_VARARGS,	"Return/Set projection origin"},
	{"palette",	(PyCFunction)Viewer_palette,	METH_VARARGS,	"Return/set palette"},
	{"pen",         (PyCFunction)Viewer_pen,	METH_VARARGS,	"Return/set pen drawing information"},
	{"pixel2uv",	(PyCFunction)Viewer_pixel2uv,	METH_VARARGS,	"Convert image pixel coordinates (i,j) to viewport (u,v)"},
	{"pixel2xyz",	(PyCFunction)Viewer_pixel2xyz,	METH_VARARGS,	"Convert image pixel coordinates (i,j) to real (x,y,z)"},
	{"planner",	(PyCFunction)Viewer_planner,	METH_VARARGS,	"Calculate integral of usrbin along a path"},
	{"project",	(PyCFunction)Viewer_project,	METH_VARARGS,	"Perform projection of viewer"},
	{"projectionChanged",(PyCFunction)Viewer_projectionChanged,METH_VARARGS,"Return/set projection change status"},
	{"rectangle",	(PyCFunction)Viewer_rectangle,	METH_VARARGS,	"Draw rectangle"},
	{"region",	(PyCFunction)Viewer_region,	METH_VARARGS,	"Set/Get region properties"},
	{"set",		(PyCFunction)Viewer_set,	METH_VARARGS,	"Return/set flag/parameter"},
	{"size",	(PyCFunction)Viewer_size,	METH_VARARGS,	"Set/Get viewer size"},
	{"snap",	(PyCFunction)Viewer_snap,	METH_VARARGS,	"Return closest rounded position aligned to grid"},
	{"state",	(PyCFunction)Viewer_state,	METH_VARARGS,	"Return thread state"},
	{"stopThread",	(PyCFunction)Viewer_stopThread,	METH_NOARGS,	"Stop projection/drawing thread"},
	{"title",	(PyCFunction)Viewer_title,	METH_VARARGS,	"Return/Set viewport title"},
	{"userdump",	(PyCFunction)Viewer_userdump,	METH_VARARGS,	"Set/Get userdump information"},
	{"usrbin",	(PyCFunction)Viewer_usrbin,	METH_VARARGS,	"Set/Get usrbin information"},
	{"uv2pixel",	(PyCFunction)Viewer_uv2pixel,	METH_VARARGS,	"Convert relative to image coordinates"},
	{"uv2xyz",	(PyCFunction)Viewer_uv2xyz,	METH_VARARGS,	"Convert relative to absolute"},
	{"vertex",	(PyCFunction)Viewer_vertex,	METH_VARARGS,	"Return all vertices projected on an axis"},
	{"viewOto0",	(PyCFunction)Viewer_viewOto0,	METH_NOARGS,	"Move view origin to 0"},
	{"viewport",	(PyCFunction)Viewer_viewport,	METH_VARARGS,	"Set/Get other viewport matrices"},
	{"voxel",	(PyCFunction)Viewer_voxel,	METH_VARARGS,	"Set/Get voxel information"},
	{"where",	(PyCFunction)Viewer_where,	METH_VARARGS,	"Get zone at location"},
	{"xyz2uv",	(PyCFunction)Viewer_xyz2uv,	METH_VARARGS,	"Convert absolute to relative"},
	{"xyz2uvw",	(PyCFunction)Viewer_xyz2uvw,	METH_VARARGS,	"Convert absolute to relative"},
	{"zone",	(PyCFunction)Viewer_zone,	METH_VARARGS,	"Manipulate selected zone"},
	{"zoom",	(PyCFunction)Viewer_zoom,	METH_VARARGS,	"Return/Set zoom level"},
//	{"lock",	(PyCFunction)Viewer_lock,	METH_NOARGS,	"Lock mutex"},
//	{"unlock",	(PyCFunction)Viewer_unlock,	METH_NOARGS,	"UnLock mutex"},
#ifdef MEM
	{"_destroy",	(PyCFunction)Viewer_destroy,	METH_NOARGS,	"Deallocate"},
#endif
	{NULL, NULL, 0, NULL}	// Sentinel
}; // Viewer_methods

PyTypeObject ViewerType = {
	PyVarObject_HEAD_INIT(NULL, 0)
	"geoviewer.Viewer",		// tp_name
	sizeof(ViewerObject),		// tp_basicsize
	0,				// tp_itemsize
	(destructor)Viewer_dealloc,	// tp_dealloc
	0,				// tp_print < 3.8 else tp_vectorcall_offset
	0,				// tp_getattr
	0,				// tp_setattr
	0,				// tp_compare
	0,				// tp_repr
	0,				// tp_as_number
	0,				// tp_as_sequence
	0,				// tp_as_mapping
	0,				// tp_hash
	0,				// tp_call
	0,				// tp_str
	0,				// tp_getattro
	0,				// tp_setattro
	0,				// tp_as_buffer
	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,	// tp_flags
	"Geometry Viewer object",	// tp_doc
	0,				// tp_traverse
	0,				// tp_clear
	0,				// tp_richcompare
	0,				// tp_weaklistoffset
	0,				// tp_iter
	0,				// tp_iternext
	Viewer_methods,			// tp_methods
	0,//Viewer_members,		// tp_members
	0,				// tp_getset
	0,				// tp_base
	0,				// tp_dict
	0,				// tp_descr_get
	0,				// tp_descr_set
	0,				// tp_dictoffset
	(initproc)Viewer_init,		// tp_init
	0,				// tp_alloc
	Viewer_new,			// tp_new
	0,				// tp_free
	0,				// tp_is_gc
	0,				// tp_bases
	0,				// tp_mro
	0,				// tp_cache
	0,				// tp_subclasses
	0,				// tp_weaklist
	0,				// tp_del
	0,				// tp_version_tag
	0,				// tp_finalize
#if PY_VERSION_HEX >= 0x03080000
	0,				// tp_vectorcall
#if PY_VERSION_HEX < 0x03090000
	0,				// tp_print (deprecated)
#endif
#endif
}; // ViewerType
