/*
 * Copyright European Organization for Nuclear Research (CERN)
 * All rights reserved
 *
 * Please look at the supplied documentation for the user's
 * license
 *
 * Author:	Vasilis.Vlachoudis@cern.ch
 * Date:	05-Apr-2011
 *
 * TODO
 * GObject has to be split in two GObject and VObject
 * otherwise we will have problems when parallel drawings are done
 */

#ifndef __XDRAW_HXDRAW_H
#define __XDRAW_HXDRAW_H

#include <assert.h>
#include <X11/X.h>
#include <X11/Xutil.h>
//#include <XRender.h>	for nicer drawing with antialias, transparencies and gradients

class Point;
class Matrix4;
class ViewPort;

// =================================== XDraw ===================================
// Wrapper for X functions
class XDraw {
public:
	Display*	display;	// graphics display
//	Drawable	drawable;	// drawable object
	GC		gc;		// graphics context

private:
	XGCValues	gcValues;	// for updating the context

	Pixmap		transparent25;	// bitmap as fill style 25% fake transparency
	Pixmap		transparent50;	// pixmap as fill style 50% fake transparency
	Pixmap		transparent75;	// pixmap as fill style 75% fake transparency

//	Font		font;		// common font for all text operations
//			dashes		// dashed pattern

public:
	XDraw() : display(NULL), gc(NULL) {/*FIXME createGC*/}
	~XDraw();

	bool	init(Drawable drawable);
//	void	free();

	// Style
	int	changeGC(unsigned long valuemask)
			{ return XChangeGC(display, gc, valuemask, &gcValues); }

	int	changeGC(unsigned long valuemask, XGCValues* values)
			{ return XChangeGC(display, gc, valuemask, values); }

	int	setFunction(unsigned function) {
				gcValues.function = function;
				return XChangeGC(display, gc, GCFunction, &gcValues);
			}

	int	setLineSolid() {
				gcValues.line_style = LineSolid;
				return changeGC(GCLineStyle);
			}
	int	setLineDashed() {
				gcValues.line_style = LineOnOffDash;
				return changeGC(GCLineStyle);
			}
	int	setLineWidth(int width) {
				gcValues.line_width = width;
				return changeGC(GCLineWidth);
			}
	int	setDashes(int dash_offset, const char dash_list[], int n)
			{ return XSetDashes(display, gc, dash_offset, dash_list, n); }

	void	setFillTransparency(int value);
	void	setFillSolid()
			{ XSetFillStyle(display, gc, FillSolid); }

	// GC and style
	int	foreground(unsigned long color) { return XSetForeground(display, gc, color); }
	int	background(unsigned long color) { return XSetBackground(display, gc, color); }

	// Points
	int	point(Drawable drawable, int x, int y)
			{ return XDrawPoint(display, drawable, gc, x, y); }

	int	points(Drawable drawable, XPoint* pts, int npts, int mode)
			{ return XDrawPoints(display, drawable, gc, pts, npts, mode); }

	// Lines
	int	line(Drawable drawable, int x1, int y1, int x2, int y2)
			{ return XDrawLine(display, drawable, gc, x1, y1, x2, y2); }

	int	lines(Drawable drawable, XPoint* pts, int npts, int mode)
			{ return XDrawLines(display, drawable, gc, pts, npts, mode); }

	int	segments(Drawable drawable, XSegment* segs, int nsegs)
			{ return XDrawSegments(display, drawable, gc, segs, nsegs); }

	// Rectangles
	int	rectangle(Drawable drawable, int x, int y, unsigned int width, unsigned int height)
			{ return XDrawRectangle(display, drawable, gc, x, y, width, height); }

	int	rectangles(Drawable drawable, XRectangle rects[], int nrects)
			{ return XDrawRectangles(display, drawable, gc, rects, nrects); }

	// Arcs
	int	arc(Drawable drawable, int x, int y, unsigned int width, unsigned int height, int angle1, int angle2)
			{ return XDrawArc(display, drawable, gc, x, y, width, height, angle1, angle2); }

	int	arcs(Drawable drawable, XArc* as, int narcs)
			{ return XDrawArcs(display, drawable, gc, as, narcs); }

	// Fill
	int	fillRectangle(Drawable drawable, int x, int y, unsigned int width, unsigned int height)
			{ return XFillRectangle(display, drawable, gc, x, y, width, height); }

	int	fillRectangles(Drawable drawable, XRectangle* rects, int nrects)
			{ return XFillRectangles(display, drawable,  gc, rects, nrects); }

	int	fillPolygon(Drawable drawable, XPoint* pts, int npts, int shape, int mode)
			{ return XFillPolygon(display, drawable,  gc, pts, npts, shape, mode); }

	int	fillArc(Drawable drawable, int x, int y, unsigned int width, unsigned int height, int angle1, int angle2)
			{ return XFillArc(display, drawable,  gc, x, y, width, height, angle1, angle2); }

	int	fillArcs(Drawable drawable, XArc* as, int narcs)
			{ return XFillArcs(display, drawable,  gc, as, narcs); }

	// Text
	int	string(Drawable drawable, int x, int y, const char* str, int length)
			{ return XDrawString(display, drawable, gc, x, y, str, length); }

	int	string16(Drawable drawable,int x, int y, const XChar2b* str, int length)
			{ return XDrawString16(display, drawable, gc, x, y, str, length); }

	// specialized functions
	void	line3D(Drawable drawable, const ViewPort& view, const Point& a, const Point& b);

	void	axes(Drawable drawable,
				const int x, const int y, const int size,
				const Matrix4& matrix, bool xyz=true);

	void	ellipse(Drawable drawable,
				const int x, const int y,
				const int major, const int minor,
				double rotation , double from, double to);

	void	trackball(Drawable drawable,
				const int x, const int y, const int radius,
				const Matrix4& matrix);

//	int	arrow(Drawable drawable, );
}; // class XDraw
#endif
